#
# This file includes and calculates some make variables that are useful
# when compiling LiS and utilities.
# The makefile that includes this one will have to define LIS_HOME to the
# root of the LiS tree.
#
# It must not contain any rules.
#

#
# Get LiS configuration, if it exists.
#
-include $(LIS_HOME)/config.in

#
# These variables could be configurable some day
#
INST_LIB = /usr/lib
INST_BIN = /usr/bin
INST_SBIN = /usr/sbin

#
# Calculate LIS_TARG
#
# This is the name of the target-specific subdirectory used when compiling.
#
ifeq ($(CONFIG_STREAMS),m)
LIS_TARG = linux
endif

ifeq ($(CONFIG_STREAMS),k)
LIS_TARG = linux
endif

ifeq ($(CONFIG_STREAMS),u)
LIS_TARG = user
endif

ifeq ($(CONFIG_STREAMS),q)
LIS_TARG = qnx
endif

#
# See if modules are supported in the kernel
#
ifdef KMODULES
ifeq ($(KMODULES),y)
CONFIG_MODULES=y
ifdef KMODVERS
ifeq ($(KMODVERS),y)
CONFIG_MODVERSIONS=y
endif
endif
endif
endif

#
# This is where we have LiS specific includes
#
LIS_INCL = $(LIS_HOME)/include

#
# Calculate the names of some often used directories.
#
LIBDIR  = $(LIS_HOME)/libc
UTILDIR = $(LIS_HOME)/util
HEADDIR = $(LIS_HOME)/head
DRVRDIR = $(LIS_HOME)/drivers/str

LIBOBJ  = $(LIBDIR)/$(LIS_TARG)
UTILOBJ = $(UTILDIR)/$(LIS_TARG)
HEADOBJ = $(HEADDIR)/$(LIS_TARG)
DRVROBJ = $(DRVRDIR)/$(LIS_TARG)

#
# Commonly used options
#
# These warning options should always be used within LiS
#
XOPTS = -Wall -Wstrict-prototypes -Wid-clash-32 -Wpointer-arith

#
# Compiler options for debugging and optimization.
#
ifeq ($(DBG_OPT),y)
XOPTS += -g -O1 -DINLINE="" -DSTATIC=""
else
XOPTS += -O3 -DINLINE=inline -DSTATIC=static -fomit-frame-pointer
endif

#
# This will make glibc define only the types defined in the good old SysV
# The only exception is caddr_t (defined in <sys/LiS/user-mdep.h>) and that
# is defined on all SVR4 systems anyway.
#
# When we do not have the types of _every_ standard, LiS will not compile
# if somebody tries to use non-SysV types. This helps to keep the LiS source
# clean, and makes it more portable.
#
ifeq ($(LIS_TARG),user)
XOPTS += -D_SVID_SOURCE -D_BSD_SOURCE
ifdef OSTYPE
ifeq ($(OSTYPE),SunOS5)
XOPTS += -DLIS_HAVE_MAJOR_T -DLIS_HAVE_O_UID_T
endif
endif
endif

#
# Use a define when running in a SMP environment.
#
ifdef CONFIG_SMP
XOPTS += -D__SMP__
endif
ifdef KSMP
ifeq ($(KSMP),y)
XOPTS += -D__SMP__
endif
endif

#
# Always use LiS includes when compiling
#
XOPTS += -I$(LIS_INCL)

#
# If running in the Linux kernel we need the kernel includes too.
# Only the Linux kernel target use this include path.
# The default link /usr/include/linux -> ../src/linux/include/linux
# could be to the wrong kernel!
#
ifeq ($(LIS_TARG),linux)
XOPTS += -I$(KSRC)/include
endif

#
# Base name of user-level library
#
ifeq ($(LIS_TARG),user)
LIBBASE = LiSuser
else
LIBBASE = LiS
endif

#
# Options for linking user-level programs
# Refer directly to the library for static linking.  Using -lLiSuser
# seems to cause "can't find -lc" from C compiler (mystery message).
#
ifeq ($(LIS_SHLIB),y)
ULOPTS = -Wl,-L$(LIBDIR) -l$(LIBBASE)
LIBDEPS =
else
ULOPTS = -static -Wl,-L$(LIBDIR) -l$(LIBBASE)
ULOPTS = $(LIBDIR)/lib$(LIBBASE).a
LIBDEPS = $(LIBDIR)/lib$(LIBBASE).a
endif

