/************************************************************************
*                           cmn_err                                     *
*************************************************************************
*									*
* This is the UNIX common-error message routine.			*
*									*
* For now we will simply turn this into a printk().  Later we can	*
* add some putbuf code.							*
*									*
* Author:	David Grothe	<dave@gcom.com>				*
*									*
* Copyright (C) 1997  David Grothe, Gcom, Inc <dave@gcom.com>		*
*									*
************************************************************************/
/*
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 * 
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330, Cambridge,
 * MA 02139, USA.
 * 
 */

#ident "@(#) LiS cmn_err.c 2.3 6/17/99 14:47:29 "

#include <sys/strport.h>
#include <sys/cmn_err.h>
#include <stdarg.h>

lis_print_trace_t      lis_print_trace;

void	cmn_err(int err_lvl, char *fmt, ...)
{
    static char	 buf[2048];
    extern int	 vsprintf (char *, const char *, va_list);
    va_list	 args;
    char	*p ;

    switch (err_lvl)
    {
    case CE_CONT:		/* continue printing */
	buf[0] = 0 ;
	break ;
    case CE_NOTE:		/* NOTICE */
	strcpy(buf, "\nNOTICE: ") ;
	break ;
    case CE_WARN:		/* WARNING */
	strcpy(buf, "\nWARNING: ") ;
	break ;
    case CE_PANIC:		/* PANIC */
	strcpy(buf, "\nPANIC: ") ;
	break ;
    default:
	sprintf(buf, "\ncmn_err:  Called with invalid arguments (0x%lx, 0x%lx)\n",
		(long) err_lvl, (long) fmt) ;
	if ( lis_print_trace )
	    (*lis_print_trace) (buf);
	return ;
    }

    for (p = buf; *p; p++) ;

    va_start (args, fmt);
    vsprintf (p, fmt, args);
    va_end (args);

    if (err_lvl==CE_PANIC)
	PANIC(buf);
    else
	if ( lis_print_trace )
	    (*lis_print_trace) (buf);

} /* cmn_err */
