/*                               -*- Mode: C -*- 
 * head.c --- LiS stream head processing
 * Author          : Graham Wheeler, Francisco J. Ballesteros
 * Created On      : Tue May 31 22:25:19 1994
 * Last Modified By: David Grothe <dave@gcom.com>
 * RCS Id          : $Id: head.c,v 1.25 1996/01/29 18:12:19 dave Exp $
 * Purpose         : stream head processing stuff
 * ----------------______________________________________________
 *
 *
 *    Copyright (C) 1995  Graham Wheeler, Francisco J. Ballesteros,
 *                        Denis Froschauer
 *    Copyright (C) 1997  David Grothe, Gcom, Inc <dave@gcom.com>
 *
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 * 
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330, Cambridge,
 * MA 02139, USA.
 * 
 *
 *    You can reach us by email to any of
 *    gram@aztec.co.za, nemo@ordago.uc3m.es, 100741.1151@compuserve.com
 */

#ident "@(#) LiS head.c 2.11 6/21/99 21:23:37 "

/* BEWARE: should check:
 * tty stuff
 */


/*  -------------------------------------------------------------------  */
/*				 Dependencies                            */


/* LiS implementation modules used */

#include <sys/stream.h>
#include <sys/poll.h>
#include <sys/LiS/errmsg.h>
#if !defined(ERANGE) && defined(LINUX)
#undef _ERRNO_H
#define __need_Emath 1
#included <errnos.h>
#endif


/*  -------------------------------------------------------------------  */
/*				   Symbols                               */


/*
 * Defines for measuring time intervals
 *
 * Each routine that is measuring its execution time will have a local
 * variable called "time_cell".  The measuring is enabled by a debug
 * bit.
 *
 * time_cell is decalred as an unsigned long initialized to zero.
 */
#define	CLOCKON()	if (LIS_DEBUG_MEAS_TIME)	\
				time_cell -= lis_hitime()

#define	CLOCKADD()	if (LIS_DEBUG_MEAS_TIME) 	\
				time_cell += lis_hitime()

#define	CLOCKOFF(item)	if (LIS_DEBUG_MEAS_TIME) 	\
				LisSetCounter(item, time_cell += lis_hitime())

/*
 * Macro to do a putnext from the stream head.  We want to set a flag
 * during this operation so that we know not to run service queues on
 * top of a "put" operation from the stream head.
 */
#define	PUTNEXT(q,m)				\
	{					\
	    lis_putnext_flag++ ;		\
	    lis_putnext( (q), (m) ) ;		\
	    lis_putnext_flag-- ;		\
	    lis_runqueues();			\
	}
/*
 * Macro to determine whether there are any processes waiting on
 * a poll for a given stream.  When using kernel version 2.1 style
 * polling the queue head is a Linux specific structure, not our
 * portable poll list head.
 */
#if defined(PORTABLE_POLL)
#define	POLL_WAITING(hdp)	( (hdp)->sd_polllist.ph_list != NULL )
#elif defined(LINUX_POLL)
#define	POLL_WAITING(hdp)	( waitqueue_active(&(hdp)->sd_task_list) )
#else
#error "Either PORTABLE_POLL or LINUX_POLL must be defined"
#endif

#define	POLL_NOT_WAITING(hdp)	( ! POLL_WAITING(hdp) )

/*  -------------------------------------------------------------------  */
/*				  Glob. Vars                             */

/* Scheduling  & scan list
 */
struct queue	*lis_qhead; /* first scheduled queue	*/
struct queue	*lis_qtail; /* last scheduled queue		*/
struct queue	*lis_scanqhead;	/* head of STREAMS scan queue	*/
struct queue	*lis_scanqtail;	/* tail of STREAMS scan queue	*/
int		 lis_queues_running ;	/* queues are being run now */

int		 lis_nstrpush;	/* maximum # of pushed modules */
int		 lis_strhold;	/* if not zero str hold feature's activated*/
unsigned long    lis_strthresh;	/* configurable STREAMS memory limit */
unsigned long    lis_iocseq;	/* ioctl id */
stdata_t	*lis_stdata_head ;	/* list of stdata structures */
long		 lis_stdata_cnt ;	/* counts stdata structs in use */
long		 lis_open_cnt ;	/* count opens */
long		 lis_close_cnt ;/* count closes */
struct semaphore lis_sem_template ;	/* blank semaphore */
int		 lis_putnext_flag ;	/* prevents svc procs during putnext */
int		 lis_negstat ;		/* stat went negative */

extern long	 lis_max_mem ;		/* maximum to allocate */
extern long	 lis_max_msg_mem ;	/* maximum to allocate */

/*  -------------------------------------------------------------------  */
/*			   Streams Queue Structures                      */
/*
 * The stream head itself acts as a streams driver (sort-of).  So
 * it needs to have qinit structures just like other drivers.
 *
 * TBD:  We need a better way to set low and high water marks.
 * TBD:  Also packet sizes.
 */
extern void	lis_strrput(queue_t *q, mblk_t *mp) ;
extern void	lis_strwsrv(queue_t *q) ;

struct module_info	strmhd_rdminfo =
		{0			/* mi_idnum */
		,"str_rput"		/* mi_idname */
		,LIS_MINPSZ		/* mi_minpsz */
		,LIS_MAXPSZ		/* mi_maxpsz */
		,0xFFFF			/* mi_hiwat */
		,0xF000			/* mi_lowat */
		} ;

struct module_info	strmhd_wrminfo =
		{0			/* mi_idnum */
		,"str_wput"		/* mi_idname */
		,LIS_MINPSZ		/* mi_minpsz */
		,LIS_MAXPSZ		/* mi_maxpsz */
		,0xFFFF			/* mi_hiwat */
		,0xF000			/* mi_lowat */
		} ;

struct qinit	strmhd_rdinit =
		{lis_strrput		/* qi_putp */
		,NULL			/* qi_srvp */
		,NULL			/* qi_qopen */
		,NULL			/* qi_qclose */
		,NULL			/* qi_qadmin */
		,&strmhd_rdminfo	/* qi_minfo */
		,NULL			/* qi_mstat */
		} ;

struct qinit	strmhd_wrinit =
		{NULL			/* qi_putp */
		,&lis_strwsrv		/* qi_srvp */
		,NULL			/* qi_qopen */
		,NULL			/* qi_qclose */
		,NULL			/* qi_qadmin */
		,&strmhd_wrminfo	/* qi_minfo */
		,NULL			/* qi_mstat */
		} ;

/*  -------------------------------------------------------------------  */
/*			   Local functions & macros                      */

/*  -------------------------------------------------------------------  */

extern void lis_runqueues(void);
extern void lis_qdetach(queue_t *q, int do_close, int flag, cred_t *creds) ;
extern void lis_wakeup_close(caddr_t arg) ;
static int  lis_strdoioctl(struct file *f, stdata_t *hd,
			    strioctl_t *ioc, cred_t *creds,
			    int do_copyin) ;
static int  lis_doclose(struct inode *i,struct file *f,
			struct stdata *head, cred_t *creds) ;
static void free_stdata( struct stdata *hd ) ;
static void lis_deallocate_polllist( stdata_t *hd ) ;

#ifdef LINUX
extern int  lis_assign_inode_to_file(struct file *f, struct inode *i) ;
extern void lis_start_qsched(void) ;
#endif						/* routine in linux-mdep.c */

#if 0
/*  -------------------------------------------------------------------  */
/*                          lis_bzero					 */
void	lis_bzero(void *ptr, int cnt)
{
    char	*p = (char *) ptr ;

    while (cnt-- > 0) *p++ = 0 ;
}
#endif

/*  -------------------------------------------------------------------  */
/*                          lis_sem_init				 */
void	lis_sem_init(struct semaphore *semp, int count)
{
    *semp = lis_sem_template ;	/* blank semaphore */
    SET_SEM_COUNT(semp,count);
}


/*  -------------------------------------------------------------------  */
/*                              str_mux_cycle				 */

#if 0
	This whole mux-cycle business, as I see it, is completely
	artificial.  In order to construct a cycle you would have
	to do something like the following:

	    ioctl(fd2, I_LINK, fd3) ;		2-->3
	    ioctl(fd1, I_LINK, fd2) ;		1-->2-->3
	    ioctl(fd3, I_LINK, fd1) ;		3-->1-->2-->3 (cycle)

	However, you can''t do this because the third ioctl will
	fail.  fd3 has its STPLEX flag set so the ioctl is disallowed.

	--DMG

/* returns non-zero if the given link leads to cyclic graph
 */
static int str_mux_cycle(stdata_t *from, stdata_t *to)
{
    (void)from;
    (void)to;
    return(0);			/* no graph detection */
}
#endif

/*  -------------------------------------------------------------------  */
/*
 * new_muxid
 *
 * Get a new muxid.  Ensure that it is not in use already.
 */
static int
new_muxid(void)
{
    static int		 muxid ;
    stdata_t		*hd ;
    int			 n ;


    for (;;)
    {
	while (++muxid < 0) muxid = 0 ;	/* keep positive */

	if (lis_stdata_head == NULL) return(muxid) ;

	for (hd = lis_stdata_head, n = 0;
	     n == 0 || hd != lis_stdata_head;
	     hd = hd->sd_next, n++)
	{
	    if (hd->sd_mux.mx_index == muxid) break ;	/* in-use */
	}

	if (n >= lis_stdata_cnt)	/* whole list, no match */
	    break ;			/* can use this one */
    }

    return(muxid) ;

} /* new_muxid */

/* 
 * Link the muxed stream under hd.  Return the l_index value.
 * 
 * Use a global running muxid so that if the same driver has multiple
 * control streams it can have unique muxids for its lowers without
 * having to know which control stream goes with which lower.
 */
static int
str_link_mux( stdata_t *hd, stdata_t *muxed, int cmd)
{
    muxed->sd_mux.mx_cmd   = cmd ;	/* save cmd used to link */
#if 0
    muxed->sd_mux.mx_index = ++hd->sd_l_index ;
#else
    muxed->sd_mux.mx_index = new_muxid() ;
#endif
    muxed->sd_mux.mx_next  = hd->sd_mux.mx_hd ;

    hd->sd_mux.mx_hd = muxed ;

    return(muxed->sd_mux.mx_index) ;

}/*str_link_mux*/


/*  -------------------------------------------------------------------  */
/* 
 * This routine unlinks the given multiplexor from the given
 * stream head.  It returns >= 0 for success, < 0 for failure.
 *
 * The parameters 'i', 'f' and 'head' refer to the control stream
 * of the multiplexor.  This stream may be open and we are unlinking
 * a specific lower indexed by l_index, or it may be closing and we
 * are unlinking all lower streams which were i_linked via this
 * control stream.
 *
 * Any mux's below hd whose reference count is currently 1 will
 * be closed.  For any such mux's we recursively unlink any mux's
 * below those.
 *
 * We do not close hd.  We decrement its ref cnt if the command
 * is I_PUNLINK.
 */
static  int
lis_i_unlink(struct inode	*i,
	     struct file	*f,
	     stdata_t		*hd,
	     int		 l_index,
	     int		 cmd,
	     cred_t		*creds)
{
    stdata_t		 *hp ;
    stdata_t		 *prev;
    stdata_t		 *next;
    strioctl_t		  ioc;
    linkblk_t		  lnk;
    int			  err ;
    int			  rtn = 0 ;
    unsigned long  	  time_cell = 0 ;

    CLOCKON() ;
    for (hp = hd->sd_mux.mx_hd, prev = NULL, next = NULL; 
	 hp != NULL; 
	 hp = next)
    {
	next = hp->sd_mux.mx_next ;		/* link to next stream */

	if (l_index >= 0 && hp->sd_mux.mx_index != l_index)
	{
	    prev = hp ;
	    continue ;				/* skip this one */
	}

	if (cmd == I_UNLINK && hp->sd_mux.mx_cmd == I_PLINK)
	{
	    prev = hp ;
	    continue ;				/* skip this one */
	}

	/*
	 * Get permission from driver before unlinking
	 */
	lnk.l_qtop   = hd->sd_wq;
	lnk.l_qbot   = hp->sd_wq;
	lnk.l_index  = hp->sd_mux.mx_index ;

			    /* note I_UNLINK vs I_PLINK eliminated above */
	if (hp->sd_mux.mx_cmd == I_LINK)
	    ioc.ic_cmd = I_UNLINK ;		/* linked w/I_LINK */
	else
	    ioc.ic_cmd = I_PUNLINK;		/* linked w/I_PLINK */

	ioc.ic_timout= LIS_LNTIME;
	ioc.ic_len   = sizeof(linkblk_t);
	ioc.ic_dp    = (char*)&lnk;
	CLOCKADD() ;
	if (   (err = lis_strdoioctl(f,hd,&ioc,creds,0)) < 0
	    && cmd == I_UNLINK			/* ignore errs for I_PUNLINK */
	   )
	{
	    CLOCKON() ;
	    rtn = err ;				/* will return error */
	    if (l_index < 0)			/* OK, skip this one */
	    {					/* lose the memory?? */
		prev = hp ;
		continue ;
	    }

	    CLOCKOFF(IOCTLTIME) ;
	    return(rtn);			/* can't unlink */
	}

	CLOCKON() ;
	if (hd->sd_mux.mx_hd == hp)		/* delinking elt at hd */
	    hd->sd_mux.mx_hd = hp->sd_mux.mx_next ;
	else					/* delinking elt down in list */
	    prev->sd_mux.mx_next = hp->sd_mux.mx_next ;

	hp->sd_mux.mx_next = NULL ;		/* clobber link */

	if ( LIS_DEBUG_IOCTL || LIS_DEBUG_LINK )
	{
	  printk(
	      "strioctl: I_UNLINK: ctl stream %s%s muxid=%d\n",
		    lis_strm_name(hd), lis_maj_min_name(hd),
		    lnk.l_index) ;
	  printk(
	      "                    l_qtop=\"%s\" l_qbot=\"%s%s\"->\"%s\"\n",
		    lis_queue_name(lnk.l_qtop),
		    lis_strm_name(hp), lis_maj_min_name(hp),
		    lis_queue_name(hp->sd_wq->q_next)) ;
	}

						/* restore queues */
	lis_setq(RD(hp->sd_wq), &strmhd_rdinit, &strmhd_wrinit);
	WR(hp->sd_wq)->q_ptr	= hp ;		/* point q back to strm hd */
	RD(hp->sd_wq)->q_ptr	= hp ;		/* assume mux used these ptrs */

	if (   hp->sd_mux.mx_cmd == I_PLINK	/* was linked w/I_PLINK */
	    && hd->sd_refcnt > 0		/* safety first */
	   )
       {
	    hd->sd_refcnt--;			/* decr ref count of head */
	    hd->sd_linkcnt-- ;			/* one less link for head */
	    lis_dec_mod_cnt() ;
	    LisDownCount(MODUSE) ;
       }

	hp->sd_linkcnt-- ;			/* one less I_LINK */
	F_CLR(hp->sd_flag,STPLEX);		/* no longer multiplexed */
	if (hp->sd_refcnt == 1)			/* about to close */
	{
	    /*
	     * Close the lower stream.  The close routine may recurse
	     * on lis_i_unlink() if the lower stream is itself the
	     * control stream of a multiplexor.
	     *
	     * We do not have access to the file and inode of the
	     * lower stream.  Because the sd_refcnt is sitting at 1,
	     * the i_link is the only user of the stream.  The file
	     * and inode have been closed.
	     */
	    CLOCKADD() ;
	    lis_doclose(NULL, NULL, hp, creds) ;	/* close lower mux */
	    CLOCKON() ;
	}
	else
	if (hp->sd_refcnt > 1)			/* be safe */
	{
	    hp->sd_refcnt-- ;			/* decr ref cnt */
	    lis_dec_mod_cnt() ;
	    LisDownCount(MODUSE) ;
	}

	if (l_index >= 0) break ;		/* done */
    }

    CLOCKOFF(IOCTLTIME) ;
    return(rtn) ;

} /* lis_i_unlink */

/*  -------------------------------------------------------------------  */
/* lis_i_link
 *
 * Called from strioctl to perform the I_LINK (I_PLINK) function.
 *
 * Note about I_PLINK:  This causes the refcnt of the top stream
 * to be incremented so that a close of that stream will not cause
 * the stream to be deallocated and the multiplexor dismantled.
 * However, if the stream does get closed then the stdata structure
 * will be laying around in memory with refcnt==1 and nothing holding
 * onto it.  It will be their until the next reboot.  Hopefully,
 * this multiplexor can take care of itself until then.
 */
int	lis_i_link(struct inode	*i,
		   struct file	*f,
		   stdata_t	*hd,
		   int		 muxfd,
		   int		 cmd,
		   cred_t	*creds)
{
    stdata_t		*muxed;
    linkblk_t		 lnk;
    strioctl_t		 ioc;
    int			 err ;
    unsigned long  	 time_cell = 0 ;

#define	RTN(v)		{lis_unlock_wopen(hd); CLOCKOFF(IOCTLTIME) ; return(v); }

    CLOCKON() ;
    err = lis_lock_wopen(hd);
    if (err < 0) RTN(err) ;

    if ( (muxed = lis_fd2str(muxfd)) == NULL )
	RTN(-EINVAL);

    lnk.l_qtop = hd->sd_wq;
    lnk.l_qbot = muxed->sd_wq;
    lnk.l_index= str_link_mux(hd, muxed, cmd);

    if ( LIS_DEBUG_IOCTL || LIS_DEBUG_LINK )
    {
	printk("strioctl: I_LINK: ctl stream %s%s l_index=%d\n",
		lis_strm_name(hd), lis_maj_min_name(hd),
		lnk.l_index) ;
	printk("                  l_qtop=\"%s\" l_qbot=\"%s%s\"->\"%s\"\n",
		lis_queue_name(lnk.l_qtop),
		lis_strm_name(muxed), lis_maj_min_name(muxed),
		lis_queue_name(muxed->sd_wq->q_next)
		) ;
    }

    lis_setq(LIS_RD(muxed->sd_wq),
	     hd->sd_strtab->st_muxrinit, hd->sd_strtab->st_muxwinit);

    ioc.ic_cmd	 = I_LINK;
    ioc.ic_timout= LIS_LNTIME;
    ioc.ic_len	 = sizeof(linkblk_t);
    ioc.ic_dp	 = (char*)&lnk;
    CLOCKADD() ;
    if ((err = lis_strdoioctl(f,hd,&ioc,creds,0)) < 0)
    {
	CLOCKON() ;
	lis_setq(LIS_RD(muxed->sd_wq),		/* put back old queues */
	     hd->sd_strtab->st_rdinit, hd->sd_strtab->st_wrinit);

	CLOCKADD() ;
	lis_i_unlink(i, f, hd, lnk.l_index,
		     cmd == I_LINK ? I_UNLINK : I_PUNLINK, creds);
	CLOCKON() ;
	RTN(err);
    }

    CLOCKON() ;
    F_SET(muxed->sd_flag,STPLEX);
    muxed->sd_refcnt++;			/* so a close won't deallocate */
    muxed->sd_linkcnt++ ;		/* one more I_LINK */
    lis_inc_mod_cnt() ;
    LisUpCount(MODUSE) ;

    /*
     * Set initial queue flags.
     *
     * We set the QWANTR flag of both sides of the queue so that
     * a putq into either queue will schedule the service procedure
     * to run.  The flag means that some service procedure "wants"
     * to "read" from the queue, that is, take messages out of it.
     */
    F_SET(muxed->sd_wq->q_flag,QWANTR);	/* yes, even tho it's the wrq */
    F_SET(LIS_RD(muxed->sd_wq)->q_flag,QWANTR);

    if (muxed->sd_mux.mx_cmd==I_PLINK)	/* linked w/I_PLINK */
    {
	hd->sd_refcnt++;		/* so a close won't deallocate */
	hd->sd_linkcnt++ ;		/* count that as a link */
	lis_inc_mod_cnt() ;
	LisUpCount(MODUSE) ;
    }

    lis_unlock_wopen(hd); 		/*wake procs in head*/
    CLOCKOFF(IOCTLTIME) ;
    return(lnk.l_index);

#undef RTN

} /* lis_i_link */


/*  -------------------------------------------------------------------  */
/* detach queues from stream and free the stream
 */
static void lis_dismantle(stdata_t *hd, cred_t *creds)
{
    queue_t	*q ;

    if (hd->sd_wq != NULL)			/* still have strm head queue */
    {
	q = LIS_RD(hd->sd_wq) ;
	if (!LIS_CHECK_Q_MAGIC(q)) return ;
	LisDownCounter(MSGQDSTRHD, lis_qsize(q)) ;
	lis_qdetach(q, 0, 0, creds);
	hd->sd_wq = NULL ;
    }

    free_stdata( hd ) ;				/* free the stdata structure */
}


/*  -------------------------------------------------------------------  */

/* set read options
 */
static int
set_readopt(int *flags, int rmode)
{
    long oflags=*flags;

    switch(rmode & RMODEMASK){
    case RNORM:
    case RMSGD:
    case RMSGN:
	*flags &= ~RMODEMASK ;			/* clear mode bits */
	*flags |= (rmode & RMODEMASK) ;		/* set mode */
	break;
    default:
	*flags=oflags;				/* restore original */
	return(-EINVAL);
    }

    switch(rmode & RPROTMASK){
    case RPROTNORM:
    case RPROTDAT:
    case RPROTDIS:
	*flags &= ~RPROTMASK ;			/* clear proto bits */
	*flags |= (rmode & RPROTMASK) ;		/* set read proto mask */
	break;
    case 0:					/* not being set */
	break;
    default:
	*flags=oflags;				/* restore original */
	return(-EINVAL);
    }

    return(0);
}

/* These will let you know if you should block on rd/wr as dictated by the
 * stream head and file flags. They should be called when the process should
 * block due to no free space or no msg avail.
 * Return non-zero if you shouldn't block and zero if you should do block 
 */
#define should_notblock_rd(hd,fp) (F_ISSET(fp->f_flags,O_NONBLOCK))

#define should_notblock_wr(hd,fp)					\
    (F_ISSET(fp->f_flags,O_NONBLOCK) && !F_ISSET(hd->sd_flag,OLDNDELAY))

/*  -------------------------------------------------------------------  */

/* Will add q to the scan list.  This happens as a result of the
 * scanq timer expiring.
 *
 * One minor problem:  If the kernel is going to run the streams
 * queues before returning from the timer interrupt then we are OK.
 * If, however, it is up to us to run ourselves then we have two
 * choices:
 *	1) wait until some user process enters the kernel on streams business
 *	2) run the queues here on top of the clock interrupt
 * I don't like either choice.  It would be best of the kernel ran
 * the streams queues from strategic places such as just before returning
 * from interrupt processing and just before returning to user level.
 *
 * NOTE: STREAMS queues are run by waking up a process so that everything
 *       is orderly. -- DMG
 */
static void
add_to_scanl( caddr_t arg )
{
  int		 psw;
  stdata_t	*hd = (stdata_t *) arg ;
  queue_t	*q ;

  /* We'll use a naive FIFO, this should be revisited
   */
  q = hd->sd_wq ;			/* stream head write queue */
  hd->sd_scantimer = 0 ;		/* note that timer is not running */
  SPLSTR(psw);
  if (F_ISSET(q->q_flag,QSCAN)) {
    SPLX(psw);
    return;
  }
  if (lis_scanqtail) 
    lis_scanqtail->q_scnxt = q;
  else
    lis_scanqhead = q;
  q->q_scnxt=NULL;
  lis_scanqtail=q;
  F_SET(q->q_flag,QSCAN);
  SPLX(psw);
  return ;

}/*add_to_scanl*/

static void
start_scanl_timer(stdata_t *hd)
{
    if (hd->sd_scantimer == 0)
	hd->sd_scantimer = lis_timeout_fcn(add_to_scanl,
					   (caddr_t) hd,
					   SECS_TO(LIS_RTIME),
					   "scan-list", MEM_TIMER
					   );  

} /* start_scanl_timer */


/*  -------------------------------------------------------------------  */
/* Returns a pointer to the stream head in the i-node after allocating it.
 * The head is locked upon return.
 * The head structure can be reused on a second open call as soon
 * as the inode is pointed to it (well, sort of).  This is really
 * more multi-processor thinking at work here, but it is safe nevertheless.
 */
static stdata_t *
alloc_stdata( struct inode *i)
{
    struct stdata *head ;
    int		   err ;

    head = (stdata_t*) ALLOCF(sizeof(stdata_t),"stream-head ");
    if (head != NULL)			/* successful allocation */
    {					/* initialize the structure */
	lis_bzero(head, sizeof(*head)) ;	/* clear to zero */
	head->magic=STDATA_MAGIC;
	lis_sem_init(&head->sd_wopen, 1) ; /* initialize open semaphore */
	err = lis_lock_wopen(head);	/* P() it */
	if (err < 0)			/* VERY busted */
	{
	    printk("aloc_stdata: error locking newly allocated queue\n") ;
	    head->magic ^= 1 ;		/* clobbers the magic number */
	    FREE(head) ;		/* free queue */
	    return(NULL) ;		/* failed allocation */
	}

	lis_sem_init(&head->sd_wioc, 1) ;
	lis_sem_init(&head->sd_wread, 0) ;	/* explicit init for emphasis */
	lis_sem_init(&head->sd_wwrite, 0) ;	/* explicit init for emphasis */
	lis_sem_init(&head->sd_wiocing, 0) ;	/* explicit init for emphasis */
	lis_sem_init(&head->sd_closing, 0) ;	/* explicit init for emphasis */

	head->sd_closetime	= LIS_CLTIME;

	/*
	 * Link sdtata structure into master list
	 */
	lis_stdata_cnt++ ;			/* count stdata structs */
	if (lis_stdata_head == NULL)
	{
	    head->sd_next   = head ;
	    head->sd_prev   = head ;
	    lis_stdata_head = head ;
	}
	else
	{
	    head->sd_next = lis_stdata_head ;
	    head->sd_prev = lis_stdata_head->sd_prev ;
	    lis_stdata_head->sd_prev->sd_next = head ;
	    lis_stdata_head->sd_prev = head ;
	}
    }

    return(head) ;

}/*alloc_stdata*/

/*  -------------------------------------------------------------------  */
/* Releases the stream head in the i-node
 */
static void
free_stdata( struct stdata *hd )
{

    /* must assert here that every pointer in the stdata struct which
     * may hold dynamic memory is NULL.
     */
    if (hd == NULL || hd->magic != STDATA_MAGIC) return ;

    SEM_DESTROY(&hd->sd_wopen) ;	/* de-initialize the semaphores */
    SEM_DESTROY(&hd->sd_wioc) ;
    SEM_DESTROY(&hd->sd_wread) ;
    SEM_DESTROY(&hd->sd_wwrite) ;
    SEM_DESTROY(&hd->sd_wiocing) ;
    SEM_DESTROY(&hd->sd_closing) ;

    if (hd->sd_wmsg != NULL)
    {
	freemsg(hd->sd_wmsg) ;
	hd->sd_wmsg = NULL ;
    }

    if (hd->sd_iocblk != NULL)
    {
	freemsg(hd->sd_iocblk) ;
	hd->sd_iocblk = NULL ;
    }

    if (hd->sd_siglist != NULL)
	lis_free_elist(&hd->sd_siglist);

    lis_deallocate_polllist(hd) ;

    if (lis_stdata_head != NULL)
    {
	if (lis_stdata_head == hd)		/* deleting head element */
	    lis_stdata_head = hd->sd_next ;	/* move down one */

	if (lis_stdata_head == hd)		/* only element */
	    lis_stdata_head = NULL ;		/* no more streams */
	else					/* link around "hd" */
	{
	    hd->sd_prev->sd_next = hd->sd_next ;
	    hd->sd_next->sd_prev = hd->sd_prev ;
	}
    }

    hd->magic ^= 1 ;			/* clobbers the magic number */
    FREE(hd) ;				/* give it back */
    lis_stdata_cnt-- ;			/* count stdata structs */

}/*free_stdata*/


/*  -------------------------------------------------------------------  */

/* Will check that inode & device are a proper STREAMS device
 *
 * Note:  This whole thing used to be a macro but the linker did
 *        something strange to the code that caused it not to
 *        work properly.  Making it into a procedure is an attempt
 *        to do an end-run around this problem. - DMG
 */
#define CHECK_INO(i,n)	{						\
			    int		rslt ;				\
									\
			    if ((rslt = lis_check_ino(i,n)) < 0)	\
				return(rslt) ;				\
			}

int	lis_global_major ;
int lis_check_ino(struct inode *i, char *msg)
{
    int		maj ;

    if (!i)
    { 					
	lis_error(LIS_ERROR,msg,"called w/ null inode"); 	
	return(-EBADF); 			
    }						

    if (!S_ISCHR(i->i_mode)) 			
	return(-ENODEV);			

    maj = STR_MAJOR(i->i_rdev) ;
    lis_global_major = maj ;
    if (!LIS_DEVOK(maj))	
	return(-ENOTTY) ;

    return(0) ;

} /* lis_check_ino */

/*  -------------------------------------------------------------------  */

/* copyin the buffer shown in usr strbuf struct sb, into *m.
 * As it's now it just copies the whole usr bufs to single blks.
 */
static  int
copyin_msgpart( struct file *f, stdata_t *hd, mblk_t **m, char *sb, int wroff )
{
    strbuf_t kbuf;
    int err;

    *m = NULL ;					/* NULL msg ptr by default */
    if ((err=lis_check_umem(f,VERIFY_READ,sb,sizeof(strbuf_t)))<0)
	return(err);

    						/* fetch strbuf */
    if ((err = lis_copyin(f,&kbuf,sb,sizeof(strbuf_t))) < 0)
	return(err) ;

    if (kbuf.len < 0 || kbuf.buf == NULL)	/* don't include this msg part */
	return(0) ;				/* not an error */

    if (   kbuf.len < (int)hd->sd_wq->q_next->q_minpsz
	|| (   hd->sd_wq->q_next->q_maxpsz != INFPSZ
	    && kbuf.len > (int)hd->sd_wq->q_next->q_maxpsz 
	   )
       )
	return(-ERANGE);

    if ( (*m=(mblk_t*)allocb(kbuf.len+wroff,BPRI_LO))==NULL)
	return(-ENOMEM);

    (*m)->b_rptr = ((*m)->b_wptr += wroff);
    if (kbuf.len>0) {
	if ((err=lis_check_umem(f,VERIFY_READ,kbuf.buf,kbuf.len))<0)
	    return(err);
	if ((err = lis_copyin(f,(*m)->b_wptr,kbuf.buf,kbuf.len)) < 0)
	    return(err);
	(*m)->b_wptr+=kbuf.len;
    }
    return(0);
}/*copyin_msgpart*/

/*  -------------------------------------------------------------------  */
/* copyin ctl & dat intro M_PROTO+M_DATA msg
 */
static int
copyin_msg(struct file *f, stdata_t *hd,
		mblk_t **m, strbuf_t *uctl, strbuf_t *udat)
{
    int err;
    mblk_t *aux = NULL;

    if (m == NULL) return(-EINVAL) ;

    if ( uctl != NULL )
    {
	if ((err=copyin_msgpart(f,hd,m,(char*)uctl,0))<0)
	    return(err);

	if (*m != NULL)
	    lis_btype(*m) = M_PROTO;
    }
    else
	*m = NULL ;				/* no ctl msg part */

    if ( udat != NULL )
    {
	if ((err=copyin_msgpart(f,hd,&aux,(char*)udat,hd->sd_wroff))<0)
	{
	    if (*m != NULL)
		lis_freeb(*m);

	    return(err);
	}
    }

    if (*m == NULL && aux == NULL)		/* no message to send */
	return(0) ;

    if (*m != NULL && aux != NULL)		/* have 2 msgs to send */
	lis_linkb(*m,aux);			/* link them */
    else
    if (*m == NULL)				/* have no ctl msg */
	*m = aux ;				/* rtn just data msg */

    return(0);
	
}/*copyin_msg*/

/*  -------------------------------------------------------------------  */

/* copyout part of msg.
 * It will assume it's ctl and return MORECTL/MOREDATA if more data is left.
 * msg won't be touched if !doit 
 * this function doesn't free any block, just adjust b_[rw]ptr's
 * is_ctl is true for the ctl part / false for the data part.
 *
 * 'm' can be NULL which means do not copy anything, but update
 * the user's strbuf.
 */
static int
copyout_msgpart( struct file *f, mblk_t *m,
		    strbuf_t *sb, strbuf_t *usb, int is_ctl, int doit)
{
    int err;
    int rtn = 0 ;
    int chunk ;
    int mlen = 0 ;		/* # of bytes for ctl or dta part */
    int size_set = 0 ;
    int maxlen = sb->maxlen;	/* how much the user wants */
    int gotsome= 0 ;
    int mtype ;

    if (doit)			/* usb points to user space */
    {				/* ensure writeable */
	if (usb && (err=lis_check_umem(f,VERIFY_WRITE,usb,sizeof(strbuf_t)))<0)
	  return(err);
	if (maxlen >= 0 && (err=lis_check_umem(f,VERIFY_WRITE,sb->buf,maxlen))<0)
	  return(err);
    }
    else			/* usb is in kernel space */
    {				/* but sb->buf is in user space */
	if (maxlen >= 0 && (err=lis_check_umem(f,VERIFY_WRITE,sb->buf,maxlen))<0)
	  return(err);
    }

    for (sb->len=0;
	 m != NULL;
	 m=m->b_cont)
    {
	chunk = 0 ;
	mtype = lis_btype(m) ;
	if (size_set == 0)				/* size not yet set */
	{
	    if (   (!is_ctl && mtype == M_DATA)		/* hit data msg */
		|| ( is_ctl && mtype != M_DATA))	/* hit ctl msg */
	    {
		mlen = lis_xmsgsize(m);			/* set length */
		size_set = 1 ;
	    }

	    if (size_set == 0) continue ;		/* still looking */
	}

	if (is_ctl && mtype == M_DATA)			/* hit data msg */
	    break ;					/* quit */

	if (!is_ctl && mtype != M_DATA)			/* non-data msg */
	    continue ;					/* keep looking */

	gotsome = 1 ;					/* blk of correct type*/
	if (maxlen < 0)					/* no place for it */
	    break ;					/* stop looking */

	chunk=min(maxlen,lis_mblksize(m));
	if (chunk > 0 && (err=lis_copyout(f,m->b_rptr,sb->buf+sb->len,chunk))<0)
	    return(err) ;

	if (doit)
	  lis_adjmsg(m,chunk);

	 sb->len += chunk ;
	 maxlen  -= chunk ;
	 if (maxlen <= 0)
	    break ;
    }					/*for blk*/

    err = 0 ;				/* return value */
    if (gotsome)			/* got some bytes from msg */
    {
	if (sb->len < mlen)		/* less than what was in msg */
	    rtn = (is_ctl?MORECTL:MOREDATA);
    }
    else
	sb->len = -1 ;			/* no blocks of that type */

    if (doit)
    {
	if (usb != NULL)
	    err = lis_copyout(f,sb,usb,sizeof(strbuf_t));
    }
    else
	*usb = *sb ;			/* update kernels rtn structure */

    return(err < 0 ? err : rtn);

}/*copyout_msgpart*/

/*  -------------------------------------------------------------------  */
/* copyout msg to usr buffers given by ctl & dta.
 * m should be a proto+data msg.
 * this function doesn't free any block, just adjust b_[rw]ptr's
 * msg won't be touched if !doit 
 */
static int
copyout_msg( struct file *f, mblk_t *m, strbuf_t *kctl, strbuf_t *kdta,
	     strbuf_t *uctl, strbuf_t *udta, int doit)
{
    int errctl,errdta;
    if ((errctl=copyout_msgpart(f,m,kctl,uctl,1,doit))<0)
	return(errctl);
    else if ((errdta=copyout_msgpart(f,m,kdta,udta,0,doit))<0)
	return(errdta);
    return(errctl|errdta);
}/*copyout_msg*/

/*  -------------------------------------------------------------------  */

/* Do we have streams resources?
 */
   
static int
stravail(struct file *fp)
{
    (void) fp ;
    return(1) ;
}

/*  -------------------------------------------------------------------  */
/*Returns non-zero if we should hold msg on write at the stream head.
 */
#define strholding(sd)	(lis_strhold && F_ISSET((sd)->sd_flag,STRHOLD))

/*  -------------------------------------------------------------------  */


/*  -------------------------------------------------------------------  */
/* qdetach - unlink queue from stream, close driver/module, and free queue
 */
void 
lis_qdetach(queue_t *q, int do_close, int flag, cred_t *creds)
{
    int		psw ;
    int		i ;
    const char	*name = NULL;
    queue_t	*downstrm_wrq ;
    queue_t	*downstrm_rdq ;
    queue_t	*upstrm_rdq ;
    queue_t	*upstrm_wrq ;
    unsigned long  time_cell = 0 ;

    CLOCKON() ;
    if (q == NULL)
    {
	CLOCKOFF(CLOSETIME) ;
	return;
    }

    q = LIS_RD(q) ;			/* operate on the read queue */
    if (q == NULL)
    {
	CLOCKOFF(CLOSETIME) ;
	return;
    }

    lis_runqueues();

    for (i = 1; i <= 2; i++, q = LIS_WR(q))		/* do twice */
    {
	if (!LIS_CHECK_Q_MAGIC(q)) return ;	/* lose the memory */

	if (LIS_DEBUG_CLOSE) name = lis_queue_name(q) ;

	if (do_close && q->q_qinfo != NULL && q->q_qinfo->qi_qclose != NULL)
	{
	    if (name)
		printk("lis_qdetach: queue \"%s\": calling close routine\n", name);

	    CLOCKADD() ;		/* user routine might sleep */
	    (*q->q_qinfo->qi_qclose)(q, (q->q_next ? 0 : flag), creds);
	    CLOCKON() ;
	}
    }

    /*
     * 'q' is the write queue
     */
    SPLSTR(psw) ;

    if (F_ISSET(q->q_flag,QSCAN)) {
      /* remove from the scan list
       */
	queue_t *p_scan = lis_scanqhead;
	queue_t *p_prev = NULL;

	if (name)
	    printk("lis_qdetach: queue \"%s\": removing from scan list\n", name);

	while (p_scan != NULL)
	{					/* only wr q in scan list */
	    if (!LIS_CHECK_Q_MAGIC(p_scan)) break ;

	    if (p_scan == q)			/* q is write queue */
	    {					/* p_scan points to our queue */
	      if (p_prev != NULL)		/* trailing ptr valid?  */
		  p_prev->q_scnxt = p_scan->q_scnxt ;	/* link around q */
	      else				/* prev still is q_head */
		  lis_scanqhead = p_scan->q_scnxt ;	/* link hd around q */
						/* NULL if last elt in list */
	      if (lis_scanqtail == p_scan)	/* last elt in list? */
		  lis_scanqtail = p_prev ;	/* back up tail ptr */
						/* NULL if only elt in list */
	      /* do not move prev ptr */
	    }
	    else				/* not our element */
		p_prev = p_scan;		/* keep trailing ptr */

	    p_scan = p_scan->q_scnxt ;		/* go to next elt of list */
	}
    }

    SPLX(psw) ;				/* open brief interrupt window */
    SPLSTR(psw) ;

    if ((q->q_flag & QENAB) || (LIS_RD(q)->q_flag & QENAB)){
	queue_t *p_scan = lis_qhead;
	queue_t *p_prev = NULL;

	if (name)
	    printk("lis_qdetach: queue \"%s\": unscheduling service procedures\n", name);

	/* Unschedule the service procedures.
	 */
	while (p_scan != NULL)
	{
	    if (!LIS_CHECK_Q_MAGIC(p_scan)) break ;

	    if (p_scan == q || p_scan == LIS_RD (q)) /* q is write queue */
	    {					/* p_scan points to our queue */
	      if (p_prev != NULL)		/* trailing ptr valid?  */
		  p_prev->q_link = p_scan->q_link ;	/* link around q */
	      else				/* prev still is q_head */
		  lis_qhead = p_scan->q_link ;	/* link hd around q */
						/* NULL if last elt in list */
	      if (lis_qtail == p_scan)		/* last elt in list? */
		  lis_qtail = p_prev ;		/* back up tail ptr */
						/* NULL if only elt in list */
	      /* do not move prev ptr */
	      LisDownCount(QSCHEDS) ;
	    }
	    else				/* not our element */
		p_prev = p_scan;		/* keep trailing ptr */

	    p_scan = p_scan->q_link ;		/* go to next elt of list */
	}
    }

    SPLX(psw) ;				/* open brief interrupt window */
    SPLSTR(psw) ;

    lis_flushq(q, FLUSHALL);
    lis_flushq(LIS_RD(q), FLUSHALL);

    downstrm_wrq	= LIS_WR(q)->q_next ;
    if (downstrm_wrq != NULL)
	downstrm_rdq= LIS_RD(downstrm_wrq) ;
    else
	downstrm_rdq= NULL ;

    upstrm_rdq	= LIS_RD(q)->q_next ;
    if (upstrm_rdq != NULL)
	upstrm_wrq	= LIS_WR(upstrm_rdq) ;
    else
	upstrm_wrq	= NULL ;

    if (upstrm_wrq != NULL)			/* have write-queue above us */
    {
	if (name)
	    printk("lis_qdetach: queue \"%s\": link around write queue\n", name);

	upstrm_wrq->q_next = downstrm_wrq ;	/* link around our queue */
    }

    if (downstrm_rdq != NULL)		/* have read-queue below us */
    {
	if (name)
	    printk("lis_qdetach: queue \"%s\": link around read queue\n", name);

	downstrm_rdq->q_next = upstrm_rdq ;	/* link around our queue */
    }

    if (name)
	printk("lis_qdetach: queue \"%s\": de-allocate queue\n", name);

    lis_freeq(LIS_RD(q));

    SPLX(psw) ;
    CLOCKOFF(CLOSETIME) ;

}/*lis_qdetach*/


/*  -------------------------------------------------------------------  */
/* lis_qattach - set up a queue pair for driver/module
 */
static int
lis_qattach (stdata_t *hd, struct streamtab *qinfo, queue_t *q, dev_t *devp,
	     int flag, cred_t *creds)
{
    queue_t *newq;
    int  open_mode,err;
    int		psw ;
    const char	*name ;
    dev_t	 dev = 0 ;

    (void) hd ;
    if (q == NULL)
	return -EIO;

    if (   qinfo == NULL
	|| qinfo->st_rdinit == NULL
	|| qinfo->st_rdinit->qi_minfo == NULL
	|| (name = qinfo->st_rdinit->qi_minfo->mi_idname) == NULL
       )
	name = "unnamed-queue" ;


    if ((newq = lis_allocq(name)) == NULL)
    {
	if (LIS_DEBUG_OPEN)
	    printk("lis_qattach: failed to allocate queue for driver\n") ;
	return -ENOMEM;
    }

    SPLSTR(psw) ;

    /* insert the queue 
     */
    newq->q_next = q;		/* LIS_RD(q), actually */
    q = LIS_WR(q);
    if ((LIS_WR(newq)->q_next = q->q_next) == NULL)
    	open_mode = DRVOPEN;
    else {
    	LIS_OTHERQ(q->q_next)->q_next = newq;
    	open_mode = MODOPEN;
    }
    q->q_next = LIS_WR(newq);

    /* Set up the queue interface stuff */
    lis_setq(newq, qinfo->st_rdinit, qinfo->st_wrinit);
    /*
     * Set initial queue flags.
     *
     * We set the QWANTR flag of both sides of the queue so that
     * a putq into either queue will schedule the service procedure
     * to run.  The flag means that some service procedure "wants"
     * to "read" from the queue, that is, take messages out of it.
     */
    F_SET(newq->q_flag,QWANTR);
    F_SET(LIS_WR(newq)->q_flag,QWANTR);		/* yes, even tho it's the wrq */

    SPLX(psw) ;
    /* Now call the new module/driver's open routine.
     */

    if (newq->q_qinfo->qi_qopen == NULL)	/* ensure open routine */
    {
	if ( LIS_DEBUG_OPEN )
	    printk("lis_qattach: driver \"%s\" does not have open routine\n",
		   lis_queue_name(newq)) ;

	return(-ENOPKG) ;			/* pkg not installed */
    }

    if (devp == NULL)				/* as for module open */
	devp = &dev ;

    if ((err = (*newq->q_qinfo->qi_qopen) (newq, devp, flag,
					    open_mode, creds)) != 0)
    {
	if ( LIS_DEBUG_OPEN )
	    printk("lis_qattach: driver \"%s\" returned error %d\n",
		   lis_queue_name(newq), err) ;

	lis_qdetach(newq, 0, 0, creds);

	if (err < 0) return(err) ;
	else	     return(-err) ;
    }
    else 
	err=0;

    return(err);

}/*lis_qattach*/

/*  -------------------------------------------------------------------  */
/* Push this module in this stream and call its open proc.
 */
static int
push_mod(stdata_t *head, streamtab_t *mod, int flags, cred_t *creds)
{
    int err;

    if (head->sd_maxpushcnt && head->sd_pushcnt == head->sd_maxpushcnt)
	return(-ENOSR);
    if ( (err=lis_qattach(head,mod,LIS_RD(head->sd_wq), NULL, flags, creds )) < 0 )
	return(err);
    else
	head->sd_pushcnt++;
    return(0);
}/*push_mod*/

/*  -------------------------------------------------------------------  */
/* Pop the top module off this stream and call its close proc.
 */
static int
pop_mod(stdata_t *head, int flags, cred_t *creds)
{
    if (!head->sd_pushcnt)
	return(-EINVAL);
    head->sd_pushcnt--;
    lis_qdetach(LIS_RD(head->sd_wq->q_next),1,flags, creds);
    return(0);
}/*pop_mod*/

/*  -------------------------------------------------------------------  */

/* Set STREAMS options
 */
static void
set_options( stdata_t *hd, stroptions_t *so)
{
    /* just set head flags as requested
     */
    if (F_ISSET(so->so_flags,SO_MREADON))	F_SET(hd->sd_flag,SNDMREAD);
    if (F_ISSET(so->so_flags,SO_MREADOFF))	F_CLR(hd->sd_flag,SNDMREAD);
    if (F_ISSET(so->so_flags,SO_NDELON))	F_SET(hd->sd_flag,OLDNDELAY);
    if (F_ISSET(so->so_flags,SO_NDELOFF))	F_CLR(hd->sd_flag,OLDNDELAY);
    if (F_ISSET(so->so_flags,SO_ISTTY))		F_SET(hd->sd_flag,STRISTTY);
    if (F_ISSET(so->so_flags,SO_ISNTTY))	F_CLR(hd->sd_flag,STRISTTY);
    if (F_ISSET(so->so_flags,SO_TOSTOP))	F_SET(hd->sd_flag,STRTOSTOP);
    if (F_ISSET(so->so_flags,SO_TONSTOP))	F_CLR(hd->sd_flag,STRTOSTOP);
    if (F_ISSET(so->so_flags,SO_STRHOLD))	F_SET(hd->sd_flag,STRHOLD);
    if (F_ISSET(so->so_flags,SO_DELIM))		F_SET(hd->sd_flag,STRDELIM);
    if (F_ISSET(so->so_flags,SO_NODELIM))	F_CLR(hd->sd_flag,STRDELIM);


    if (F_ISSET(so->so_flags,SO_READOPT))
	set_readopt(&hd->sd_rdopt,so->so_flags);

    /* This chunk will set parameters
     */
    if (F_ISSET(so->so_flags,SO_WROFF))		
        hd->sd_wroff=so->so_wroff;
    if (F_ISSET(so->so_flags,SO_MINPSZ))	
        RD(hd->sd_wq)->q_minpsz= so->so_minpsz;
    if (F_ISSET(so->so_flags,SO_MAXPSZ))
	RD(hd->sd_wq)->q_maxpsz= so->so_maxpsz;
    if (F_ISSET(so->so_flags,SO_BAND))
    {
	if (F_ISSET(so->so_flags,SO_HIWAT))
	    lis_strqset(RD(hd->sd_wq), (qfields_t) QHIWAT,
						    so->so_band, so->so_hiwat);
	if (F_ISSET(so->so_flags,SO_LOWAT))
	    lis_strqset(RD(hd->sd_wq), (qfields_t) QLOWAT,
						    so->so_band, so->so_lowat);
    }
    else					/* set in q struct */
    {
	if (F_ISSET(so->so_flags,SO_HIWAT))
	    RD(hd->sd_wq)->q_hiwat= so->so_hiwat;
	if (F_ISSET(so->so_flags,SO_LOWAT))
	    RD(hd->sd_wq)->q_lowat= so->so_lowat;
    }
}/*set_options*/

/*  -------------------------------------------------------------------  */

/* Wake up polling processes
 */
static void
lis_wake_up_poll(stdata_t *hd, int ev)
{
#if defined(PORTABLE_POLL)
  polldat_t *pd;
  int	     psw ;

  if ( LIS_DEBUG_POLL )
    printk("lis_wake_up_poll: stream %s%s: revents:%s\n",
	lis_strm_name(hd), lis_maj_min_name(hd), lis_poll_events((short)ev)) ;

  SPLSTR(psw) ;					/* may be unnecessary */

  for (pd=hd->sd_polllist.ph_list; pd != NULL; pd=pd->pd_next) 
  {
    pd->pd_events = (short) ev ;		/* events being caused */
    if (pd->pd_fn != NULL)			/* careful */
	(*pd->pd_fn)(pd->pd_arg);		/* this will wake up the
						 * polling process.
	  					 */
  }

  SPLX(psw) ;

  lis_select_wakeup(hd) ;			/* wake up selectors */

#elif defined(LINUX_POLL)

  if ( LIS_DEBUG_POLL )
    printk("lis_wake_up_poll: stream %s%s: revents:%s\n",
	lis_strm_name(hd), lis_maj_min_name(hd), lis_poll_events((short)ev)) ;

  wake_up_interruptible(&hd->sd_task_list);

#else
#error "Either PORTABLE_POLL or LINUX_POLL must be defined"
#endif

}/*lis_wake_up_poll*/

/*  -------------------------------------------------------------------  */

/* deallocate poll list
 * 
 * Called from stdata structure deallocation code.  Shortly after
 * this routine returns the stdata structure will be deallocated.
 * We wake up any processes that are sleeping on poll data structures
 * chained off of this stdata structure.  Immediately upon return
 * from the wakeup, we go through the list and nullify the pointer
 * to the list head in all the pdat structures.  This means that
 * the poll code, when it finally wakes up, may find this head
 * pointer to be NULL.  If the poll code has not delinked the
 * poll entry by then, it may be in trouble.
 */
static void
lis_deallocate_polllist( stdata_t *hd)
{
#if defined(PORTABLE_POLL)
  polldat_t *pd;
  int	     psw ;

  SPLSTR(psw) ;					/* may be unnecessary */

  if (hd->sd_polllist.ph_list != NULL)
      lis_wake_up_poll(hd, POLLHUP) ;
  else						/* no pollers */
      lis_select_wakeup(hd) ;			/* wake up selectors */

  for (pd=hd->sd_polllist.ph_list; pd != NULL; pd = pd->pd_next) 
  {
      pd->pd_headp = NULL ;
  }

  SPLX(psw) ;

#elif defined(LINUX_POLL)

  hd->sd_task_list = (struct wait_queue *) NULL ;  /* nothing to deallocate */

#else
#error "Either PORTABLE_POLL or LINUX_POLL must be defined"
#endif
}/*lis_deallocate_polllist*/

/*  -------------------------------------------------------------------  */
/* kill procs in elist w/ sig
 */
static void 
kill_procs( struct strevent *elist, int sig, short e)
{
    strevent_t *ev;
    int res;

    (void) sig ;
    for(ev = elist->se_next ; ev != elist; ev=ev->se_next )
	if ((ev->se_evs & e) != 0){
	    if (F_ISSET(ev->se_evs,S_BANDURG) && F_ISSET(ev->se_evs,S_RDBAND))
	    {
		if (LIS_DEBUG_SIG)
		    printk("kill_proc(pid=%d, sig=0x%x)\n", ev->se_pid, SIGURG);
		if ((res=kill_proc(ev->se_pid,SIGURG,1))<0)
		    lis_error(LIS_ERROR,
			      "kill_procs","kill_proc: errno %d",res);

		if (ev->se_evs& e& ~S_RDBAND) /* other events -> sigpoll too */
		{
		    if (LIS_DEBUG_SIG)
			printk("kill_proc(pid=%d, sig=0x%x)\n",
				ev->se_pid, SIGPOLL);
		    if ((res=kill_proc(ev->se_pid,SIGPOLL,1))<0)
			lis_error(LIS_ERROR,
				  "kill_procs","kill_proc: errno %d",res);
		}
	    }
	    else
	    {
		if (LIS_DEBUG_SIG)
		    printk("kill_proc(pid=%d, sig=0x%x)\n",
			    ev->se_pid, SIGPOLL);
		if ((res=kill_proc(ev->se_pid,SIGPOLL,1))<0)
		    lis_error(LIS_ERROR,"kill_procs","kill_proc: errno %d",res);
	    }
	}
    
}/*kill_procs*/
/*  -------------------------------------------------------------------  */

/* Will send relevant signals to the usr on arrival of M_[PC]SIG containing
 * `sig' at the stream head.
 */
static void 
signalusr(int sig, stdata_t *hd)
{
    if (sig==SIGPOLL)
	kill_procs(hd->sd_siglist,SIGPOLL,SIGPOLL); /* ?? !!! */
    else if (F_ISSET(hd->sd_flag,STRISTTY))
	kill_pg(hd->sd_pgrp,sig,1);
}/*signalusr*/
/*  -------------------------------------------------------------------  */

/* Send an M_READ msg downstream
 */
static int
snd_mread(queue_t *wq, unsigned int ulen)
{
    mblk_t *mread = allocb(sizeof(ulen),BPRI_LO);
    if (mread == NULL)
	return(-ENOMEM);
    else {
	lis_btype(mread)=M_READ;
	memcpy(mread->b_wptr,(char*)&ulen,sizeof(ulen));
	mread->b_wptr+=sizeof(ulen);
	PUTNEXT(wq,mread);
	return(0);
    }
}/*snd_mread*/

/* Send an M_FLUSH msg downstream
 */
static int
lis_snd_mflush(queue_t *wq, int flag, int band)
{
    mblk_t *mf = allocb(2*sizeof(int),BPRI_LO); /*sizeof mflush msg (?)*/
    if (mf == NULL)
	return(-ENOMEM);
    else {
	lis_btype(mf)=M_FLUSH;
	lis_putbyte(&(mf->b_wptr),(char)flag);
	if (F_ISSET(flag,FLUSHBAND))
	    lis_putbyte(&(mf->b_wptr),(char)band);
	PUTNEXT(wq,mf);
	return(0);
    }
}/*lis_snd_mflush*/

/*  -------------------------------------------------------------------  */
/* Send an ioctl negative ack  msg downstream reusing mp (M_IOCTL) msg.
 */
static int
snd_iocnak(queue_t *wq, mblk_t *mp,int code)
{
    iocblk_t *iocb = (iocblk_t*)(mp->b_rptr);
    mblk_t   *dtab;
    lis_btype(mp)=M_IOCNAK;
    iocb->ioc_error=code;
    dtab=lis_unlinkb(mp);
    lis_freemsg(dtab);
    PUTNEXT(wq,mp);
    return(0);
}/*snd_iocnak*/


/*  -------------------------------------------------------------------  */

/* Is this ioc blk out of sequence?
 */
static int
bad_ioc_seq( mblk_t *mp, stdata_t *hd)
{
    iocblk_t *iocb = (iocblk_t*)mp->b_rptr;

    return(iocb->ioc_id != (uint) hd->sd_iocseq);

}/*bad_ioc_seq*/

/*  -------------------------------------------------------------------  */

/* Call every open procedure in this queue
 */
static int
open_mods( queue_t *q, dev_t dev, int flags, cred_t *creds)
{
    int 	 err;
    queue_t	*rdq ;

    if (q == NULL)
	return(-ENOPKG) ;			/* pkg not installed */

    for (q=LIS_WR(q),err=0; !err && q != NULL; q=q->q_next)
    {
	if (!LIS_CHECK_Q_MAGIC(q)) return(-EINVAL) ;

	rdq = LIS_RD(q) ;
	if (   rdq->q_qinfo == NULL		/* ensure qinfo */
	    || rdq->q_qinfo->qi_qopen == NULL)	/* ensure open routine */
	    return(-ENOPKG) ;			/* pkg not installed */

	if ( ( err =
		  (*(rdq->q_qinfo->qi_qopen))
		  		(rdq, &dev, flags,
				 q->q_next != NULL ? MODOPEN : DRVOPEN,
				 creds)
	     ) != 0
	   )
	    return(err > 0 ? -err : err);
    }
    return(0);
}/*open_mods*/

/*  -------------------------------------------------------------------  */
/* Call every close procedure in this queue
 */
#if 0				/* not used */
static void
close_mods( queue_t *q, dev_t dev, int flags, cred_t *creds)
{
  int err;
    (void) dev ;
    for (q=LIS_WR(q),err=0; !err && q != NULL; q=q->q_next)
    {
      if (!LIS_CHECK_Q_MAGIC(q)) break ;
      if (q->q_qinfo != NULL && q->q_qinfo->qi_qclose != NULL)
	(*(q->q_qinfo->qi_qclose))(LIS_RD(q),flags,creds);
    }

}/*close_mods*/
#endif

/*  -------------------------------------------------------------------  */
/* lis_strrput - stream head read put procedure
 *
 * TBD:  Must consider interrupt protection
 */
void
lis_strrput(queue_t *q, mblk_t *mp)
{
    stdata_t *shead ;
    int typ ;

    if (q==NULL){
	lis_error(LIS_ERROR,"lis_strrput","no queue, is this vacation?");
	lis_freemsg(mp);
	return;
    }
    else if ( (shead=q->q_ptr)==NULL){
	lis_error(LIS_ERROR,"lis_strrput","no head (null) for queue 0x%x",q);
	lis_freemsg(mp);
	return;
    }
    else if (shead->magic != STDATA_MAGIC)
    {
	lis_error(LIS_ERROR,"lis_strrput","stream head invalid magic number");
	lis_freemsg(mp);
	return;
    }
    else if (!mp || !mp->b_datap){
	lis_error(LIS_ERROR,"lis_strrput","absent msg for queue 0x%x",q);
	lis_freemsg(mp);
	return;
    }
    else if (!LIS_CHECK_Q_MAGIC(q))
    {
	lis_freemsg(mp);
	return;
    }

    /* ok, do the job */

    typ = lis_btype(mp);
    if (F_ISSET(shead->sd_flag, STRCLOSE))	/* stream is closing */
    {						/* no file above stream */
	switch (typ)
	{
	case M_PCPROTO:
	case M_DATA:
	case M_PROTO:
	case M_PASSFP:
	case M_SIG:
	case M_PCSIG:
	case M_COPYIN: 
	case M_COPYOUT:
	    if (LIS_DEBUG_STRRPUT)
	      printk(
	       "lis_strrput: discarding %s from \"%s\" on closed stream %s%s\n",
			lis_msg_type_name(mp), lis_queue_name(backq(q)),
			lis_strm_name(shead), lis_maj_min_name(shead)) ;
	    lis_freemsg(mp) ;			/* no hope of retrieving msg */
	    return ;
	}
    }

    if (LIS_DEBUG_STRRPUT)
	printk("lis_strrput: received %s (%d) from \"%s\" on stream %s%s size %d\n",
		lis_msg_type_name(mp), typ, lis_queue_name(backq(q)),
		lis_strm_name(shead), lis_maj_min_name(shead),
		lis_msgsize(mp)) ;

    switch(typ){
    case M_PCPROTO:
	if (F_ISSET(shead->sd_flag,STRPRI)){
	    lis_freemsg(mp);
	    return;
	}
	else 
	    F_SET(shead->sd_flag,STRPRI);
	/* fall thru... */
    case M_DATA:
    case M_PROTO:
    case M_PASSFP:
	LisUpCount(MSGQDSTRHD) ;
        putq(q,mp);
	/*
	 * TBD: should this really be done first??
	 */
	if (F_ISSET(shead->sd_flag,RSLEEP)){
	    lis_wake_up_wread(shead);
	    F_CLR(shead->sd_flag,RSLEEP);
	}

	if (   shead->sd_sigflags == 0			/* no signals */
	    && POLL_NOT_WAITING(shead)			/* no polling */
	    && !F_ISSET(shead->sd_flag,STRSELPND)	/* no selecting */
	   )
	    break ;

	if (typ == M_PCPROTO || mp == q->q_first)
	{				/* hipri msg or 1st msg in queue */
	    short evs=0;
	    short polls=0 ;

	    if (lis_hipri(typ))
	    {
		F_SET(polls,POLLPRI) ;
		F_SET(evs,S_HIPRI);
	    }
	    else
	    {
		F_SET(polls,POLLIN) ;
		F_SET(evs,S_INPUT);

		if (lis_bband(mp)==0)
		{
		    F_SET(polls,POLLRDNORM) ;
		    F_SET(evs,S_RDNORM);
		}

		if (lis_bband(mp)>0)
		{
		    F_SET(polls,POLLRDBAND) ;
		    F_SET(evs,S_RDBAND);
		}
	    }

	    if (evs &= shead->sd_sigflags)
		kill_procs(shead->sd_siglist,SIGPOLL,evs);

	    if (   polls
		&& (   POLL_WAITING(shead)
		    || F_ISSET(shead->sd_flag,STRSELPND)
		   )
	       )
		lis_wake_up_poll(shead,polls);
	}
	break;
    case M_ERROR:
	{
	    unsigned char	errnum ;

	    switch (mp->b_wptr - mp->b_rptr) 	/* how many bytes?  */
	    {
	    default:				/* not valid format */
		mp->b_rptr = mp->b_datap->db_base;
		mp->b_wptr = mp->b_rptr ;
		*mp->b_wptr++ = EINVAL ;	/* supply default errnum */
		/*
		 * Fall into the one-byte case
		 */
	    case 1:				/* set errnum for both sides */
		errnum = *mp->b_rptr++;
		if (errnum != NOERROR)
		{
		    shead->sd_rerror = errnum ;
		    shead->sd_werror = errnum ;

		    if (errnum == 0)
		    {
			F_CLR(shead->sd_flag,STRDERR);
			F_CLR(shead->sd_flag,STWRERR);
		    }
		    else
		    {
			F_SET(shead->sd_flag,STRDERR);
			F_SET(shead->sd_flag,STWRERR);
		    }
		}
		break ;

	    case 2:				/* set two sides individually */
		errnum = *mp->b_rptr++;
		if (errnum != NOERROR)
		{
		    shead->sd_rerror = errnum ;

		    if (errnum == 0)
			F_CLR(shead->sd_flag,STRDERR);
		    else
			F_SET(shead->sd_flag,STRDERR);
		}

		errnum = *mp->b_rptr++;
		if (errnum != NOERROR)
		{
		    shead->sd_werror = errnum ;

		    if (errnum == 0)
			F_CLR(shead->sd_flag,STWRERR);
		    else
			F_SET(shead->sd_flag,STWRERR);
		}

		break ;
	    }

	    lis_wake_up_wwrite(shead);
	    lis_wake_up_wread(shead);
	    F_CLR(shead->sd_flag,RSLEEP);
	    F_CLR(shead->sd_flag,WSLEEP);
	    if (shead->sd_werror || shead->sd_rerror)
		lis_snd_mflush(LIS_WR(q),FLUSHW,0);
	    if (F_ISSET(shead->sd_sigflags,S_ERROR))
		kill_procs(shead->sd_siglist,SIGPOLL,S_ERROR);

	    if (   POLL_WAITING(shead)
		|| F_ISSET(shead->sd_flag,STRSELPND)
	       )
		lis_wake_up_poll(shead,POLLERR);

	    lis_freemsg(mp);
	    break;
	}
    case M_HANGUP:
	F_SET(shead->sd_flag,STRHUP);
	lis_wake_up_wwrite(shead);
	lis_wake_up_wread(shead);
	F_CLR(shead->sd_flag,RSLEEP);
	F_CLR(shead->sd_flag,WSLEEP);
	if (F_ISSET(shead->sd_sigflags,S_HANGUP))
	    kill_procs(shead->sd_siglist,SIGPOLL,S_HANGUP);
	if (   POLL_WAITING(shead)
	    || F_ISSET(shead->sd_flag,STRSELPND)
	   )
	    lis_wake_up_poll(shead,POLLHUP);

	lis_freemsg(mp);
	break;

    case M_FLUSH:
	if (*mp->b_rptr & FLUSHR)
	{					/* flush read queue */
	    int		msgs_before = lis_qsize(q) ;

	    if (F_ISSET(*mp->b_rptr,FLUSHBAND))
		lis_flushband(LIS_RD(q),mp->b_rptr[1],FLUSHDATA);
	    else
		lis_flushq(LIS_RD(q),FLUSHDATA);
	    *mp->b_rptr &= ~FLUSHR ;		/* turn off flush read bit */
	    LisDownCounter(MSGQDSTRHD, msgs_before - lis_qsize(q)) ;
	}

	if (*mp->b_rptr & FLUSHW)
	{					/* flush write queue */
	    if (F_ISSET(*mp->b_rptr,FLUSHBAND))
		lis_flushband(LIS_WR(q),mp->b_rptr[1],FLUSHDATA);
	    else
		lis_flushq(LIS_WR(q),FLUSHDATA);

	    if (!(mp->b_flag & MSGNOLOOP))
	    {
		mp->b_flag |= MSGNOLOOP;
		PUTNEXT(LIS_WR(q),mp);	/* send back downstream */
	    }
	    else
	        lis_freemsg(mp);	/* free message */
	}
	else					/* not flushing write */
	    lis_freemsg(mp);

	break;

    case M_SIG:
	LisUpCount(MSGQDSTRHD) ;
	putq(q,mp);			/* strread will signalusr() */
	if (F_ISSET(shead->sd_flag,RSLEEP))
	{
	    lis_wake_up_wread(shead);
	    F_CLR(shead->sd_flag,RSLEEP);
	}

	if (   POLL_WAITING(shead)
	    || F_ISSET(shead->sd_flag,STRSELPND)
	   )
	    lis_wake_up_poll(shead,POLLMSG);
	break;
    case M_PCSIG:
	signalusr(*mp->b_rptr,shead);
	if (   POLL_WAITING(shead)
	    || F_ISSET(shead->sd_flag,STRSELPND)
	   )
	    lis_wake_up_poll(shead,POLLMSG);
	lis_freemsg(mp);
	break;
    case M_IOCNAK: 
    case M_IOCACK: 
	if (!F_ISSET(shead->sd_flag,IOCWAIT)|| 
	    F_ISSET(shead->sd_flag,(STWRERR|STRDERR)) ||
	    bad_ioc_seq(mp,shead)
	   )
	    lis_freemsg(mp);
	else {			/* everything's ok: now we process it */
	    shead->sd_iocblk=mp;
	    lis_wake_up_wiocing(shead); /* this is not an unlock! */
	}
	break ;
    case M_COPYIN: 
	if (!F_ISSET(shead->sd_flag,IOCWAIT))
	    lis_freemsg(mp);
	else if (F_ISSET(shead->sd_flag,STWRERR) ||
		 F_ISSET(shead->sd_flag,STRDERR) || bad_ioc_seq(mp,shead)){
	    copyresp_t *res=(copyresp_t*)mp->b_rptr;
	    mp->b_wptr=(char*)(res+1);
	    mp->b_datap->db_type = M_IOCDATA ;
	    res->cp_rval=(caddr_t)1;
	    PUTNEXT(LIS_WR(q),mp);
	}
	else {			/* everything's ok: now we process it */
	    shead->sd_iocblk=mp;
	    lis_wake_up_wiocing(shead); /* this is not an unlock! */
	}
	break;
    case M_COPYOUT:
	if (!F_ISSET(shead->sd_flag,IOCWAIT))
	    lis_freemsg(mp);
	else if (F_ISSET(shead->sd_flag,STWRERR) ||
		 F_ISSET(shead->sd_flag,STRDERR) ||  bad_ioc_seq(mp,shead)){
	    copyresp_t *res=(copyresp_t*)mp->b_rptr;
	    mp->b_wptr=(char*)(res+1);
	    mp->b_datap->db_type = M_IOCDATA ;
	    res->cp_rval=(caddr_t)1;
	    PUTNEXT(LIS_WR(q),mp);
	}
	else {			/* everything's ok: now we process it */
	    shead->sd_iocblk=mp;
	    lis_wake_up_wiocing(shead); /* this is not an unlock! */
	}
	break;
    case M_SETOPTS:
	set_options(shead,(stroptions_t*)mp->b_rptr);
	lis_freemsg(mp);
	break ;
    case M_IOCTL:
	snd_iocnak(LIS_WR(q),mp,1);
	break;
    default:
	lis_freemsg(mp);
	break ;
    }

}/*lis_strrput*/


/*  -------------------------------------------------------------------  */
/*  lis_strwsrv
 *
 * This is the stream head write service procedure.  It is used
 * for detecting that the stream below the stream head no longer
 * has messages queued for output so the stream can be closed.
 */

void	lis_strwsrv(queue_t *q)
{
    stdata_t	*hd ;
    mblk_t	*mp ;
    int		 couldput ;

    if (!LIS_CHECK_Q_MAGIC(q)) return ;

    hd = (stdata_t *) q->q_ptr ;
    if (hd == NULL)				/* should not happen */
    {
	printk("lis_strwsrv: queue %s: q_ptr is NULL\n", lis_queue_name(q));
	return ;
    }
    
    if (hd->magic != STDATA_MAGIC || hd->sd_wq != q) return ;

    while (   (couldput = lis_bcanputnext(q,0))	/* can send downstream */
	   && (mp = getq(q)) != NULL		/* have queued messages */
	  )
    {
	PUTNEXT(q, mp) ;			/* send downstream */
    }

    if (   F_ISSET(hd->sd_flag, STRCLOSE)	/* stream is closing */
	&& lis_qcountstrm(q) == 0 		/* queues are empty */
       )
   {
	if ( LIS_DEBUG_CLOSE )
	    printk("lis_strwsrv: stream %s%s empty, wake up close routine\n",
		    lis_strm_name(hd),
		    lis_maj_min_name(hd)) ;

	lis_wakeup_close((caddr_t) hd) ;	/* wake up close */
   }

   if (hd->sd_sigflags != 0)
   {
	if (F_ISSET(hd->sd_sigflags,S_OUTPUT)
	    && lis_bcanputnext(q,0)
	   )
	    kill_procs(hd->sd_siglist,SIGPOLL,S_OUTPUT);

	/*
	 * AT&T says in STREAMS Programmer's Guide:
	 *
	 * "S_WRBAND
	 *
	 * "A priority band greater than 0 of a queue downstream exists
	 * and is writeable.  This notifies the user that there is room
	 * on the queue for sending (or writing) priority data downstream."
	 */
	if (F_ISSET(hd->sd_sigflags,S_WRBAND)
	    && lis_bcanputnext_anyband(q)
	   )
	    kill_procs(hd->sd_siglist,SIGPOLL,S_WRBAND);
   }

   if (couldput && F_ISSET(hd->sd_flag,WSLEEP))	/* write waiting flow ctrl */
   {
	F_CLR(hd->sd_flag,WSLEEP);		/* clear waiting flag */
	lis_wake_up_wwrite(hd);			/* wake up write/putmsg */
   }

    if (  (   POLL_WAITING(hd)
	   || F_ISSET(hd->sd_flag,STRSELPND)
          )
	&& lis_bcanputnext(q,0)			/* can still send downstream */
       )
    {
	lis_wake_up_poll(hd,POLLOUT);
    }

} /* lis_strwsrv */


/*  -------------------------------------------------------------------  */
/* This is the tmout handling function.
 * it will be called on response to a tmout and is responsible to
 * take the right actions.
 * It should be called w/ an stream head as argument.
 */
static void
lis_do_tmout( unsigned long arg )
{
    stdata_t *hd = (stdata_t*)arg;

    if (hd->magic != STDATA_MAGIC){
	lis_error(LIS_ERROR,"lis_do_tmout","timeout with bad stream 0x%x",arg);
	return;
    }
    if (F_ISSET(hd->sd_flag,STIOCTMR)){     /* ioctl timer: awake sleeper*/
	F_CLR(hd->sd_flag,STIOCTMR);
	lis_wake_up_wiocing(hd);
    }
}/*lis_do_tmout*/


/*  -------------------------------------------------------------------  */
/* lis_wait_for_wiocing - wait until ioc response arrived or timeout occur
 */
static int
lis_wait_for_wiocing(stdata_t *hd, int tmout)
{
    struct timer_list tl;
    int		      rslt ;

    if (tmout == 0)
	tmout = LIS_DFLT_TIM ;

    while( hd->sd_iocblk == NULL )
    {
	if ( F_ISSET(hd->sd_flag,(STRDERR|STWRERR)) )
	    return(-hd->sd_rerror) ;
	F_SET(hd->sd_flag,STIOCTMR);
	if (tmout != INFTIM)
	    lis_tmout(&tl,lis_do_tmout,(long)hd, SECS_TO(tmout));
	rslt = lis_sleep_on_wiocing(hd);
	if (rslt < 0)
	{
	    if (tmout != INFTIM)
		lis_untmout(&tl);
	    F_CLR(hd->sd_flag,STIOCTMR);
	    return(rslt) ;
	}

	if (!F_ISSET(hd->sd_flag,STIOCTMR)) /* timer expired */
	    return(-ETIME);
	else
	{
	    if (tmout != INFTIM)
		lis_untmout(&tl);
	    F_CLR(hd->sd_flag,STIOCTMR);
	}
    }
    return(0);			/* got msg */
}/*lis_wait_for_wiocing*/

/*  -------------------------------------------------------------------  */

/* copyout bmlks for reads
 */
static int 
copyout_blks(struct file *f, char *ubuff, long count, mblk_t *mp)
{
    mblk_t *mb;
    long ocount=count;
    int len;
    int err = 0 ;

    for (mb=mp ; mb && count > 0; mp=mb)
    {
	mb=mp->b_cont;				/* next buffer in chain */
	mp->b_cont = NULL ;			/* unchain this buffer */
	if (err == 0)
	{
	    len=min(count,lis_mblksize(mp));	/* size of current bfr */
	    err = lis_copyout(f,mp->b_rptr,ubuff,len); /* cpy out current bfr */
	    count -=len;
	    ubuff +=len;
	}
	lis_freeb(mp);				/* free current buffer */
    }

    if (err < 0)
	return(err) ;				/* something went wrong */

    if (count != 0)			/* hopefully, copied whole buffer */
	lis_error(LIS_WARN,"copyout_blks",
		  "count (%d) doesn't match data (%d)", ocount, ocount-count);
    return(0) ;
}

/*  -------------------------------------------------------------------  */
/* strdoioctl - ioctl handler used by strioctl
 *
 * There is a special circumstance in which this routine is called
 * with a NULL pointer for 'f'.  This is from the i_unlink code in
 * which the file is not known.  We do not worry about null pointer
 * reference here since our only use of 'f' occurs with 'do_copyin'
 * is passed as a "true", which it is not from i_unlink code.
 */
static int
lis_strdoioctl(struct file *f, stdata_t *hd,
		strioctl_t *ioc, cred_t *creds, int do_copyin)
{
    mblk_t *mioc,*mdta;
    iocblk_t *iocb;
    int err = 0 ;
    unsigned long  time_cell = 0 ;

    CLOCKON() ;
    if (ioc->ic_len != TRANSPARENT && (ioc->ic_len < 0 || ioc->ic_timout < -1))
    {
	ioc->ic_len = 0 ;		/* no data */
	CLOCKOFF(IOCTLTIME) ;
	return(-EINVAL);
    }
    if ((mioc=allocb(sizeof(copyreq_t),BPRI_LO))==NULL)
    {
	ioc->ic_len = 0 ;		/* no data */
	CLOCKOFF(IOCTLTIME) ;
	return(-ENOMEM);
    }
    if (ioc->ic_len){
	if (do_copyin && 
	    (err=lis_check_umem(f,VERIFY_WRITE,ioc->ic_dp,ioc->ic_len))<0)
	{
	    ioc->ic_len = 0 ;		/* no data */
	    CLOCKOFF(IOCTLTIME) ;
	    return(err);
	}
	if ((mdta=allocb(
	            ioc->ic_len == TRANSPARENT ? sizeof(ioc->ic_dp)
					       : ioc->ic_len,BPRI_LO))==NULL)
	{
	    ioc->ic_len = 0 ;		/* no data */
	    lis_freemsg(mioc);
	    CLOCKOFF(IOCTLTIME) ;
	    return(-ENOMEM);
	}
	if (do_copyin)
	{
	    if ((err = lis_copyin(f,mdta->b_wptr,ioc->ic_dp,ioc->ic_len)) < 0)
	    {
		ioc->ic_len = 0 ;		/* no data */
		lis_freemsg(mioc);
		CLOCKOFF(IOCTLTIME) ;
		return(err) ;
	    }
	    mdta->b_wptr += ioc->ic_len;
	}
	else
	if (ioc->ic_len != TRANSPARENT)
	{
	    memcpy(mdta->b_wptr,ioc->ic_dp,ioc->ic_len);
	    mdta->b_wptr += ioc->ic_len;
	}
	else
	{
	    void	**ptr_ptr = (void **) mdta->b_wptr ;
	    *ptr_ptr = ioc->ic_dp ;		/* pass in ptr */
	    mdta->b_wptr += sizeof(*ptr_ptr) ;
	}
	lis_linkb(mioc,mdta);
    }
    iocb=(iocblk_t*)mioc->b_wptr;
    mioc->b_wptr += sizeof(iocblk_t);
    mioc->b_datap->db_type = M_IOCTL ;
    iocb->ioc_cmd=ioc->ic_cmd;
    iocb->ioc_uid=creds->cr_uid;
    iocb->ioc_gid=creds->cr_gid;
    iocb->ioc_count=ioc->ic_len;
    iocb->ioc_error=0;
    iocb->ioc_rval=0;
    iocb->ioc_id = hd->sd_iocseq =  ++lis_iocseq;
  again:
    if (err < 0 || (err = lis_lock_wioc(hd)) < 0)
    {
	ioc->ic_len = 0 ;		/* no data */
	lis_freemsg(mioc);
	CLOCKOFF(IOCTLTIME) ;
	return(err) ;
    }

    if ( LIS_DEBUG_IOCTL )
	printk("lis_strdoioctl: send %s to \"%s\" on stream %s%s, cmd(0x%x)\n",
			lis_msg_type_name(mioc),
			lis_queue_name(hd->sd_wq->q_next),
			lis_strm_name(hd), lis_maj_min_name(hd), iocb->ioc_cmd) ;
    F_SET(hd->sd_flag,IOCWAIT);		/* for strrput */
    PUTNEXT(hd->sd_wq,mioc);
    mioc=NULL;			
    CLOCKADD() ;
    if (   (err=lis_wait_for_wiocing(hd,ioc->ic_timout))<0
        || (mioc = hd->sd_iocblk) == NULL		/* rtnd ioctl msg */
       )
    {
	F_CLR(hd->sd_flag,IOCWAIT);	/* now completed */
	ioc->ic_len = 0 ;		/* no data */
	lis_unlock_wioc(hd);
	CLOCKOFF(IOCTLTIME) ;
	return(err);
    }
    /* Here we have! */
    CLOCKON() ;
    F_CLR(hd->sd_flag,IOCWAIT);		/* now completed */
    hd->sd_iocblk = NULL ;
    if ( LIS_DEBUG_IOCTL )
	printk("lis_strdoioctl: %s from \"%s\" on stream %s%s\n",
		    lis_msg_type_name(mioc),
		    lis_queue_name(hd->sd_wq->q_next),
		    lis_strm_name(hd), lis_maj_min_name(hd)) ;

    switch(lis_btype(mioc))
    {
    case M_IOCACK:
	iocb=(iocblk_t*)mioc->b_rptr;
	if (iocb->ioc_error){
	    ioc->ic_len = 0 ;		/* no data */
	    lis_freemsg(mioc);
	    lis_unlock_wioc(hd);
	    CLOCKOFF(IOCTLTIME) ;
	    return(-(iocb->ioc_error));
	}

	{
	    int rval=iocb->ioc_rval;
	    int len = iocb->ioc_count ;
	    mblk_t *dat=lis_unlinkb(mioc);

	    if (do_copyin)
	    {
		if (len > 0)
		{				/* data to return */
		    if ((err=lis_check_umem(f,VERIFY_WRITE,ioc->ic_dp,len)) < 0)
			rval = err ;		/* oops... */
		    else
		    {
			copyout_blks(f,ioc->ic_dp, len, dat);
			dat = NULL;		/* see copyout_blks() */
			ioc->ic_len = len ;	/* length */
		    }
		}
		else				/* no data to return */
		if (dat)			/* however, have empty bfr */
		    ioc->ic_len = 0 ;		/* so set len to zero */
	    }

	    if (dat)
		lis_freemsg(dat);

	    lis_freemsg(mioc);
	    lis_unlock_wioc(hd);
	    CLOCKOFF(IOCTLTIME) ;
	    return(rval);
	}
    case M_IOCNAK:
	iocb=(iocblk_t*)mioc->b_rptr;
	{
	    int res=-(iocb->ioc_error);
	    ioc->ic_len = 0 ;		/* no data */
	    lis_freemsg(mioc);
	    lis_unlock_wioc(hd);
	    CLOCKOFF(IOCTLTIME) ;
	    return(res ? res : -EINVAL);
	}
	break ;
    case M_COPYIN:
	{
	    copyreq_t *cp=(copyreq_t*)mioc->b_rptr;
	    copyresp_t*cr=(copyresp_t*)mioc->b_rptr;
	    u_int      len=cp->cq_size;
	    mblk_t    *dat=NULL;
	    char      *ubuf=cp->cq_addr;

	    err = 0 ;

	    dat = mioc->b_cont ;
	    if (dat != NULL)
	    {
		mioc->b_cont = NULL ;
		freemsg(dat) ;
		dat = NULL ;
	    }
	    lis_btype(mioc)=M_IOCDATA;
	    if (   (err=lis_check_umem(f,VERIFY_READ,ubuf,len)) < 0
		|| (dat=allocb(len,BPRI_LO))==NULL
		|| (err = lis_copyin(f,dat->b_wptr,ubuf,len)) < 0
	       )
	    {
		if (err < 0)
		    cr->cp_rval=(caddr_t)(-err);
		else
		    cr->cp_rval=(caddr_t)ENOMEM;
		if (dat != NULL)
		    lis_freemsg(dat);
	    }
	    else
	    {
		dat->b_wptr += len ;
		lis_linkb(mioc,dat);
		cr->cp_rval=0;
	    }

	    lis_unlock_wioc(hd);
	    mioc->b_wptr = mioc->b_rptr + sizeof(*cr) ;
	    goto again;
	}
    case M_COPYOUT:
	{
	    copyreq_t *cp=(copyreq_t*)mioc->b_rptr;
	    copyresp_t*cr=(copyresp_t*)mioc->b_rptr;
	    char      *ubuf=cp->cq_addr;
	    mblk_t    *dat;
	    u_int      len=cp->cq_size;

	    lis_btype(mioc)=M_IOCDATA;
	    if ((err=lis_check_umem(f,VERIFY_WRITE,ubuf,len)) >= 0){
		dat=lis_unlinkb(mioc);
		if ((err = copyout_blks(f,ubuf,len,dat)) < 0)
		    cr->cp_rval=(caddr_t)(-err);
		else
		    cr->cp_rval=0;
	    } 
	    mioc->b_wptr = mioc->b_rptr + sizeof(*cr) ;
	    lis_unlock_wioc(hd);
	    goto again;
	}
    default:
	ioc->ic_len = 0 ;		/* no data */
	lis_freemsg(mioc);
	lis_unlock_wioc(hd);
	CLOCKOFF(IOCTLTIME) ;
	return(-EINVAL) ;
    }

}/*lis_strdoioctl*/

/*  -------------------------------------------------------------------  */
/*
 * lis_lookup_stdata
 *
 * Look through all the stdata structures for one that has the
 * given device id.  Return a pointer to it.
 */
static
stdata_t *lis_lookup_stdata(dev_t dev, stdata_t *not_this_one)
{
    stdata_t		*hd ;
    int			 n ;

    if (lis_stdata_head == NULL)
	return(NULL) ;			/* no strdata structs allocated */

    for (hd = lis_stdata_head, n = 0;
	 n == 0 || hd != lis_stdata_head;
	 hd = hd->sd_next, n++)
    {
	if (hd != not_this_one && hd->sd_dev == dev)
	    return(hd) ;		/* found it */
    }

    return(NULL) ;			/* not found */

} /* lis_lookup_stdata */




/*  -------------------------------------------------------------------  */
/*			Exported functions & macros                      */
/*  -------------------------------------------------------------------  */





/*  -------------------------------------------------------------------  */
/* lis_stropen - open a stream (see Magic Garden 7.9.3 and 7.9.4)
 *
 * The stream head represents a single stream.  It has the status of
 * a file in the operating system.  As such it can be referenced by
 * multiple files and multiple inodes.  Therefore, several inodes
 * can be pointing to the same stdata structure.  The use count
 * in the stdata structure keeps track of the number of such inward
 * pointers.  There is NO back pointer from stdata to inodes.
 */

int 
lis_stropen( struct inode *i, struct file *f )
{
    struct stdata *head = NULL;
    struct queue  *q = NULL;
    struct queue  *modq = NULL;
#if 0
    streamtab_t	  *st;
#endif
    const char	  *name = "(unknown)";
    const char	  *majmin = "(unknown)";
    cred_t	   creds;
    dev_t	   odev ;
    dev_t	   ndev ;
    int		   err = 0;
    int		   hd_locked = 0 ;
    int		   maj;
#if 0
    int		   a;
#endif
    int		   psw ;
    unsigned long  time_cell = 0 ;

    CLOCKON() ;

#ifdef LINUX
    lis_start_qsched() ;		/* ensure q running process going */
#endif

#ifdef KERNEL_2_1				/* linux >= 2.1.0 */
#define	PRINT_ADDRS				\
	    printk("             i=%lx f=%lx f->f_dentry=%lx\n", \
			  (long) i, (long) f, (long) f->f_dentry)
#else						/* linux < 2.1.0 or portable */
#define	PRINT_ADDRS				\
	    printk( "             i=%lx f=%lx f->f_inode=%lx\n", \
			   (long) i, (long) f, (long) f->f_inode)
#endif

    lis_open_cnt++ ;
    if (LIS_DEBUG_OPEN)
    {
	printk("lis_stropen: "
	       "call number %ld i_rdev=(%d,%d) i_count=%ld f_count=%d"
#ifdef KERNEL_2_1				/* linux >= 2.1.0 */
	       " d_count=%d\n",
#else
	       "\n",
#endif
	       lis_open_cnt, STR_MAJOR(i->i_rdev), STR_MINOR(i->i_rdev),
	       (long) i->i_count, f->f_count
#ifdef KERNEL_2_1				/* linux >= 2.1.0 */
	       , f->f_dentry->d_count
#endif
	       ) ;

	if (LIS_DEBUG_ADDRS) PRINT_ADDRS ;
    }

    CHECK_INO(i,"lis_stropen");	/* may return */
    odev = i->i_rdev;
    maj=STR_MAJOR(odev);
    head=INO_STR(i);
    if (head == NULL)			/* opening new stream */
	head = lis_lookup_stdata(i->i_rdev, NULL) ;
					/* chk for duplicate dev open */
    creds.cr_uid  = (uid_t) EUID(f);
    creds.cr_gid  = (gid_t) EGID(f);
    creds.cr_ruid = (uid_t) UID(f);
    creds.cr_rgid = (gid_t) GID(f);
    if (head != NULL)			/* via inode or via lookup */
    {
	if (head->magic != STDATA_MAGIC)
	{
	    printk("lis_stropen: "
	    	   "head@%lx has magic number %lx, should be %lx\n",
	    		(long)head, head->magic, STDATA_MAGIC) ;
	    CLOCKOFF(OPENTIME) ;
	    return(-EINVAL) ;
	}

	/* Stream was created by previous open of device: sleep if still
           opening, then call every module open() and awake sleepers */

	SPLSTR(psw) ;
	head->sd_refcnt++;		/* another use counted */
	SPLX(psw) ;
	lis_inc_mod_cnt() ;
	LisUpCount(MODUSE) ;

	INO_STR(i) = head ;		/* point inode to strm hd */
	err = lis_lock_wopen(head);	/* get exclusive use of the head */
	if (err < 0)
	    goto error_rtn ;

	hd_locked = 1 ;

	if ( LIS_DEBUG_OPEN )
	{
	    name = lis_strm_name(head) ;
	    majmin = lis_maj_min_name(head) ;
	}

	modq = LIS_RD(head->sd_wq->q_next);

	if ((err=open_mods(modq,i->i_rdev,f->f_flags,&creds))<0)
	    goto error_rtn ;

	goto successful_rtn ;
    }

    /* First open of the stream: allocate a queue pair and stream head if
       STREAMS is not already using too much memory */
    if (!stravail(f))
    {
	if (LIS_DEBUG_OPEN)
	    printk("lis_stropen: not enough memory to open a stream\n") ;

	CLOCKOFF(OPENTIME) ;
	return(-ENOSR);
    }

    q=lis_allocq("stream-head");
    if (!q){
	if (LIS_DEBUG_OPEN)
	    printk("lis_stropen: failed to allocate a queue for stream head\n") ;

	CLOCKOFF(OPENTIME) ;
	return(-ENOMEM);
    }

    /* initialize the stream head and set the STWOPEN flag set
       vp->v_stream to point to stream head set q_ptr fields in queue
       pair to point to stream head */
    if ((head=alloc_stdata(i))==NULL){		/* allocate and lock */
	if (LIS_DEBUG_OPEN)
	    printk("lis_stropen: failed to allocate stream head\n") ;

	err = -ENOMEM;
	goto error_rtn ;
    }

    /*
     * alloc_stdata() returned with the stdata structure locked
     */
    INO_STR(i) = head ;				/* point inode to strm hd */
    hd_locked = 1 ;
    head->sd_refcnt=1;
    lis_inc_mod_cnt() ;
    LisUpCount(MODUSE) ;

    LIS_RD(q)->q_ptr=head;
    LIS_WR(q)->q_ptr=head;
    head->sd_wq = LIS_WR(q);
    lis_setq(q, &strmhd_rdinit, &strmhd_wrinit);
    head->sd_rdopt = RPROTNORM | RNORM ;
    head->sd_wropt = 0 ;

    /* head->sd_session=SESSION();	DMG not used??  */
    head->sd_pgrp= PGRP(f);

    /* TBD: need a better default for LIS_MAXPSZ */
    head->sd_maxpsz   = LIS_MAXPSZ; /* probably not to be done so */
    head->sd_minpsz   = LIS_MINPSZ;

#ifdef FIFO_IMPL
//	lis_fifo_strtab not yet defined !!
    if (LIS_ISDEV(maj,LIS_FIFO))
	head->sd_strtab= &lis_fifo_strtab;
    else
#endif
	head->sd_strtab= LIS_DEVST(maj).f_str;	/* use strmtab of opened strm */

    head->sd_dev = odev ;
    ndev = odev ;
    name = lis_strm_name(head) ;
    majmin = lis_maj_min_name(head) ;
    lis_mark_mem(head, name, MEM_STRMHD) ;	/* plug in name */
    head->sd_open_flags = f->f_flags ;
    err = lis_qattach(head,head->sd_strtab, q, &ndev, f->f_flags, &creds);
    if (err < 0)				/* failed queue attach */
	goto error_rtn ;			/* go cleanup */

    if (ndev != odev)				/* clone open */
    {						/* changed the device */
	stdata_t	*other_hd ;
	const char	*clone_name = name ;
	const char	*clone_majmin = majmin ;

	head->sd_dev = ndev ;
	head->sd_strtab = LIS_DEVST(STR_MAJOR(ndev)).f_str;
	name = lis_strm_name(head) ;		/* update names */
	majmin = lis_maj_min_name(head) ;
	lis_mark_mem(head, name, MEM_STRMHD) ;	/* plug in name */

	other_hd = lis_lookup_stdata(ndev, head) ;
						/* chk for other strm */
	if (other_hd != NULL)
	{					/* use it instead of head */
	    printk("lis_stropen: clone open stream to '%s%s' same as '%s%s'\n",
		    name, lis_maj_min_name(head),
		    lis_strm_name(other_hd), lis_maj_min_name(other_hd)
		    ) ;
	}
	else
	{
	    struct inode	*rtni ;

						/* allocate new inode */
	    rtni = lis_pick_inode(f, i, NULL, ndev) ;	/* callout routine */
	    if (rtni == NULL)
	    {
		err = -ENOMEM ;
		goto error_rtn ;
	    }

	    i = rtni ;				/* use new inode */
#ifndef LINUX					/* linux kernel */
	    STR_FILE_INODE(f) = i ;		/* hook file to new inode */
#endif
	}

	if ( LIS_DEBUG_OPEN )
	{
	    printk("lis_stropen: open clone '%s%s' to '%s%s'\n",
		    clone_name, clone_majmin,
		    name, majmin
		    ) ;
	    if (LIS_DEBUG_ADDRS)
#ifdef KERNEL_2_1				/* linux >= 2.1.0 */
	      printk(
		"lis_stropen: clone: f=%lx i=%lx i->i_rdev=%x\n",
			   (long) f, (long) i, i->i_rdev) ;
#else						/* linux < 2.1.0 or portable */
	      printk(
		"lis_stropen: clone: f=%lx f->f_inode=%lx i=%lx i->i_rdev=%x\n",
			   (long) f, (long) f->f_inode, (long) i, i->i_rdev) ;
#endif
	}
    }						/* ndev != odev */

#if 0
    for (a=0; a<MAX_APUSH && (st=lis_apushm(a,i->i_rdev))!=NULL; a++ )
    {
	if ( (err = push_mod(head,st,f->f_flags,&creds)) < 0 )
	    goto error_rtn ;			/* go cleanup */
    }
#else
    { /* Do autopush */
	const char *mods[MAXAPUSH];
	int j, cnt;

	cnt = lis_apushm(i->i_rdev, mods);
	for (j = 0; j < cnt; ++j) {
	    modID_t mod = lis_findmod(mods[j]);
	    streamtab_t *st;

printk("autopushing %s\n", mods[j]);
	    if (mod == LIS_NULL_MID) {
		/*
		 *  Module not present. Probably one of:
		 *  a) Module open for a previous autopush slept, giving
		 *     kerneld(8) or root time to unload our loadable module.
		 *  b) A configured loadable module could not be loaded.
		 */
		err = -ENOSR; /* Assume (a), retry may succeed */
		goto error_rtn;
	    }

	    st = lis_fmod_sw[mod].f_str;
	    ASSERT(st != NULL);
	    
	    if ( (err = push_mod(head, st, f->f_flags, &creds)) < 0 )
	        goto error_rtn; /* push failure */
	}
    }
#endif

successful_rtn:					/* returning success */

#ifdef S_ZERO_WR				/* Linux >= 2.0.36 */
    i->i_flags |= S_ZERO_WR ;
#endif

    if ( LIS_DEBUG_OPEN )
    {
	printk("lis_stropen: "
	       "open number %d on stream to '%s%s' (%ld open streams)\n",
		head->sd_refcnt, name, majmin, lis_stdata_cnt) ;
	printk("             head-q:\"%s\" driver-q:\"%s\"\n",
		lis_queue_name(head->sd_wq),
		lis_queue_name(head->sd_wq->q_next)) ;
	if (LIS_DEBUG_ADDRS) PRINT_ADDRS ;
    }

    lis_unlock_wopen(head);
    CLOCKOFF(OPENTIME) ;
    return(0);

error_rtn:				/* come here if fail */
					/* no need to unlock the head */
    if ( LIS_DEBUG_OPEN )
    {
	printk("lis_stropen: "
	       "open stream to \"%s\" returning error %d (%ld open streams)\n",
		name, err, lis_stdata_cnt) ;
	if (LIS_DEBUG_ADDRS) PRINT_ADDRS ;
    }

    if (q != NULL) lis_freeq(q);
    if (head != NULL)			/* head was allocated */
    {
	if (hd_locked)			/* head is locked */
	    lis_unlock_wopen(head);	/* unlock it */

	if (   --head->sd_refcnt <= 0	/* last use of head */
	    && INO_STR(i) == head	/* inode points to head */
	   )
        {
#ifdef S_ZERO_WR				/* Linux >= 2.0.36 */
	    i->i_flags &= ~S_ZERO_WR ;
#endif
	    free_stdata(head) ;	/* free & clear i_str ptr */
	    INO_STR(i)=NULL;
        }

	lis_dec_mod_cnt() ;
	LisDownCount(MODUSE) ;
    }

    CLOCKOFF(OPENTIME) ;
    return(err) ;

#undef	PRINT_ADDRS

}/*lis_stropen*/

/*  -------------------------------------------------------------------  */
/* strwrite - write to a stream (see Magic Garden 7.9.5)
 */
#if defined(KERNEL_2_1)
ssize_t
lis_strwrite(struct file *fp, const char *ubuff, size_t ulen, loff_t *op)
#else
int
lis_strwrite(struct inode *i, struct file *fp, const char *ubuff,  int ulen)
#endif
{
    stdata_t *hd;
    mblk_t   *held;
    const char  *name = "" ;
    int		 chunk;
    int		 written=0; 
    int		 flags ;
    int err,newmsg;
    unsigned long  time_cell = 0 ;

    /* Just in case we're holding messages in the write side we use
       head->sd_wmsg msg to hold data to be sent. When it's time to send it we
       just lis_putnext() such message.  If we're not holding then we just
       lis_putnext() the message w/ the usr data.       */

#if defined(KERNEL_2_1)
    (void) op ;			/* compiler happiness */
#else
    CHECK_INO(i,"lis_strwrite"); /* may return */
#endif

    CLOCKON() ;

#if defined(KERNEL_2_1)
    hd = INO_STR(fp->f_dentry->d_inode) ;
#else
    hd = INO_STR(i) ;
#endif
    if (hd == NULL)
    {
	CLOCKOFF(WRITETIME) ;
	return(-ENODEV);
    }

    if (LIS_DEBUG_WRITE) name = lis_strm_name(hd) ;

    if (!stravail(fp))
    {
	if (LIS_DEBUG_WRITE)
	  printk(
	    "strwrite: stream %s%s: out of streams resources\n",
			name, lis_maj_min_name(hd)) ;

	CLOCKOFF(WRITETIME) ;
	return(-ENOSR); 
    }

    if (ulen < hd->sd_minpsz)
    {
	if (LIS_DEBUG_WRITE)
	  printk(
	    "strwrite: stream %s%s: write length %d less than min size of %ld\n",
			name, lis_maj_min_name(hd),
			ulen, hd->sd_minpsz) ;

	CLOCKOFF(WRITETIME) ;
	return(-EINVAL);
    }

    if (!ulen && !F_ISSET(hd->sd_wropt,SNDZERO)) /* we don't send anything */
    {
	if (LIS_DEBUG_WRITE)
	  printk(
	    "strwrite: stream %s%s: write length zero and SNDZERO not set\n",
			name, lis_maj_min_name(hd)) ;

	CLOCKOFF(WRITETIME) ;
	return(0);
    }

    if ((err=lis_check_umem(fp,VERIFY_READ,ubuff,ulen))<0)
    {
	if (LIS_DEBUG_WRITE)
	  printk(
	    "strwrite: stream %s%s: user address 0x%lx cannot be read from\n",
			name, lis_maj_min_name(hd), (long)ubuff) ;

	CLOCKOFF(WRITETIME) ;
	return(err);
    }

    if (F_ISSET(hd->sd_flag,STWRERR))
    {
	if (LIS_DEBUG_WRITE)
	  printk(
	    "strwrite: stream %s%s: stream has been I_LINKed\n",
		name, lis_maj_min_name(hd)) ;
	CLOCKOFF(WRITETIME) ;
	return(-EINVAL) ;
    }

    do {
	if (F_ISSET(hd->sd_flag,STWRERR))
	{
	    if (LIS_DEBUG_WRITE)
	      printk(
		"strwrite: stream %s%s: stream has received M_ERROR %d\n",
			    name, lis_maj_min_name(hd), hd->sd_werror) ;

	    CLOCKOFF(WRITETIME) ;
	    return(-hd->sd_werror);
	}

	if (F_ISSET(hd->sd_flag,STRHUP))
	{
	    if (LIS_DEBUG_WRITE)
	      printk(
		"strwrite: stream %s%s: stream has received M_HANGUP\n",
			    name, lis_maj_min_name(hd)) ;
	    CLOCKOFF(WRITETIME) ;
	    return(-ENODEV);
	}

	/* write packets until all data has been written. */
	if (!lis_bcanputnext(hd->sd_wq,0))
	{
	    if (should_notblock_wr(hd,fp))
	    {
		if (LIS_DEBUG_WRITE)
		    printk(
		     "strwrite: stream %s%s: flw cntrl blocked, return EAGAIN\n",
			name, lis_maj_min_name(hd)) ;
		CLOCKOFF(WRITETIME) ;
		return(-EAGAIN);
	    }
	    else {
		F_SET(hd->sd_flag,WSLEEP);
		if (LIS_DEBUG_WRITE)
		    printk(
		     "strwrite: stream %s%s: flw cntrl blocked, sleep\n",
			name, lis_maj_min_name(hd)) ;
		CLOCKADD() ;
		err = lis_wait_on_wwrite(hd);
		CLOCKON() ;
		if (err < 0)
		{
		    F_CLR(hd->sd_flag,WSLEEP);
		    CLOCKOFF(WRITETIME) ;
		    return(err) ;
		}

		continue ;
	    }
	}
	/* get mblk if there's not one */
	held=hd->sd_wmsg;
	chunk=min((int)hd->sd_maxpsz,ulen-written);
	newmsg= (held==NULL);
	if ( held == NULL && 
	     (held=hd->sd_wmsg=allocb(chunk+hd->sd_wroff,BPRI_LO))==NULL)
	{
	    lis_runqueues();
	    CLOCKOFF(WRITETIME) ;
	    return(written?written:-ENOMEM);
	}
	/* now copy from usr to mblk (type M_DATA by default) */
	if (newmsg)
	{
	    held->b_wptr=held->b_datap->db_base + hd->sd_wroff;
	    if (F_ISSET(hd->sd_flag,STRDELIM))		/* delimited msg */
		held->b_flag |= MSGDELIM ;
	}

	if (held->b_datap->db_lim - held->b_wptr >= chunk)    /* enough room */
	{
	    if ((err = lis_copyin(fp,held->b_wptr,ubuff,chunk)) < 0)
	    {
		CLOCKOFF(WRITETIME) ;
		return(err) ;
	    }

	    ubuff        += chunk;
	    written      += chunk;
	    held->b_wptr += chunk;

	    if (strholding(hd) && 
		held->b_datap->db_lim - held->b_wptr > written)
	    {					/* hold for more data */
		if (LIS_DEBUG_WRITE)
		    printk(
		     "strwrite: stream %s%s: hold message %d bytes for scan timer\n",
			name, lis_maj_min_name(hd), chunk) ;

		if (newmsg) 
		    start_scanl_timer(hd); 

		lis_runqueues();
		CLOCKOFF(WRITETIME) ;
		return(written);
	    }
	}        

        /*
         * Either there was not enough room in the existing message
         * to hold the next write, or we wrote to a new message
         * and the write-holding mechanism is not active.
         *
         * In either case, now is the time to send the message downstream.
         */
        SPLSTR(flags) ;
        if (hd->sd_scantimer != 0)		/* scantimer running */
        {
	    untimeout(hd->sd_scantimer) ;	/* cancel the timer */
	    hd->sd_scantimer = 0 ;		/* note timer is not running */
        }
        SPLX(flags) ;

	if (LIS_DEBUG_WRITE)
	    printk(
		"strwrite: stream %s%s: send %d bytes downstream to %s\n",
		name, lis_maj_min_name(hd), chunk,
		lis_queue_name(hd->sd_wq->q_next)) ;

	CLOCKADD() ;				/* exclude driver time */
        PUTNEXT(hd->sd_wq,hd->sd_wmsg);
	CLOCKON() ;
        hd->sd_wmsg = NULL ;			/* don't reuse the message */

    } while (written<ulen);


    lis_runqueues();
    CLOCKOFF(WRITETIME) ;
    return(written);

}/*lis_strwrite*/

/*  -------------------------------------------------------------------  */
/* lis_strread - read from a stream head
 */
#if defined(KERNEL_2_1)
ssize_t 
lis_strread(struct file *fp, char *ubuff, size_t ulen, loff_t *op)
#else
int 
lis_strread(struct inode *i, struct file *fp, char *ubuff,  int ulen)
#endif
{
    stdata_t *hd;
    queue_t  *hd_rq;
    mblk_t   *mp;
    const char     *name = "" ;
    int		 count=0;
    unsigned int chunk;
    int		 nbytes ;
    int		 msg_lgth ;
    int		 msg_marked ;
    int		 msgs_read = 0 ;
    int		 mread_sent ;
    int err;
    int		 psw = 0 ;
    unsigned long  time_cell = 0 ;

#define	RTN(val)	{ err = val; goto return_point; }


#if defined(KERNEL_2_1)
    (void) op ;			/* compiler happiness */
#else
    CHECK_INO(i,"lis_strrread");
#endif

    CLOCKON() ;

#if defined(KERNEL_2_1)
    hd = INO_STR(fp->f_dentry->d_inode) ;
#else
    hd = INO_STR(i) ;
#endif
    if (hd == NULL)
	RTN(-ENODEV);

    if (LIS_DEBUG_READ) name = lis_strm_name(hd) ;

    if ((err=lis_check_umem(fp,VERIFY_WRITE,ubuff,ulen))<0)
    {
	if (LIS_DEBUG_READ)
	  printk(
	    "strread: stream %s%s: user buffer at 0x%lx cannot be written into\n",
		name, lis_maj_min_name(hd), (long)ubuff) ;

	RTN(err);
    }

    if (F_ISSET(hd->sd_flag,STPLEX))
    {
	if (LIS_DEBUG_READ)
	  printk(
	    "strread: stream %s%s: stream has been I_LINKed\n",
		name, lis_maj_min_name(hd)) ;

	RTN(-EINVAL) ;
    }

    hd_rq=LIS_RD(hd->sd_wq);
    if (!LIS_CHECK_Q_MAGIC(hd_rq)) RTN(-EINVAL) ;

    for (;/*ever*/;)
    {
	mread_sent = 0 ;
	SPLSTR(psw) ;		/* exclude strrput called from interrupt */
	while (hd_rq->q_first == NULL && msgs_read == 0)
	{					/* nothing to read */
	    if (F_ISSET(hd->sd_flag,STRDERR))
	    {
		SPLX(psw) ;
		if (LIS_DEBUG_READ)
		  printk(
		    "strread: stream %s%s: stream has received M_ERROR %d\n",
			name, lis_maj_min_name(hd), hd->sd_rerror) ;
		RTN(-hd->sd_rerror);
	    }
	    
	    if (F_ISSET(hd->sd_flag,STRHUP))
	    {
		SPLX(psw) ;
		if (LIS_DEBUG_READ)
		  printk(
		    "strread: stream %s%s: stream has received M_HANGUP\n",
			name, lis_maj_min_name(hd)) ;

		RTN(0);
	    }

	    if (!mread_sent && F_ISSET(hd->sd_flag,SNDMREAD))
	    {				/* only send once per user call */
		SPLX(psw) ;
		if (LIS_DEBUG_READ)
		    printk(
			"strread: stream %s%s: send M_READ downstream to %s\n",
			name, lis_maj_min_name(hd),
			lis_queue_name(hd->sd_wq->q_next)) ;

		err = snd_mread(hd->sd_wq,ulen);
		if (err < 0)		/* probably ENOMEM */
		    RTN(err) ;

		mread_sent = 1 ;
		SPLSTR(psw) ;		/* re-evaluate w/intrs off */
		continue ;		/* go around again */
	    }

	    if (should_notblock_rd(hd,fp))
	    {
		SPLX(psw) ;
		if (LIS_DEBUG_READ)
		    printk(
			"strread: stream %s%s: O_NONBLOCK set, return EAGAIN\n",
			name, lis_maj_min_name(hd)) ;

		RTN(-EAGAIN);
	    }

	    F_SET(hd->sd_flag,RSLEEP);		/* strrput will wake us up */
	    SPLX(psw) ;				/* intrs OK now */

	    if (LIS_DEBUG_READ)
		printk(
		    "strread: stream %s%s: O_NONBLOCK not set, sleep\n",
		    name, lis_maj_min_name(hd)) ;
	    CLOCKADD() ;
	    err = lis_sleep_on_wread(hd);
	    CLOCKON() ;
	    if (err < 0)
	    {
		F_CLR(hd->sd_flag,RSLEEP);	/* strrput did not run */
		RTN(err) ;
	    }

	    SPLSTR(psw) ;		/* re-evaluate w/intrs off */
	}

	SPLX(psw) ;			/* intrs ok now */

	mp = lis_getq(hd_rq);
	if (mp == NULL)				/* out of data to read */
	{
	    RTN(count);				/* rtn amount read */
	}

	LisDownCount(MSGQDSTRHD) ;
	msg_marked = mp->b_flag & MSGMARK;
	switch(lis_btype(mp)){
	case M_DATA:
	    msg_lgth = 0 ;
	    err = 0 ;
	    msgs_read++ ;
	    do
	    {
		/*
		 * If rptr > wptr the message block is ill-formed
		 * and nbytes will be negative.
		 */
		if (lis_btype(mp) != M_DATA)
		    nbytes = -1 ;		/* pretend invalid */
		else
		    nbytes = mp->b_wptr - mp->b_rptr;

		if (nbytes > 0)			/* safety check */
		{
		    chunk=min(ulen-count,nbytes);
		    if ((err = lis_copyout(fp,mp->b_rptr,ubuff,chunk)) < 0)
			break ;			/* stop reading */

		    msg_lgth   += nbytes ;	/* keep track of msg lgth */
		    mp->b_rptr += chunk;	/* message read ptr */
		    ubuff      += chunk ;	/* user buffer ptr */
		    count      += chunk;	/* bytes copied to user */
		    nbytes     -= chunk ;	/* bytes left in msg */
		}

		if (nbytes <= 0)		/* time for nxt blk */
		{				/* includes bad blocks */
		    mblk_t *dust=mp;		/* and zero-lgth msgs */
		    mp=lis_unlinkb(mp);
		    lis_freeb(dust);
		}

	    } while(count < ulen && mp);

	    if (mp != NULL)			/* not at msg boundary */
	    {
		if ( err == 0 && (hd->sd_rdopt & RMODEMASK) == RMSGD )
		    lis_freemsg(mp);		/* discard rest of msg */
		else				/* RMSGN or RNORM */
		{
		    lis_putbq(hd_rq,mp); 	/* save rest of msg */
		    LisUpCount(MSGQDSTRHD) ;
		}
	    }

	    if (LIS_DEBUG_READ)
	      printk(
	      "strread: stream %s%s: RDOPT=0x%x user buffer %d bytes, "
	      "read %d error %d\n",
		   name, lis_maj_min_name(hd), hd->sd_rdopt, ulen, count, err) ;

	    if (   ulen == count		/* filled user buffer */
		|| (hd->sd_rdopt & RMODEMASK) != RNORM
		|| msg_lgth == 0		/* 0-lgth msg */
		|| msg_marked			/* at marked msg */
	       )				/* return to user */
	    {
	        RTN(count);
	    }

	    if (err < 0)			/* something went wrong */
		RTN(err) ;

	    break;
	case M_PCPROTO:
	    F_CLR(hd->sd_flag,STRPRI);
	    /*fall thru...*/
	case M_PROTO: 
	    msgs_read++ ;
	    if (LIS_DEBUG_READ)
		printk(
		    "strread: stream %s%s: RDOPT=0x%x",
		    name, lis_maj_min_name(hd), hd->sd_rdopt) ;

	    if (F_ISSET(hd->sd_rdopt,RPROTNORM))
	    {					/* return with EBADMSG */
		lis_freemsg(mp);		/* discard the messasge */
		if (LIS_DEBUG_READ)
		    printk(" discard M_PROTO msg\n") ;
		RTN(-EBADMSG);			/* return */
	    }

	    if (F_ISSET(hd->sd_rdopt,RPROTDAT))/* treat M_PROTO as M_DATA */
	    {
		if (LIS_DEBUG_READ)
		    printk(" change M_PROTO msg to M_DATA\n") ;
		lis_btype(mp) = M_DATA;		/* change to M_DATA */
		lis_putbq(hd_rq, mp);		/* put back in queue */
		LisUpCount(MSGQDSTRHD) ;
		break ;				/* go around again */
	    }

	    if (F_ISSET(hd->sd_rdopt,RPROTDIS))/* discard M_PROTO */
	    {					/* deliver M_DATA */
		mblk_t	*proto = mp;		/* the M_PROTO */

		if (LIS_DEBUG_READ)
		    printk(" discard M_PROTO block, retain M_DATA\n") ;
		mp = lis_unlinkb(mp);		/* mp is M_DATA or NULL */
		lis_freeb(proto);		/* discard proto */
		if (mp != NULL)			/* still have a msg */
		{
		    lis_putbq(hd_rq, mp);	/* put back in queue */
		    LisUpCount(MSGQDSTRHD) ;
		}
	    }					/* go around again */
	    else
	    {
		if (LIS_DEBUG_READ)
		    printk(" invalid rdopt, discard M_PROTO msg\n") ;

		lis_freemsg(mp);		/* discard the messasge */
		RTN(-EBADMSG);			/* return */
	    }

	    break;

	case M_READ:		/* add to-be-discarded msgs here */
	    if (LIS_DEBUG_READ)
		printk(
		    "strread: stream %s%s: read %s: discard\n",
		    name, lis_maj_min_name(hd), lis_msg_type_name(mp)) ;
	    lis_freemsg(mp);
	    break;

	case M_SIG:
	    {
		int	signo = *mp->b_rptr ;

		if (LIS_DEBUG_READ)
		    printk(
			"strread: stream %s%s: read M_SIG, signo=%d\n",
			name, lis_maj_min_name(hd), signo) ;
		signalusr(signo,hd);
		lis_freemsg(mp);
		RTN(-EAGAIN);			/* return from system call */
	    }
	    break;

	default:
	    if (LIS_DEBUG_READ)
		printk(
		    "strread: stream %s%s: read %s: return EBADMSG\n",
		    name, lis_maj_min_name(hd), lis_msg_type_name(mp)) ;
	    lis_freemsg(mp);
	    RTN(-EBADMSG);
	}/*sw db_type*/
    }

return_point:
    lis_runqueues();				/* run the streams queues */
    CLOCKOFF(READTIME) ;
    return(err) ;				/* return to user */

#undef	RTN

}/*lis_strrread*/

/*  -------------------------------------------------------------------  */
/* strputmsg - stream head handler for putmsg
 */
int
lis_strputpmsg(struct inode *i, struct file *fp, 
	       void *ctlp, void *datp, int band, int flags)
{
    int		    err;
    int		    couldput ;
    mblk_t	   *msg;
    stdata_t	   *hd;
    const char     *name = "" ;
    strbuf_t	   *ctl = (strbuf_t *) ctlp ;
    strbuf_t	   *dat = (strbuf_t *) datp ;
    unsigned long  time_cell = 0 ;

#if !defined(KERNEL_2_1)
    CHECK_INO(i,"lis_strputmsg"); /* may return */
#endif

    CLOCKON() ;

    hd = INO_STR(i) ;
    if (hd == NULL)
    {
	CLOCKOFF(PUTMSGTIME) ;
	return(-ENODEV);
    }

    if (LIS_DEBUG_PUTMSG) name = lis_strm_name(hd) ;

    if (!stravail(fp))
    {
	if (LIS_DEBUG_PUTMSG)
	  printk(
	    "strputmsg: stream %s%s: out of streams resources\n",
			name, lis_maj_min_name(hd)) ;

	CLOCKOFF(PUTMSGTIME) ;
	return(-ENOSR); 
    }

    if (F_ISSET(hd->sd_flag,STWRERR))
    {
	if (LIS_DEBUG_PUTMSG)
	  printk(
	    "strputmsg: stream %s%s: stream has been I_LINKed\n",
		name, lis_maj_min_name(hd)) ;

	CLOCKOFF(PUTMSGTIME) ;
	return(-EINVAL) ;
    }

    if (band >= 0)				/* putpmsg */
    {
	if ( (flags & (MSG_HIPRI|MSG_BAND)) == (MSG_HIPRI|MSG_BAND) )
	{
	    CLOCKOFF(PUTMSGTIME) ;
	    return(-EINVAL) ;			/* flags are exclusive */
	}
	if ( flags & ~(MSG_HIPRI|MSG_BAND) )
	{
	    CLOCKOFF(PUTMSGTIME) ;
	    return(-EINVAL) ;			/* other flags set */
	}
	if (flags == 0)
	{
	    CLOCKOFF(PUTMSGTIME) ;
	    return(-EINVAL) ;			/* flags must be non-zero */
	}
	if (F_ISSET(flags,MSG_HIPRI))		/* usr wants hi priority msg */
	    band = 0 ;				/* must be band zero */
    }
    else					/* putmsg */
    {
	if (flags & ~RS_HIPRI)
	{
	    CLOCKOFF(PUTMSGTIME) ;
	    return(-EINVAL) ;			/* other flags set */
	}
    }

    if ( (err=copyin_msg(fp,hd,&msg,ctl,dat))<0)	/* allocates msg */
    {
	CLOCKOFF(PUTMSGTIME) ;
	return(err);
    }

    if (msg == NULL)			/* no message to send */
    {
	if (band >= 0)			/* putpmsg */
	{
	    if (flags & ~MSG_BAND)	/* asking for hi-pri */
	    {
		CLOCKOFF(PUTMSGTIME) ;
		return(-EINVAL) ;	/* no can do */
	    }
	}
	else				/* putmsg */
	{
	    if (flags & RS_HIPRI)	/* asking for hi-pri */
	    {
		CLOCKOFF(PUTMSGTIME) ;
		return(-EINVAL) ;	/* no can do */
	    }
	}

	CLOCKOFF(PUTMSGTIME) ;
	return(0) ;			/* no options, no error */
    }

    if (band >= 0)			/* putpmsg */
    {
	if (F_ISSET(flags,MSG_HIPRI))		/* usr wants hi priority msg */
	{
	    if (lis_btype(msg) == M_PROTO)	/* is there a ctl part */
		lis_btype(msg)=M_PCPROTO;	/* yes */
	    else
	    {
		lis_freemsg(msg);
		CLOCKOFF(PUTMSGTIME) ;
		return(-EINVAL) ;		/* no, sorry */
	    }
	}
	else					/* regular priority msg */
	if (F_ISSET(flags,MSG_BAND))		/* non-zero band */
	{
	    if (band >=0 && band <=LIS_MAX_BAND)	/* range check */
		lis_bband(msg) = (unsigned char) band;
	    else
	    {
		lis_freemsg(msg);
		CLOCKOFF(PUTMSGTIME) ;
		return(-EINVAL);
	    }
	}
    }
    else					/* putmsg */
    {
	if (F_ISSET(flags,RS_HIPRI))		/* usr wants hi priority msg */
	{
	    if (lis_btype(msg) == M_PROTO)	/* is there a ctl part */
		lis_btype(msg)=M_PCPROTO;	/* yes */
	    else
	    {
		lis_freemsg(msg);
		CLOCKOFF(PUTMSGTIME) ;
		return(-EINVAL) ;		/* no, sorry */
	    }
	}
    }

    /*
     * wait for flow control 
     *
     * If we are sending a high priority message (M_PCPROTO) ignore flow
     * control and send the messasge downstream unconditionally.
     *
     * Always check for error or hangup.
     */
    do
    {
	if (F_ISSET(hd->sd_flag,STWRERR))
	{
	    if (LIS_DEBUG_PUTMSG)
	      printk(
		"strputmsg: stream %s%s: stream has received M_ERROR %d\n",
			    name, lis_maj_min_name(hd), hd->sd_werror) ;

	    freemsg(msg) ;
	    CLOCKOFF(PUTMSGTIME) ;
	    return(-hd->sd_werror);
	}

	if (F_ISSET(hd->sd_flag,STRHUP))
	{
	    if (LIS_DEBUG_PUTMSG)
	      printk(
		"strputmsg: stream %s%s: stream has received M_HANGUP\n",
			    name, lis_maj_min_name(hd)) ;
	    freemsg(msg) ;
	    CLOCKOFF(PUTMSGTIME) ;
	    return(-ENODEV);
	}

	if (lis_btype(msg) == M_PCPROTO)	/* high priority msg */
	{
	    couldput = 1 ;			/* pretend flw ctl enabled */
	    break ;
	}

	couldput = lis_bcanputnext(hd->sd_wq,
				   (unsigned char) (band < 0 ? 0 : band));
						    /* check the band */
	if (   !couldput			/* flow control blocked */
	    && should_notblock_wr(hd,fp)	/* NDELAY or NONBLOCK */
	   )
	{
	    if (LIS_DEBUG_PUTMSG)
		printk(
		 "strputmsg: stream %s%s: flw cntrl blocked, return EAGAIN\n",
		    name, lis_maj_min_name(hd)) ;
	    freemsg(msg) ;

	    CLOCKOFF(PUTMSGTIME) ;
	    return(-EAGAIN);
	}

	if (!couldput)				/* flw cntrl blocked */
	{
	    if (LIS_DEBUG_PUTMSG)
		printk(
		 "strputmsg: stream %s%s: flw cntrl blocked, sleep\n",
		    name, lis_maj_min_name(hd)) ;

	    F_SET(hd->sd_flag,WSLEEP);		/* wait until unblocked */
	    CLOCKADD() ;
	    err = lis_wait_on_wwrite(hd);
	    CLOCKON() ;
	    if (err < 0)
	    {
		F_CLR(hd->sd_flag,WSLEEP);
		lis_freemsg(msg) ;
		CLOCKOFF(PUTMSGTIME) ;
		return(err) ;
	    }
	}

    } while (!couldput) ;
	    
    if (LIS_DEBUG_PUTMSG)
    {
	printk(
	    "strputmsg: stream %s%s: send %d/%d ctl/data bytes downstream to %s\n",
	    name, lis_maj_min_name(hd),
	    msg->b_datap->db_type == M_DATA ? 0 : lis_xmsgsize(msg),
	    lis_msgdsize(msg),
	    lis_queue_name(hd->sd_wq->q_next)) ;

	if (LIS_DEBUG_ADDRS)
	    printk("strputmsg: q = 0x%lx, msg = 0x%lx\n",
		    (long)hd->sd_wq, (long)msg);
    }

    CLOCKOFF(PUTMSGTIME) ;
    PUTNEXT(hd->sd_wq,msg);			/* send downstream */

    return(0);

}/*lis_strputpmsg*/

/*  -------------------------------------------------------------------  */
/* strgetmsg - stream head handler for getmsg
 *
 * If 'doit' is set then this is a getmsg or getpmsg system call.
 * In this case 'ctl', 'dat' and 'flagsp' are all user space addresses.
 * 'bandp' can be NULL if the call is from getmsg().  It will be
 * a non-null user-space address if the call is from getpmsg().
 *
 * If 'doit' is not set then this is an I_PEEK call.  In this case
 * 'ctl', 'dat' and 'flagsp' and 'bandp' point to kernel memory.
 *
 * Note that getmsg() and I_PEEK use one encoding of the flags
 * (the RS_...) and that getpmsg() uses another (the MSG_...).
 */
int
lis_strgetpmsg(struct inode *i, struct file *fp, 
	       void *ctlp, void *datp, int *bandp, int *flagsp,
	       int doit)
{
    strbuf_t	*ctl = (strbuf_t *) ctlp ;
    strbuf_t	*dat = (strbuf_t *) datp ;
    const char	*ctl_type = "" ;
    const char	*dat_type = "" ;
    int		 err= 0 ;
    int		 rtn = 0 ;
    stdata_t	*hd;
    strbuf_t	 kctl,kdat;
    mblk_t	*mp;
    queue_t	*rdq ;
    int		 band = 0 ;
    int		 flags = 0 ;
    int		 mtype ;
    int		 psw = 0 ;
    unsigned long  time_cell = 0 ;

#define RTN(v)	{ CLOCKOFF(GETMSGTIME) ; return(v); }

#if !defined(KERNEL_2_1)
    CHECK_INO(i,"lis_strgetmsg"); /* may return */
#endif

    CLOCKON() ;

    hd = INO_STR(i) ;
    if (hd == NULL)
	RTN(-ENODEV)

    if (!stravail(fp))
	RTN(-ENOSR)		 /* no more streams resources */
    else if (F_ISSET(hd->sd_flag,STRDERR))
	RTN(-hd->sd_rerror)
    else if (doit && ctl != NULL &&
            (err=lis_check_umem(fp,VERIFY_READ,ctl,sizeof(strbuf_t)))<0)
       RTN(err)
    else if (doit && dat != NULL &&
            (err=lis_check_umem(fp,VERIFY_READ,dat,sizeof(strbuf_t)))<0)
       RTN(err)
    else if (flagsp == NULL)
	RTN(-EFAULT)
    else if (doit && (err=lis_check_umem(fp,VERIFY_READ,flagsp,sizeof(int)))<0)
	RTN(err)
    else if (doit && bandp != NULL
	     && (err=lis_check_umem(fp,VERIFY_READ,bandp,sizeof(int)))<0)
	RTN(err);

    if (F_ISSET(hd->sd_flag,STPLEX))
    {
	if (LIS_DEBUG_GETMSG)
	  printk(
	    "strgetmsg: stream %s%s: stream has been I_LINKed\n",
		lis_strm_name(hd), lis_maj_min_name(hd)) ;

	RTN(-EINVAL) ;
    }

    rdq = LIS_RD(hd->sd_wq) ;
    if (!LIS_CHECK_Q_MAGIC(rdq)) RTN(-EINVAL) ;

    if (doit)
    {					/* fetch structs from usr space */
	if (ctl != NULL)
	{
	    if ((err = lis_copyin(fp,&kctl,ctl,sizeof(strbuf_t))) < 0)
		RTN(err) ;
	}
	else
	{
	    kctl.maxlen = -1 ;
	    kctl.len    = -1 ;
	    kctl.buf    = NULL ;
	}

	if (dat != NULL)
	{
	    if ((err = lis_copyin(fp,&kdat,dat,sizeof(strbuf_t))) < 0)
		RTN(err) ;
	}
	else
	{
	    kdat.maxlen = -1 ;
	    kdat.len    = -1 ;
	    kdat.buf    = NULL ;
	}

	if ((err = lis_copyin(fp,&flags,flagsp,sizeof(flags))) < 0)
	    RTN(err) ;
	if (doit && bandp != NULL)
	{
	    if ((err = lis_copyin(fp,&band,bandp,sizeof(band))) < 0)
		RTN(err) ;
	}
	else
	    band  = 0 ;

	if (kctl.maxlen >= 0 || kdat.maxlen >= 0)
	{
	    SPLSTR(psw) ;	/* exclude strrput called from interrupt */
	    while (   rdq->q_first == NULL
	           && (hd->sd_flag & (STRHUP | STRDERR)) == 0
	          )
	    {
		if (should_notblock_rd(hd,fp))
		{
		    SPLX(psw) ;
		    RTN(-EAGAIN);
		}

		F_SET(hd->sd_flag,RSLEEP);	/* strrput will wake us up */
		SPLX(psw) ;			/* now the flag is set */
		CLOCKADD() ;
		err = lis_sleep_on_wread(hd);
		CLOCKON() ;
		if (err < 0) 
		{
		    F_CLR(hd->sd_flag,RSLEEP);
		    RTN(err) ;
		}

		SPLSTR(psw) ;			/* intrs off to iterate */
	    }

	    SPLX(psw) ;
	}
    }
    else				/* fetch structs from kernel space */
    {
	kctl  = *ctl ;
	kdat  = *dat ;
	flags = *flagsp ;
	band  = *bandp ;
    }

    /*
     * Disable interrupts.  Every exit from here on should go out
     * via the "return_point" label for cleanup purposes.
     *
     * Remember, interrupt routines could putnext() messages up
     * to the stream head read queue and change our idea of the
     * head message in the queue while we are looking at it.
     * We must keep interrupts disabled at least until we remove
     * the message from the queue.
     *
     * If we are peeking we do not need the interrupt exclusion
     * since once we have the mp that is all we need.
     */
    if (doit) SPLSTR(psw) ;

    if ((mp = rdq->q_first) == NULL)		/* mp is 1st msg */
    {						/* empty read queue */
	if (F_ISSET(hd->sd_flag,STRHUP))
	{
	    kctl.len=0;			      /* so sez STREAMS pgrmr's guide */
	    kdat.len=0;
	    rtn = 0 ;
	}
	else
	if (F_ISSET(hd->sd_flag,STRDERR))
	{
	    kctl.len = -1;			/* just a guess */
	    kdat.len = -1;
	    err = -hd->sd_rerror;
	}
	else					/* return NULL msg action */
	    rtn=copyout_msg(fp,NULL,&kctl,&kdat,
			    (strbuf_t*)ctl,(strbuf_t*)dat,doit) ;

	if (doit) SPLX(psw) ;
	goto return_point ;			/* go return 'rtn' */
    }

    if (LIS_DEBUG_GETMSG && LIS_DEBUG_DMP_MBLK && LIS_DEBUG_ADDRS)
    {
	printk("lis_strgetpmsg: q=%lx, mp=%lx, mp->b_rptr=%lx\n",
				(long)rdq, (long)mp, (long)mp->b_rptr);
    }

    /*
     * At this point we have a message.  Ensure that it is
     * of the type that we want to process.
     */
    mtype = lis_btype(mp) ;
    if (LIS_DEBUG_GETMSG)
    {
	if (mtype != M_DATA)
	{
	    ctl_type = lis_msg_type_name(mp) ;
	    dat_type = lis_msg_type_name(mp->b_cont) ;
	}
	else
	{
	    ctl_type = "NULL-MSG" ;
	    dat_type = lis_msg_type_name(mp) ;
	}
    }

    switch (mtype)
    {
    case M_DATA:
    case M_PROTO:
    case M_PCPROTO:
    case M_SIG:
	break ;				/* OK, keep going */

    case M_PASSFP:
    default:
	if (doit) SPLX(psw) ;
	err = -EBADMSG ;		/* unsure of rtn value here */
	goto return_point ;
    }

    if (doit)					/* it's get[p]msg() */
    {
	if (lis_hipri(mtype))			/* M_PCPROTO */
	{
	    F_CLR(hd->sd_flag,STRPRI);		/* about to remove the msg */
	    flags = MSG_HIPRI ;
	    band  = 0 ;
	}
	else					/* normal priority msg */
	{
	    if (flags == MSG_HIPRI)		/* user wanted hi-pri only */
	    {
		SPLX(psw) ;
		err = -EBADMSG;
		goto return_point ;		/* go return 'rtn' */
	    }

	    if (flags == MSG_BAND)		/* wants from particular band */
	    {
		if (mp->b_band < (unsigned char) band)	/* msg in wrong band */
		{
		    SPLX(psw) ;
		    err = -EBADMSG;
		    goto return_point ;		/* go return 'rtn' */
		}
	    }

	    if (flags != 0 && flags != MSG_ANY)
	    {
		SPLX(psw) ;
		err = -EINVAL;			/* invalid flags */
		goto return_point ;		/* go return 'rtn' */
	    }

	    flags = MSG_BAND ;
	    band  = mp->b_band ;
	}

	/*
	 * Now we commit to removing the message from the read queue
	 */
	mp = lis_getq(rdq);			/* remove from queue */

	if ( mp == NULL )
	{
	    printk("lis_strgetpmsg: empty queue: addr of queue 0x%lx\n",
		    (long) rdq);
	    err = -EINVAL;
	    SPLX(psw) ;				/* OK to enable now */
	    goto return_point;
	}

	LisDownCount(MSGQDSTRHD) ;
	if (LIS_DEBUG_GETMSG && LIS_DEBUG_DMP_DBLK)
	{
	    lis_print_msg(mp, "lis_strgetpmsg", PRINT_DATA_RDWR) ;
	}

	SPLX(psw) ;				/* OK to enable now */

	if (mtype == M_SIG)			/* process M_SIG */
	{
	    signalusr(*mp->b_rptr,hd);		/* signal nr is a byte */
	    lis_freemsg(mp);
	    err = -EAGAIN ;
	    goto return_point ;
	}
    }
    else					/* it's I_PEEK */
    {
	if (mtype == M_SIG)			/* not suitable msg */
	{
	    err = -EBADMSG ;
	    goto return_point ;
	}

	if (lis_hipri(mtype))			/* M_PCPROTO */
	    flags = MSG_HIPRI ;
	else					/* M_DATA or M_PROTO */
	if (flags == RS_HIPRI)			/* wants hi-priority only */
	{
	    err = -EBADMSG;
	    goto return_point ;			/* go return 'rtn' */
	}
	else
	    flags = MSG_BAND ;			/* regular priority */
    }

    /* 
     * copyout_msg will copy data to usr returning MORECTL,MOREDATA status
     */
    if ((rtn=copyout_msg(fp,mp,&kctl,&kdat,(strbuf_t*)ctl,(strbuf_t*)dat,doit))<0)
    {
	if (doit)
	    lis_freemsg(mp);
    } 
    else 
    if (doit)				/* getmsg/getpmsg */
    {
	if (rtn)			/* there's still ctl or data in msg */
	{
	  lis_putbq(rdq, mp);
	  LisUpCount(MSGQDSTRHD) ;
	}
	else
	  lis_freemsg(mp);
    }
    else				/* I_PEEK */
	rtn = kctl.len >= 0 || kdat.len >= 0 ;	/* got something */


return_point:

    if (err == 0 && rtn >= 0)			/* return flags to user */
    {
	if (!doit || bandp == NULL)	/* I_PEEK or getmsg */
	{				/* different encoding of flags */
	    switch (flags)
	    {
	    case MSG_HIPRI:
		flags = RS_HIPRI ;
		break ;
	    case MSG_BAND:
	    case MSG_ANY:
	    case 0:
	    default:
		flags = 0 ;
		break ;
	    }
	}

	if (!doit)
	    *flagsp = flags ;		/* return flags to kernal space */
	else				/* return flags to user space */
	if ((err=lis_check_umem(fp,VERIFY_WRITE,flagsp,sizeof(int))) == 0)
	    err = lis_copyout(fp,&flags,flagsp,sizeof(flags));
    }

    if (err == 0 && rtn >= 0 && doit && bandp != NULL)		/* getpmsg() */
    {
	if ((err=lis_check_umem(fp,VERIFY_WRITE,bandp,sizeof(int))) == 0)
	    err = lis_copyout(fp,&band,bandp,sizeof(band));
    }
    else				/* getmsg or I_PEEK */
    if (!doit && bandp != NULL)		/* kernel pointer */
	*bandp = band ;

    if (LIS_DEBUG_GETMSG)
    {
      printk("strgetmsg:    maxlen     len    return=0x%x flags=0x%x band=%d",
					  err ? err : rtn,
							  flags,     band) ;
      printk(" stream %s%s\n", lis_strm_name(hd), lis_maj_min_name(hd)) ;
      printk("  control:    %6d  %6d    %s\n",
		  kctl.maxlen, kctl.len, ctl_type) ;
      printk("     data:    %6d  %6d    %s\n",
		  kdat.maxlen, kdat.len, dat_type) ;
    }

    lis_runqueues();
    CLOCKOFF(GETMSGTIME) ;
    return(err != 0 ? err : rtn);

#undef RTN

}/*lis_strgetpmsg*/


/*  -------------------------------------------------------------------  */
/* strioctl - handle IOCTLs at the stream head
 *
 * NOTE: This functions needs pretty much cleanup. Perhaps ioctling stuff
 * should be moved to strioc.c (or alike). THe lis_runqueues() calls should
 * be relocated, i.e. place them only when you think there could be
 * things to be [STREAMS] scheduled.
 */
int
lis_strioctl(struct inode *i, struct file *f, unsigned int cmd, unsigned long arg)
{
    stdata_t *hd;
    int err=0,flush_pri=0;
    cred_t creds;
    streamtab_t *st;
    const char        *name = "(unknown)" ;
    const char	      *lis_maj_min = "" ;
    unsigned long  time_cell = 0 ;
    pid_t pid=(pid_t)PID(f);

#define RTN(v)	{ CLOCKOFF(IOCTLTIME) ; return(v); }

#if !defined(KERNEL_2_1)
    CHECK_INO(i,"lis_strioctl"); /* may return */
#endif

    CLOCKON() ;

    if (LIS_DEBUG_IOCTL && LIS_DEBUG_ADDRS)
#ifdef KERNEL_2_1				/* linux >= 2.1.0 */
	printk("lis_strioctl: f=%lx i=%lx i->i_rdev=%x\n",
	       (long) f, (long) i, i->i_rdev) ;
#else						/* linux < 2.1.0 or portable */
	printk("lis_strioctl: f=%lx f->f_inode=%lx i=%lx i->i_rdev=%x\n",
	       (long) f, (long) f->f_inode, (long) i, i->i_rdev) ;
#endif

    hd = INO_STR(i) ;
    if (hd == NULL)
	RTN(-ENODEV);
    if (!stravail(f))
	RTN(-ENOSR); 
    creds.cr_uid  = (uid_t) EUID(f);
    creds.cr_gid  = (gid_t) EGID(f);
    creds.cr_ruid = (uid_t) UID(f);
    creds.cr_rgid = (gid_t) GID(f);
    if ( LIS_DEBUG_IOCTL )
    {
	name = lis_strm_name(hd) ;
	lis_maj_min = lis_maj_min_name(hd) ;
    }

    if (F_ISSET(hd->sd_flag,STPLEX))
    {
	if (LIS_DEBUG_IOCTL)
	  printk(
	    "strioctl: stream %s%s: stream has been I_LINKed\n",
		name, lis_maj_min) ;

	RTN(-EINVAL) ;
    }


    switch( cmd ){		/* the often an ioc is called the 
				 * closer it should be to the begin of the sw*/
    case I_PUSH:		/* tested */
	if (F_ISSET(hd->sd_flag,STRHUP))
	    RTN(-ENXIO);
#ifdef FIFO_IMPL
	if (LIS_ISDEV(STR_MAJOR(i->i_rdev),LIS_FIFO)) /* not ok for a fifo  */
	    RTN(-ENXIO);
#endif
	if ((err=lis_check_umem(f,VERIFY_READ,(char*)arg,FMNAMESZ+1))<0)
	    RTN(err)
	else {
	    char *mname;
	    if((err=lis_copyin_str(f,(char*)arg,&mname,FMNAMESZ+1))<0)
		RTN(err);

	    if ( LIS_DEBUG_IOCTL )
		printk("strioctl: I_PUSH \"%s\" below %s%s: ",
			mname, name, lis_maj_min) ;

	    if ((st = lis_modstr(lis_findmod(mname))) == NULL)
	    {
		err=-EINVAL;
		if ( LIS_DEBUG_IOCTL )
		    printk("module not found\n") ;

		FREE((caddr_t)mname);
	    }
	    else
	    {
		FREE((caddr_t)mname);
		err = lis_lock_wopen(hd);
		if (err < 0) break ;

		err=push_mod(hd,st,f->f_flags,&creds);
		if ( LIS_DEBUG_IOCTL )
		{
		    if (err)	printk("error %d\n", err) ;
		    else	printk("OK\n") ;
		}

		if (F_ISSET(hd->sd_flag,STRISTTY)){
		    /* 
		       we should make this the controling terminal...
		     */
		    hd->sd_pgrp = PGRP(f);
		}
		lis_unlock_wopen(hd);
	    }
	}
	break;
    case I_POP:		/* tested */
      {
	const char	*qname = "";

	if (F_ISSET(hd->sd_flag,STRHUP))
	    RTN(-ENXIO);
	err = lis_lock_wopen(hd);
	if (err < 0) break ;

	if ( LIS_DEBUG_IOCTL )
	    qname = lis_queue_name(hd->sd_wq->q_next) ;

	err=pop_mod(hd,f->f_flags, &creds); 

	if ( LIS_DEBUG_IOCTL )
	{
	    printk("strioctl: I_POP \"%s\" from stream %s%s: ",
		    qname, name, lis_maj_min) ;

	    if (err)	printk("error %d\n", err) ;
	    else	printk("OK\n") ;
	}

	lis_unlock_wopen(hd);
      }
      break;
    case I_LOOK:		/* tested */
	err = lis_lock_wopen(hd);
	if (err < 0)
	    RTN(err) ;
	if ((err=lis_check_umem(f,VERIFY_WRITE,(char*)arg,FMNAMESZ+1))<0)
	    RTN(err);
	if (!hd->sd_pushcnt)
	    RTN(-EINVAL);
	if ( LIS_DEBUG_IOCTL )
	    printk("strioctl: I_LOOK: top driver/module of stream %s%s is \"%s\"\n",
		    name, lis_maj_min, lis_queue_name(hd->sd_wq->q_next)) ;

	name = lis_queue_name(hd->sd_wq->q_next) ;
	err = lis_copyout(f,name, (char*)arg, strlen(name)+1);
	lis_unlock_wopen(hd);
	break;
    case I_FLUSHBAND:
	{
	    struct bandinfo bi;
	    if (   (err=lis_check_umem(f,VERIFY_READ,
				    (char*)arg,sizeof(bandinfo_t)))<0
		|| (err=lis_copyin(f,&bi,(char*)arg,sizeof(bandinfo_t))) < 0
	       )
		RTN(err);
	    arg= bi.bi_flag|FLUSHBAND;
	    flush_pri=bi.bi_pri;
	}
        /*fall thru...*/
    case I_FLUSH:
	if (F_ISSET(hd->sd_flag,STRHUP))
	    RTN(-ENXIO);
	if ( LIS_DEBUG_IOCTL )
	    printk("strioctl: I_FLUSH arg=0x%lx\n", arg) ;

	switch(arg & FLUSHRW){
	case FLUSHR:
#ifdef FIFO_IMPL
	    if (LIS_ISDEV(STR_MAJOR(i->i_rdev),LIS_FIFO)){
	      /*flush rw/other-wr for fifo;*/
	    }
	    else
#endif
	    {
	      int res = lis_snd_mflush(LIS_WR(hd->sd_wq),FLUSHR,flush_pri);
	      lis_runqueues();
	      RTN(res);
	    }
	    break;
	case FLUSHW:
#ifdef FIFO_IMPL
	    if (LIS_ISDEV(STR_MAJOR(i->i_rdev),LIS_FIFO)){
		/*flush rw/other-wr for fifo;*/
	    }
	    else
#endif
	    {
	      int res=lis_snd_mflush(LIS_WR(hd->sd_wq),FLUSHW,flush_pri);
	      lis_runqueues();
	      RTN(res);
	    }
	case FLUSHRW:
#ifdef FIFO_IMPL
	    if (LIS_ISDEV(STR_MAJOR(i->i_rdev),LIS_FIFO)){
		/*flush rd/other-rd;*/
	    }
	    else
#endif
	    {
		int res=lis_snd_mflush(LIS_WR(hd->sd_wq),FLUSHRW,flush_pri);
		lis_runqueues();
		RTN(res);
	    }
	default:
	    RTN(-EINVAL);
	}
	break;
    case I_SETSIG:
	if (!arg){
	    if (lis_del_from_elist(&(hd->sd_siglist),pid,S_ALL))
		err = -EINVAL;
	    else 
		if (!hd->sd_siglist)
		    hd->sd_sigflags=0;
	}
	else 
	    if (lis_add_to_elist(&(hd->sd_siglist),pid, (short) arg))
		err = -EAGAIN;
	    else
		F_SET(hd->sd_sigflags,arg);

	if ( LIS_DEBUG_IOCTL || LIS_DEBUG_SIG )
	{
	    printk("strioctl: I_SETSIG: stream %s%s signal %ld: ",
		    name, lis_maj_min, arg) ;
	    if (err)	printk("error %d\n", err) ;
	    else	printk("OK\n") ;
	}

	lis_runqueues();
	break;
    case I_GETSIG:
	{
	    int evs;

	    if ( LIS_DEBUG_IOCTL )
		printk("strioctl: I_GETSIG: stream %s%s: ",
			name, lis_maj_min) ;

	    err=lis_check_umem(f,VERIFY_WRITE,(char*)arg,sizeof(int)) ;
	    if (   err == 0
		&& (evs=(int)lis_get_elist_ent(hd->sd_siglist,pid)) == 0
	       )
		err = -EINVAL;

	    if ( LIS_DEBUG_IOCTL )
	    {
		if (err)	printk("error %d\n", err) ;
		else		printk("signal %d\n", evs) ;
	    }

	    if (err == 0)
		err = lis_copyout(f,&evs,(char*)arg,sizeof(int));
	}
	break;
    case I_FIND:		/* tested */
	{
	    char *mname = NULL;
	    queue_t *q;

	    err = lis_lock_wopen(hd);
	    if (err < 0)
		RTN(err) ;
	    if ((err=lis_check_umem(f,VERIFY_READ,(char*)arg,FMNAMESZ+1))<0)
		RTN(err);
	    if (hd->sd_pushcnt){
		if ((err=lis_copyin_str(f,(char*)arg,&mname,FMNAMESZ+1))<0)
		    RTN(err);
		if ( LIS_DEBUG_IOCTL )
		    printk("strioctl: I_FIND: stream %s%s module \"%s\"\n",
					    name, lis_maj_min, mname);
		for (q=hd->sd_wq; 
		     q && strncmp(lis_queue_name(q),mname,FMNAMESZ);
		     q=q->q_next )
		{
		    if (!LIS_CHECK_Q_MAGIC(q))
		    {
			FREE(mname) ;
			lis_unlock_wopen(hd);
			lis_runqueues();
			RTN(-EINVAL);
		    }
		}
		if (q){		
		    FREE(mname) ;
		    lis_unlock_wopen(hd);
		    lis_runqueues();
		    RTN(1);	/* found! */
		}
	    }
	    lis_unlock_wopen(hd);
	    if (mname == NULL)
	    {
		lis_runqueues();
	        RTN(0);		/* not present */
	    }

	    if (!lis_findmod(mname))
	    {
		FREE(mname) ;
		lis_runqueues();
		RTN(-EINVAL);	/* not a valid module */
	    }

	    FREE(mname) ;
	    lis_runqueues();
	    RTN(0);		/* not present */
	}
    case I_PEEK:
	{
	    strpeek_t pb;
	    strbuf_t *ctl=&(pb.ctlbuf), *dat=&(pb.databuf);
	    int band =0;
	    int res,flags;
	    if (   (err=lis_check_umem(f,VERIFY_READ,
				    (char*)arg,sizeof(strpeek_t)))<0
		|| (err=lis_check_umem(f,VERIFY_WRITE,
				    (char*)arg,sizeof(strpeek_t)))<0
		|| (err=lis_copyin(f,&pb,(char*)arg,sizeof(strpeek_t)))<0
	       )
		RTN(err);
            flags=(int)pb.flags;
	    res=lis_strgetpmsg(i,f,ctl,dat,&band,&flags,0); /*0-> just peek!*/
            pb.flags=(long)flags;
	    if (res >= 0)
		res = lis_copyout(f,&pb,(char*)arg,sizeof(strpeek_t));
	    if ( LIS_DEBUG_IOCTL )
		printk("strioctl: I_PEEK: stream %s%s flags=0x%x rslt=%d\n",
					name, lis_maj_min, flags, res) ;
	    lis_runqueues();
	    RTN(res);
	}
	break;
    case I_SRDOPT:
      {
	int res = set_readopt(&hd->sd_rdopt,arg);
	if ( LIS_DEBUG_IOCTL )
	    printk("strioctl: I_SRDOPT: stream %s%s arg=%ld rslt=%d\n",
					name, lis_maj_min, arg, res) ;
	lis_runqueues();
	RTN(res);
      }
    case I_GRDOPT:
	{
	    int rmode;
	    if ((err=lis_check_umem(f,VERIFY_WRITE,(char*)arg,sizeof(int)))<0)
		RTN(err);
	    rmode=hd->sd_rdopt;
	    if ( LIS_DEBUG_IOCTL )
		printk("strioctl: I_GRDOPT: stream %s%s mode=0x%x\n",
					    name, lis_maj_min, rmode);
	    err = lis_copyout(f,&rmode,(char*)arg,sizeof(int));
	}	
	break;
    case I_NREAD:		/* tested */
	{ 
	    int bytes=0;
	    if ((err=lis_check_umem(f,VERIFY_WRITE,(char*)arg,sizeof(int)))<0)
		RTN(err);
	    if (LIS_RD(hd->sd_wq)->q_first)
		bytes=lis_msgdsize(LIS_RD(hd->sd_wq)->q_first);
	    if ( LIS_DEBUG_IOCTL )
		printk("strioctl: I_NREAD: stream %s%s bytes=%d\n",
				name, lis_maj_min, bytes);
	    err = lis_copyout(f,&bytes,(char*)arg,sizeof(int));
	}
	break;
    case I_FDINSERT:
      {
	/* This code compliments of Eric Levenez <levenez@club-internet.fr> */
	/* see magic garden 7.9.7 */
	strfdinsert_t	 strfdinsert;
	stdata_t	*ihd;
	queue_t		*q;
	mblk_t		*msg;
	int		 couldput;
	strbuf_t	*ctl;
	strbuf_t	*dat;

	/* primary sanity checks */

	if (F_ISSET (hd->sd_flag, STRDERR) || F_ISSET (hd->sd_flag, STWRERR))
	  RTN(-EINVAL);

	if (!hd->sd_wq->q_next)
	  RTN(-ENOSR);

	if (   (err = lis_check_umem (f, VERIFY_READ, (char *) arg,
				   sizeof (strfdinsert_t))) < 0
	    || (err = lis_copyin (f, &strfdinsert, (char *) arg,
				   sizeof (strfdinsert_t))) < 0
	   )
	    RTN(err);

	/* read structure command */


	/* sanity checks */
	if (   (strfdinsert.offset < 0)
	    || (strfdinsert.offset > 4096)
	    || (strfdinsert.ctlbuf.len < strfdinsert.offset + sizeof (char *))
	    || (strfdinsert.offset % sizeof (char *))
	    || (strfdinsert.fildes < 0)
	    || (strfdinsert.flags & ~RS_HIPRI))
	    RTN(-EINVAL);

	if ((ihd = lis_fd2str(strfdinsert.fildes)) == NULL)
	  RTN(-ENOSTR);

	if (   F_ISSET (ihd->sd_flag, STRDERR)
	    || F_ISSET (ihd->sd_flag, STWRERR))
	      RTN(-EINVAL);

	if (F_ISSET (ihd->sd_flag, STPLEX))
	  RTN(-ENXIO);

	if (   (   (strfdinsert.databuf.len > 0)
		&& (strfdinsert.databuf.len < hd->sd_minpsz)
	       )
	    || (   (hd->sd_maxpsz != INFPSZ)
		&& (strfdinsert.databuf.len > hd->sd_maxpsz)
	       )
	   )
	      RTN(-ERANGE);

	/* find driver queue */

	for (q = ihd->sd_wq; q->q_next; q = q->q_next)
	{
	    if (!LIS_CHECK_Q_MAGIC(q)) RTN(-EINVAL) ;
	}

	/* create the message a-la putpmsg */

	ctl = (strbuf_t *) (arg +
			    (unsigned long) &((strfdinsert_t *) NULL)->ctlbuf);
	dat = (strbuf_t *) (arg +
			    (unsigned long) &((strfdinsert_t *) NULL)->databuf);

	if ((err = copyin_msg (f, hd, &msg, ctl, dat)) < 0)
	  RTN(err);

	if (!msg)
	  RTN(-EINVAL);

	/* for priority message */

	if (F_ISSET (strfdinsert.flags, RS_HIPRI))
	{
	  if (lis_btype (msg) == M_PROTO)
	    lis_btype (msg) = M_PCPROTO;
	  else
	  {
	    lis_freemsg (msg);
	    RTN(-EINVAL);
	  }
	}

	/* insert the read queue */

	*((queue_t **) (msg->b_rptr + strfdinsert.offset)) = LIS_RD (q);

	/* wait for flow control (code from lis_strputpmsg) */

	do
	{
	  if (F_ISSET (hd->sd_flag, STWRERR))
	  {
	    freemsg (msg);
	    lis_runqueues();
	    RTN(-EINVAL);
	  }
	  if (F_ISSET (hd->sd_flag, STRHUP))
	  {
	    freemsg (msg);
	    lis_runqueues();
	    RTN(-ENODEV);
	  }
	  couldput = lis_bcanputnext (hd->sd_wq, (unsigned char) 0);
	  if (   !couldput
	      && (   should_notblock_wr (hd, f)
		  || F_ISSET (strfdinsert.flags, MSG_HIPRI)
		 )
	     )
	  {
	    freemsg (msg);
	    lis_runqueues();
	    RTN(-EAGAIN);
	  }

	  if (!couldput)
	  {
	    F_SET (hd->sd_flag, WSLEEP);
	    if ((err = lis_wait_on_wwrite (hd)) < 0)
	    {
	      F_CLR (hd->sd_flag, WSLEEP);
	      lis_freemsg (msg);
	      lis_runqueues();
	      RTN(err);
	    }
	  }
	}
	while (!couldput);

	/* send the message */

	if (LIS_DEBUG_IOCTL)
	  printk ("strioctl: I_FDINSERT: stream %s%s: q=%lx q2 = %lx \n",
		  name, lis_maj_min, (unsigned long) hd->sd_wq, (unsigned
							    long) LIS_RD (q));
	PUTNEXT (hd->sd_wq, msg);
	RTN(0);
      }
      break;

    case I_STR:		/* tested */
	if (F_ISSET(hd->sd_flag,STRHUP))
	    RTN(-ENXIO)
	else {
	    strioctl_t ioc;
	    int res;
	    if (   (err=lis_check_umem(f,VERIFY_READ,
				    (char*)arg,sizeof(strioctl_t)))<0
		|| (err=lis_check_umem(f,VERIFY_WRITE,
				    (char*)arg,sizeof(strioctl_t)))<0
		|| (err=lis_copyin(f,&ioc,(char*)arg,sizeof(strioctl_t)))<0
	       )
		RTN(err);
	    CLOCKADD() ;
	    res=lis_strdoioctl(f,hd,&ioc,&creds,1);
	    CLOCKON() ;
	    if (res == 0)
		res=lis_copyout(f,&ioc,(char*)arg,sizeof(strioctl_t));
	    if ( LIS_DEBUG_IOCTL )
	     printk(
	      "strmioctl: I_STR: stream %s%s ic_cmd=0x%x ic_timout=%d ic_len=%d rslt=%d\n",
			    name, lis_maj_min,
			    ioc.ic_cmd, ioc.ic_timout, ioc.ic_len, res) ;
		
	    lis_runqueues();
	    RTN(res);
	}
	break;
    case I_SWROPT:
	if ( LIS_DEBUG_IOCTL )
	    printk("strioctl: I_SWROPT: stream %s%s arg=0x%lx\n",
					name, lis_maj_min, arg) ;
	hd->sd_wropt = arg ;
	lis_runqueues();
	RTN(0) ;
    case I_GWROPT:
	if ((err=lis_check_umem(f,VERIFY_WRITE,(char*)arg,sizeof(int)))<0)
	    RTN(err);

	if ( LIS_DEBUG_IOCTL )
	    printk("strioctl: I_GWROPT: stream %s%s mode=0x%x\n",
				name, lis_maj_min, hd->sd_wropt);
	err = lis_copyout(f,&hd->sd_wropt,(char*)arg,sizeof(int));
	lis_runqueues();
	RTN(err);
    case I_SENDFD:
	if ( LIS_DEBUG_IOCTL )
	    printk("strioctl: I_SENDFD: stream %s%s: not implemented",
	    				name, lis_maj_min) ;

	RTN(-EINVAL);	/* not implemented */

        /* copy strrecvfd (arg) struct from usr and get fd value; put it in a
	   M_PASSFP msg and give it to the hd->q_next; it's assumed this is a
	   streams pipe --see mg 7.10 */
	
	break;
    case I_RECVFD:
	if ( LIS_DEBUG_IOCTL )
	    printk("strioctl: I_RECVFD: stream %s%s: not implemented",
	    				name, lis_maj_min) ;
	RTN(-EINVAL);	/* not implemented --see I_SENDFD*/
	break;
    case I_LIST:
	if (!arg)		/* want number of modules */
	{
	    if ( LIS_DEBUG_IOCTL )
		printk("strioctl: I_LIST: stream %s%s arg=NULL, return %d\n",
				    name, lis_maj_min, hd->sd_pushcnt+1) ;
	    RTN(hd->sd_pushcnt+1); /* +1 for driver  */
	}
	else {
	    str_list_t list;
	    queue_t   *q;
	    int	       nmods ;
	    if (   (err=lis_check_umem(f,VERIFY_READ,(char*)arg,
						sizeof(str_list_t)))<0
	        || (err=lis_check_umem(f,VERIFY_WRITE,(char*)arg,
						sizeof(str_list_t)))<0
		|| (err=lis_copyin(f,&list,(char*)arg,sizeof(str_list_t)))<0
	       )
		RTN(err);
	    if (list.sl_nmods<=0)
		RTN(-EINVAL);
	    if ((err=lis_check_umem(f,VERIFY_WRITE,
				 list.sl_modlist,
				 sizeof(struct str_mlist)*list.sl_nmods)) <0)
		RTN(err);

	    if (hd->sd_wq == NULL || (q=hd->sd_wq->q_next) == NULL)
	    {
		printk("strioctl: I_LIST: stream %s%s: NULL queue ptrs\n",
					name, lis_maj_min) ;
		RTN(0) ;			/* really a bug */
	    }

	    for (nmods = 0; 
		 q && list.sl_nmods-- ; 
		 q=q->q_next, list.sl_modlist++, nmods++
		)
	    {
		if (!LIS_CHECK_Q_MAGIC(q)) break ;
		if ((err = lis_copyout(f,lis_queue_name(q),
			    list.sl_modlist->l_name,
			    FMNAMESZ)) < 0)
		    RTN(err) ;
	    }
	    if ((err = lis_copyout(f, &nmods,
			    &((str_list_t *)arg)->sl_nmods, sizeof(int))) < 0)
		RTN(err) ;

	    if ( LIS_DEBUG_IOCTL )
		printk("strioctl: I_LIST: stream %s%s arg=non-NULL, "
		       "%d modules, return 0\n", name, lis_maj_min, nmods) ;
	    RTN(0);
	}
	break;
    case I_ATMARK:
	{
	    mblk_t	*mp ;
	    int		 rtn = 0 ;

	    /* Switch on arg and return whether the first msg in the
	     * stream head read queue is marked if arg is ANYMARK.
	     * Return whether is it the last marked message if arg is LASTMARK.
	     */
	    if (   hd->sd_wq == NULL		/* no queue */
		|| LIS_RD(hd->sd_wq) == NULL	/* no rd queue */
		|| (mp = LIS_RD(hd->sd_wq)->q_first) == NULL	/* no msg */
	       )
	    {
		rtn = 0 ;			/* obviously not marked */
	    }
	    else
	    switch (arg)
	    {
	    case ANYMARK:			/* is 1st msg marked?  */
		rtn = (mp->b_flag & MSGMARK) ? 1 : 0 ;
		break ;

	    case LASTMARK:			/* is 1st msg the last marked */
		if ( !(mp->b_flag & MSGMARK) )	/* 1st msg not marked at all */
		{
		    rtn = 0 ;			/* is not the last one, then */
		    break ;
		}

		rtn = 1 ;			/* assume no other msg marked */
		for (mp = mp->b_next; mp != NULL; mp = mp->b_next)
		{
		    if (mp->b_flag & MSGMARK)	/* some other msg is marked */
		    {
			rtn = 0 ;		/* condition not met */
			break ;
		    }
		}

		break ;

	    default:
		rtn = -EINVAL;			/* invalid argument */
	    }

	    if ( LIS_DEBUG_IOCTL )
		printk("strioctl: I_ATMARK: stream %s%s: arg=%ld, return %d\n",
				    name, lis_maj_min, arg, rtn) ;
	    RTN(rtn) ;
	}
	break;
    case I_CKBAND:
	if ( LIS_DEBUG_IOCTL )
	    printk("strioctl: I_CKBAND: stream %s%s arg=%ld\n",
				    name, lis_maj_min, arg) ;
	if ( arg > LIS_MAX_BAND )
	    RTN(-EINVAL)
	else {
	    long val = 0;
	    
	    err = lis_strqget(LIS_RD(hd->sd_wq),QCOUNT,(char)arg, &val) ; 
	    if (err) 
		val = -err ;			/* failed */
	    else
		val = (val != 0) ? 1 : -1 ;	/* AT&T STREAMS pgrmr's guide */

	    lis_runqueues();
	    RTN(val);
	}
	break;
    case I_GETBAND:
	{
	    int res;
	    if ((err=lis_check_umem(f,VERIFY_WRITE,(char*)arg,sizeof(int)))<0)
		RTN(err);
	    if (!LIS_RD(hd->sd_wq)->q_first)
		RTN(-ENODATA);
	    res=LIS_RD(hd->sd_wq)->q_first->b_band;
	    if ( LIS_DEBUG_IOCTL )
		printk("strioctl: I_GETBAND: stream %s%s rslt=%d\n",
					name, lis_maj_min, res) ;
	    err = lis_copyout(f,&res,(char*)arg,sizeof(int));
	    lis_runqueues();
	    RTN(err);
	}
	break;
    case I_CANPUT:
        {
	    int	res ;
	    lis_runqueues();
	    if ( arg > LIS_MAX_BAND )
		RTN(-EINVAL)
	    else 
	    {
		res = lis_bcanputnext(hd->sd_wq, (unsigned char) arg);
		if ( LIS_DEBUG_IOCTL )
		    printk("strioctl: I_CANPUT: stream %s%s rslt=%d\n",
				    name, lis_maj_min, res) ;
		lis_runqueues();
		RTN(res) ;
	    }
        }
	break;
    case I_SETCLTIME:
	if (arg >LIS_MAX_CLTIME)
	    RTN(-EINVAL);
	hd->sd_closetime=arg;
	if ( LIS_DEBUG_IOCTL )
	    printk("strioctl: I_SETCLTIME: stream %s%s arg=%ld\n",
				    name, lis_maj_min, arg) ;
	RTN(0);
	break;
    case I_GETCLTIME:
	{
	    int res;
	    if ((err=lis_check_umem(f,VERIFY_WRITE,(char*)arg,sizeof(int)))<0)
		RTN(err);
	    res=hd->sd_closetime;
	    if ( LIS_DEBUG_IOCTL )
		printk("strioctl: I_GETCLTIME: stream %s%s rslt=%d\n",
					name, lis_maj_min, res) ;
	    err = lis_copyout(f,&res,(char*)arg,sizeof(int));
	    RTN(err);
	}
	break;
    case I_PLINK:		
	/*fall through...  */
    case I_LINK:		/* link arg(=fd) stream below this one */
				/* see magic garden 7.5.1              */
				/* also AT&T STREAMS Pgrmr's Guide Chapt 10 */
	if (F_ISSET(hd->sd_flag,STRDERR)||
	    F_ISSET(hd->sd_flag,STWRERR))
	    RTN(-EINVAL);
	if (F_ISSET(hd->sd_flag,STRHUP))
	    RTN(-ENXIO);

	if (   hd->sd_strtab->st_muxrinit == NULL
	    || hd->sd_strtab->st_muxwinit == NULL
	   )					/* must have lower qinits */
	    RTN(-EINVAL);

	CLOCKADD() ;
	err = lis_i_link(i, f, hd, (int) arg, cmd, &creds) ;
	CLOCKON() ;
	lis_runqueues();
	RTN(err) ;
	
    case I_PUNLINK:
	/*fall through...  */
    case I_UNLINK:
	if (F_ISSET(hd->sd_flag,STRHUP))
	    RTN(-ENXIO);
	if ((long) arg < -1)		/* compare as signed quantity */
	    RTN(-EINVAL);

	err = lis_lock_wopen(hd);
	if (err >= 0)
	{
	    CLOCKADD() ;
	    err = lis_i_unlink(i, f, hd, (int)arg, cmd, &creds ) ;
	    CLOCKON() ;
	    lis_unlock_wopen(hd);
	}

	lis_runqueues();
	RTN(err) ;

    case I_LIS_GETSTATS:		/* LiS only - return stats */
	err = lis_check_umem(f,VERIFY_WRITE,(char*)arg, sizeof(lis_strstats)) ; 
	if (err < 0)
	    RTN(err);

	err = lis_copyout(f,lis_strstats,(char*)arg,sizeof(lis_strstats));
	RTN(err);

    case I_LIS_PRNTSTRM:		/* LiS only - print stream */
	lis_print_stream(hd) ;
	break;

    case I_LIS_PRNTQUEUES:		/* LiS only - print queues */
	lis_print_queues() ;
	break;

    case I_LIS_PRNTMEM:			/* LiS only - print memory */
	lis_print_mem() ;		/* maybe an suser() check */
	break;

    case I_LIS_PRNTSPL:
	lis_print_spl_track() ;
	break ;

    case I_LIS_SDBGMSK:			/* LiS only - set debug mask */
        CLOCKOFF(IOCTLTIME) ;		/* may reset timing debug bit */
	lis_debug_mask = arg ;		/* maybe an suser() check */
        CLOCKON() ;
	break;

    case I_LIS_GET_MAXMSGMEM:		/* LiS only - get max msg mem */
	err = lis_check_umem(f,VERIFY_WRITE,(char*)arg, sizeof(long)) ; 
	if (err < 0)
	    RTN(err);

	err=lis_copyout(f,&lis_max_msg_mem,(char*)arg,sizeof(lis_max_msg_mem));
	break;

    case I_LIS_SET_MAXMSGMEM:		/* LiS only - set max msg mem */
	if (!lis_suser(f))
	    RTN(-EINVAL) ;

	lis_max_msg_mem = arg ;
	lis_strstats[MSGMEMLIM][MAXIMUM] = lis_max_msg_mem ;
	break;

    case I_LIS_GET_MAXMEM:		/* LiS only - get max mem */
	err = lis_check_umem(f,VERIFY_WRITE,(char*)arg, sizeof(long)) ; 
	if (err < 0)
	    RTN(err);

	err = lis_copyout(f,&lis_max_mem,(char*)arg,sizeof(lis_max_mem));
	break;

    case I_LIS_SET_MAXMEM:		/* LiS only - set max mem */
	if (!lis_suser(f))
	    RTN(-EINVAL) ;

	lis_max_mem = arg ;
	lis_strstats[MEMLIM][MAXIMUM] = lis_max_mem ;
	break;


    default:
	/* encapsulate ioctl(fd,cmd,arg)as if the user specified 
	 * instead     ioctl(fd,I_STR,{cmd,arg})
	 * and pass down to strdoioctl()
	 */
	{
	    strioctl_t ioc;
	    int res;
	    ioc.ic_cmd=cmd;
	    ioc.ic_timout=LIS_LNTIME;
	    ioc.ic_len=TRANSPARENT;
	    ioc.ic_dp=(char*)arg;
	    CLOCKADD() ;
	    res=lis_strdoioctl(f,hd,&ioc,&creds,0);	/* no copyin */
	    CLOCKON() ;
	    if ( LIS_DEBUG_IOCTL )
		printk("strioctl: stream %s%s ic_cmd=0x%x ic_timout=%d ic_len=0 rslt=%d\n",
			    name, lis_maj_min,
			    ioc.ic_cmd, ioc.ic_timout, res) ;
	    lis_runqueues();
	    RTN(res);
	}
    }/*sw cmd*/
    lis_runqueues();
    CLOCKOFF(IOCTLTIME) ;
    return(err);

#undef RTN

}/*lis_strioctl*/

/*  -------------------------------------------------------------------  */
/* lis_poll_bits - evaluate poll bits for a stream			 */
/*
 * Used by lis_strpoll and by lis_poll_2_1 to evaluate a stream and
 * return poll events appropriate to the stream.
 */
unsigned lis_poll_bits(stdata_t *hd)
{
    int			 revents   = 0 ;
    int			 mtype ;
    mblk_t		*mp ;

#define	RTN(msk)	{ revents = msk; goto rtn_point; }

    if ( F_ISSET(hd->sd_flag,(STRDERR|STWRERR)) )
	RTN(POLLERR) ;

    if (F_ISSET(hd->sd_flag,(STRHUP|STRCLOSE)))
	RTN(POLLHUP) ;

    if (F_ISSET(hd->sd_flag,STPLEX))
    {
	if (LIS_DEBUG_POLL)
	  printk(
	    "strpoll: stream %s%s: stream has been I_LINKed\n",
		lis_strm_name(hd), lis_maj_min_name(hd)) ;

	RTN(POLLNVAL) ;
    }

    if ((mp = LIS_RD(hd->sd_wq)->q_first) != NULL)	/* mp is 1st msg */
    {						/* a msg can be read */
	mtype = lis_btype(mp) ;
	if (lis_hipri(mtype))
		revents |= POLLPRI ;
	else
	{
	    revents |= POLLIN ;			/* other than hi pri msg */
	    switch (mtype)
	    {
	    case M_DATA:
	    case M_PROTO:
		if (mp->b_band > 0)
		    revents |= POLLRDBAND ;
		else
		    revents |= POLLRDNORM ;
		break ;

	    case M_SIG:
		if (*mp->b_rptr == SIGPOLL)
		    revents |= POLLMSG ;
		break ;
	    }
	}
    }

    if (lis_bcanputnext(hd->sd_wq, 0))
	revents |= POLLWRNORM ;			/* can write band 0 data */

    if (lis_bcanputnext_anyband(hd->sd_wq->q_next))
	revents |= POLLWRBAND ;			/* can write band > 0 */

rtn_point:
    if ( LIS_DEBUG_POLL )
	printk("lis_poll_bits: stream %s%s: revents:%s\n",
		lis_strm_name(hd), lis_maj_min_name(hd),
	        lis_poll_events((short)revents)) ;

    return(revents) ;				/* return events */

#undef RTN

} /* lis_poll_bits */

/*  -------------------------------------------------------------------  */
/* strpoll - called from lis_syspoll() poll() syscall handler
 *
 * The STREAMS poll function is called with the file info and
 * a pointer to a polldat structure.  The STREAMS routine
 * interrogates the file for the requested conditions according
 * to the 'pd_events' field in the polldat structure.  It returns
 * a bit-mask of the events that satisfy the conditions, if any.
 * The STREAMS routine also sets the pd_headp pointer to point to
 * its list head in its stream structure for the polldat list.
 * The caller takes care of linking the structure into the list.
 */
#ifdef PORTABLE_POLL
int
lis_strpoll(struct inode *i, struct file *f, void *ptr)
{
    polldat_t		*pdat_ptr  = (polldat_t *) ptr ;
    int			 revents   = 0 ;
    int			 events    = pdat_ptr->pd_events ;
    stdata_t		*hd;
    unsigned long  	 time_cell = 0 ;

    (void) f ;				/* reference 'f' for compiler */

    if (   !(hd = INO_STR(i))
	|| hd->magic != STDATA_MAGIC
       )
	return(POLLNVAL);

    CLOCKON() ;

    revents = lis_poll_bits(hd) & events ;	/* evaluate the stream */

    pdat_ptr->pd_headp = &hd->sd_polllist ;	/* ptr to list hd struct */
						/* allows insertion */
    if ( LIS_DEBUG_POLL )
	printk("strpoll: stream %s%s: events:%s revents:%s\n",
		lis_strm_name(hd), lis_maj_min_name(hd),
	      lis_poll_events((short)events), lis_poll_events((short)revents)) ;

    CLOCKOFF(POLLTIME) ;

    return(revents) ;				/* return events */

#undef RTN

} /* lis_strpoll */
#endif

/*  -------------------------------------------------------------------  */
/*  lis_wakeup_close
 *
 * This routine is used as a timout routine and as a utility function.
 * The argument is really a pointer to a streamhead structure.
 */
void	lis_wakeup_close(caddr_t arg)
{
    stdata_t	*hd = (stdata_t *) arg ;

    if (hd->magic == STDATA_MAGIC)
    {
	untimeout(hd->sd_close_timer) ;		/* cancel timer, just in case */
	lis_up(&hd->sd_closing) ;			/* semaphore wakeup */
    }

} /* lis_wakeup_close */

/*  -------------------------------------------------------------------  */
/* doclose - dec refcnt & dismantle the stream when sd_refcnt gets 0.
 *
 * This routine is called from two different contexts.  One is from
 * lis_strclose which is on the straightforward close path in closing
 * a file.  The other is from the unlink code when a stream is unlinked
 * from below a multiplexor and this is the last use of the stream.  In this
 * latter case, both file and inode pointers will be NULL since the file
 * and inode are both long gone.
 */
static int
lis_doclose(struct inode *i, struct file *f, struct stdata *head, cred_t *creds)
{
    int		   err ;
    int		   refcnt ;
    int		   psw;
    unsigned long  time_cell = 0 ;

    CLOCKON() ;
    err = lis_lock_wopen(head);
    if (err < 0)
	printk("lis_strclose: error %d locking stream head, closing anyway\n",
		err) ;

    lis_del_from_elist(&head->sd_siglist,PID(f),S_ALL);

    if ( LIS_DEBUG_CLOSE )
    {
	printk("lis_strclose: closing stream %s%s refcnt=%d linkcnt=%d\n",
		lis_strm_name(head),
		lis_maj_min_name(head),
		head->sd_refcnt, head->sd_linkcnt ) ;
	printk("              head-q:\"%s\" driver-q:\"%s\"\n",
		lis_queue_name(head->sd_wq),
		lis_queue_name(head->sd_wq->q_next)) ;
    }

    head->sd_refcnt-- ;				/* decr ref count */
    refcnt = head->sd_refcnt ;			/* will return this */
    lis_dec_mod_cnt() ;
    LisDownCount(MODUSE) ;

    /*
     * If the reference count is now down to the plink count
     * then all that is left are the plinks.  This means that
     * this is really the last close of the stream and the inode
     * is about to disappear.  We clear the inode ptr in the
     * stdata structure.  There will be no more accesses of
     * this stream from above via the file system.
     *
     * The stream remains allocated but cannot be reached via an
     * open.  Open will allocate a new stdata structure upon a
     * re-open.
     */
    if (head->sd_linkcnt != 0 && head->sd_refcnt <= head->sd_linkcnt)
    {
	lis_free_elist(&head->sd_siglist);	/* do in signal masks */
    }
    
    if (head->sd_refcnt == 0)			/* dismantle stream */
    {
	head->sd_flag |= STRCLOSE ;
	if (   head->sd_closetime != 0		/* have close time set */
	    && lis_qcountstrm(head->sd_wq) != 0	/* messages queued */
	   )
	{					/* set a timer */
	    if ( LIS_DEBUG_CLOSE )
	      printk(
		"lis_strclose: stream %s%s waiting %dms for queues to drain\n",
			lis_strm_name(head), lis_maj_min_name(head),
			head->sd_closetime);

	    SPLSTR(psw);			/* just in case */
	    head->sd_wq->q_next->q_flag |= QCLOSING ;
	    SPLX(psw) ;
	    head->sd_close_timer = lis_timeout_fcn(lis_wakeup_close,
					   (caddr_t) head,
					   (head->sd_closetime*HZ)/1000,
					   "stream-close", MEM_TIMER) ;
	    lis_runqueues();
	    CLOCKADD() ;
	    lis_down(&head->sd_closing) ;		/* sleep for awhile */
	    CLOCKON() ;
	}

	lis_free_elist(&head->sd_siglist);	/* do in signal masks */
	if (head->sd_mux.mx_hd != NULL)		/* is a ctl stream */
	{
	    CLOCKADD() ;
	    lis_i_unlink(i, f, head, -1,
			 I_PUNLINK, creds) ;	/* unlink all lowers */
	    CLOCKON() ;
	}

	/* if M_PASSFD implemented then 
	      we should close fds; */
	while (!pop_mod(head,head->sd_open_flags,creds))
	    ;

	CLOCKADD() ;
	lis_qdetach(LIS_RD(head->sd_wq->q_next),1,head->sd_open_flags,creds);
	CLOCKON() ;

	lis_dismantle(head, creds);		/* deallocate head */
						/* must be last thing */
    }
    else				/* still has users, unlock it */
	lis_unlock_wopen(head);

    if ( LIS_DEBUG_CLOSE )
	printk("lis_strclose: %ld open streams\n", lis_stdata_cnt) ;

    CLOCKOFF(CLOSETIME) ;
    return(refcnt);
								     
} /*lis_doclose*/

#if defined(KERNEL_2_1)
int lis_strclose(struct inode *i, struct file *f)
#define RTN	return(0)
#else
void lis_strclose(struct inode *i, struct file *f)
#define RTN	return
#endif
{
    struct stdata *head;
    cred_t	   creds;
    unsigned long  time_cell = 0 ;
    
    CLOCKON() ;
    lis_close_cnt++ ;
    if ( LIS_DEBUG_CLOSE )
    {
	printk("lis_strclose: "
	       "call number %ld i_rdev=(%d,%d) i_count=%ld f_count=%d"
#ifdef KERNEL_2_1				/* linux >= 2.1.0 */
	       " d_count=%d\n",
#else
	       "\n",
#endif
	       lis_close_cnt, STR_MAJOR(i->i_rdev), STR_MINOR(i->i_rdev),
	       (long) i->i_count, f->f_count
#ifdef KERNEL_2_1				/* linux >= 2.1.0 */
	       , f->f_dentry->d_count
#endif
	       ) ;

	if (LIS_DEBUG_ADDRS)
#ifdef KERNEL_2_1				/* linux >= 2.1.0 */
	      printk("lis_strclose: f=%lx i=%lx\n",
			   (long) f, (long) i) ;
#else						/* linux < 2.1.0 or portable */
	      printk("lis_strclose: f=%lx f->f_inode=%lx i=%lx\n",
			   (long) f, (long) f->f_inode, (long) i) ;
#endif
    }

    if (!i) {
	lis_error(LIS_ERROR,"lis_strclose","called w/ null inode");
	CLOCKOFF(CLOSETIME) ;
	RTN;
    }

    if (!S_ISCHR(i->i_mode))
    {
	CLOCKOFF(CLOSETIME) ;
	RTN;
    }

    if (!LIS_DEVOK(STR_MAJOR(i->i_rdev)))
    {
	CLOCKOFF(CLOSETIME) ;
	RTN;
    }

    if (!(head=INO_STR(i))){
	lis_error(LIS_ERROR,"lis_strclose",
		  "called with null stream (inode 0x%x)",i);
	CLOCKOFF(CLOSETIME) ;
	RTN;
    }    

    creds.cr_uid  = (uid_t) EUID(f);
    creds.cr_gid  = (gid_t) EGID(f);
    creds.cr_ruid = (uid_t) UID(f);
    creds.cr_rgid = (gid_t) GID(f);

    CLOCKADD() ;
    if (   lis_doclose(i, f, head, &creds) == 0	/* last close, gone */
#if defined(KERNEL_2_1)
	|| (   f->f_dentry->d_count == 1	/* dentry about to be 'put' */
	    && INO_CNT(i) == 1			/* inode about to be 'put' */
	   )
#else
        || INO_CNT(i) == 1			/* inode about to be 'put' */
#endif
       )
    {
#ifdef S_ZERO_WR				/* Linux >= 2.0.36 */
	i->i_flags &= ~S_ZERO_WR ;
#endif
	INO_STR(i) = NULL ;			/* unhook inode ptr */
    }

    RTN ;

} /* lis_strclose */

#undef RTN

/*  -------------------------------------------------------------------  */

/* Initialize some glob vars...
 */
void lis_init_head( void )
{
    /* We'd better use defaults from strconfig instead of magic #s.
     */
    lis_qhead = lis_qtail =  lis_scanqhead= lis_scanqtail = NULL;
    lis_queues_running = 0;
    lis_nstrpush = 0;
    lis_strhold = 0;
    lis_strthresh = 0;
    /*lis_iocseq=whichever get's in the air -- so we have some kind of random
     *number for this seq # <-should be fixed*/

    lis_init_bufcall() ;
    lis_init_mod() ;
}/*lis_init_head*/

/*  -------------------------------------------------------------------  */
/*
 * Terminate code
 */
void lis_terminate_head(void)
{
    lis_terminate_mod() ;
    lis_terminate_msg() ;
    lis_terminate_bufcall() ;
    lis_terminate_dki() ;
    lis_terminate_mem() ;
} /* lis_terminate_head */


/*----------------------------------------------------------------------
# Local Variables:      ***
# change-log-default-name: "~/src/prj/streams/src/NOTES" ***
# End: ***
  ----------------------------------------------------------------------*/
