/*                               -*- Mode: C -*- 
 * linux-mdep.c --- Linux kernel dependent support for LiS.
 * Author          : Francisco J. Ballesteros
 * Created On      : Sat Jun  4 20:56:03 1994
 * Last Modified By: David Grothe
 * RCS Id          : $Id: linux-mdep.c,v 1.8 1996/01/26 16:53:01 dave Exp $
 * Purpose         : provide Linux kernel <-> LiS entry points.
 * ----------------______________________________________________
 *
 *    Copyright (C) 1995  Francisco J. Ballesteros, Denis Froschauer
 *    Copyright (C) 1997  David Grothe, Gcom, Inc <dave@gcom.com>
 *
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 * 
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330, Cambridge,
 * MA 02139, USA.
 * 
 *
 *    You can reach me by email to
 *    nemo@ordago.uc3m.es, 100741.1151@compuserve.com
 *    dave@gcom.com
 */

#ident "@(#) LiS linux-mdep.c 2.6 6/11/99 19:20:16 "

/*  -------------------------------------------------------------------  */
/*				 Dependencies                            */

#include <sys/LiS/linux-mdep.h>
#include <linux/module.h>
#include <linux/version.h>

#ifdef USE_BH
#include <linux/stddef.h>		/* for NULL */
#include <linux/tqueue.h>		/* task queues */
#include <asm/ptrace.h>			/* for pt_regs */
#include <linux/interrupt.h>
#else
#include <linux/sched.h>
#define	__KERNEL_SYSCALLS__	1	/* to make kernel_thread visible */
#include <linux/unistd.h>		/* for kernel_thread */
#endif

#include <asm/signal.h>
#include <asm/io.h>
#include <sys/strport.h>	/* interface */
#include <sys/LiS/errmsg.h>	/* LiS err msg types */
#include <sys/LiS/buffcall.h>	/* bufcalls */
#include <sys/LiS/head.h>	/* stream head */
#include <sys/stream.h>		/* LiS entry points */
#include <sys/cmn_err.h>
#include <sys/poll.h>
#include <linux/fs.h>		/* linux file sys externs */
#include <linux/time.h>
#ifdef VERSION_2
#include <asm/semaphore.h>	/* kernel semaphore routines */
#endif

#if 0 /* OSH: Now using lis_module_info, and a #define in <sys/LiS/mod.h> */
#ifdef KERNEL_2_1		/* kernel version 2.1 or later */
#define module_info	xx_module_info
#endif
#endif

/*  -------------------------------------------------------------------  */

#ifdef KERNEL_2_1
#define	SEM_COUNT(semp)		atomic_read(&((semp)->count))
#else
#define	SEM_COUNT(semp)		((semp)->count)
#endif

/************************************************************************
*                            lis_assert_fail                            *
*************************************************************************
*									*
* This is called when the ASSERT() macro fails.                         *
*									*
* It sends a warning message to the kernel logger, but does nothing     *
* else.                                                        		*
*									*
************************************************************************/
void lis_assert_fail(const char *expr, const char *objname,
		     const char *file, unsigned int line)
{
        printk(KERN_CRIT "%s: assert(%s) failed in file %s, line %u\n",
	       objname, expr, file, line);
        /* We cannot just abort() the kernel here :-( */
}

/************************************************************************
*                            Prototypes                                 *
************************************************************************/
extern void do_gettimeofday(struct timeval *tv) ;	/* kernel fcn */
extern void lis_print_trace_func(char *msg) ;	/* forward decl */
extern void lis_runqueues(void);

/************************************************************************
*                          vremap                                       *
*************************************************************************
*									*
* The 2.1 and 2.2 kernels have de-implemented the function vremap.	*
* They replaced it with ioremap, which does the same thing.  For	*
* backward compatibility we provide a vremap that calls ioremap.	*
*									*
************************************************************************/
#if 0				/* now done via osif.[ch] */
#ifdef KERNEL_2_1		/* kernel version 2.1 or later */

void *vremap(unsigned long offset, unsigned long size)
{
    return(ioremap(offset, size)) ;
}

#endif
#endif

/************************************************************************
*                           lis_phys_to_virt                            *
*************************************************************************
*									*
* This routine maps physical addresses to virtual addresses and is	*
* portable across kernel versions.					*
*									*
* This may not look like much, but the secret is that the routine	*
* phys_to_virt is an inline function that is different in the different	*
* kernel versions.  Because we do this here, driver writers do not	*
* necessarily have to maintain a special version of their drivers	*
* for kernel version 2.2.						*
*									*
************************************************************************/
#if 0				/* now done via osif.[ch] */
void *lis_phys_to_virt(unsigned long address)
{
    return(phys_to_virt(address)) ;
}
#endif

/************************************************************************
*                         SPL Tracking                                  *
*************************************************************************
*									*
* The following structure is used to track SPL usage.			*
*									*
************************************************************************/
typedef struct
{
    int			 type ;
    unsigned long	 tim ;
    unsigned long	 dtim ;
    int			 flags ;
    int			 state ;
    char		*file ;
    int			 line ;

} spl_track_t ;

/*
 * Types
 */
#define	TRACK_SPLSTR	1		/* splstr call */
#define	TRACK_SPLX	2		/* splx call */

#define	SPL_TRACK_SIZE	4096		/* power of 2 */
spl_track_t		lis_spl_track[SPL_TRACK_SIZE] ;
int			lis_spl_track_inx ;

/************************************************************************
*                       Storage Declarations                            *
************************************************************************/

int				lis_runq_open_flag ; /* open has been called */
int				lis_runq_sched ;     /* q's are scheduled */
struct semaphore		lis_runq_sem ;
struct semaphore		lis_runq_kill_sem ;
int				lis_runq_pid ;	/* procss id of kernel thread */
						/* must be declared even if   */
						/* not using kernel thread.   */

extern void			lis_run_queues(void) ;
extern int			lis_queues_running ;
extern int              	lis_putnext_flag ;


/*  -------------------------------------------------------------------  */

/* This should be entry points from the kernel into LiS
 * kernel should be fixed to call them when appropriate.
 */


struct file_operations
lis_streams_fops = {
	NULL,			/* lseek   -- no lseek  */
	lis_strread,		/* read    		*/
	lis_strwrite,		/* write to stream      */
	NULL,			/* readdir -- no readdir*/
#if defined(KERNEL_2_1)
	lis_poll_2_1,		/* poll  		*/
#else
	lis_select,		/* select  		*/
#endif
	lis_strioctl,		/* ioctl   		*/
	NULL,			/* mmap    -- no mmap   */
	lis_stropen,		/* open the stream      */
#if defined(KERNEL_2_1)
	NULL,			/* flush   -- no flush  */
#endif
	lis_strclose,		/* close   		*/
	NULL,			/* fsync   -- no fsync  */
	NULL,			/* fasync  -- no fasync */
	NULL,			/* check_media_change	*/
	NULL			/* revalidate		*/
#if defined(KERNEL_2_1)
       ,NULL			/* lock			*/
#endif
};

/* some kernel memory has been free'd 
 * tell STREAMS
 */
void lis_memfree( void )
{
}/*lis_memfree*/

/* This will copyin usr string pointed by ustr and return the result  in
 * *kstr. It will stop at  '\0' or max bytes copyed in.
 * caller should call FREE(*kstr) on success.
 * Will return 0 or errno
 * STATUS: complete, untested
 */
int 
lis_copyin_str(struct file *f, const char *ustr, char **kstr, int max)
{
	int    error;
#if !defined(KERNEL_2_1)
	char  *tmp, c;
	int    i=max;
#endif
	char  *mem ;

	(void) f ;
	if (max <= 0)
	    return(-ENOMEM);
	error = -EFAULT;
#if !defined(KERNEL_2_1)
	c = get_fs_byte(ustr++);
	if (!c)
	    return(-EINVAL);
#endif

	if((mem = ALLOCF(max,"copyin-buf ")) == NULL)
	    return(-ENOMEM);

#if defined(KERNEL_2_1)
	*kstr = mem ;
	error = strncpy_from_user(mem, ustr, max) ;
	if (error < 0)
	{
	    FREE(mem);
	    return(error) ;
	}

	if (error >= max)
	{
	    FREE(mem);
	    return(-ENAMETOOLONG) ;
	}

	return(0) ;
#else
	*kstr = tmp = mem ;
	while (--i)
	{
	    *(tmp++) = c;
	    c = get_fs_byte(ustr++);
	    if (!c) {
		*tmp = '\0';
		return(0);
	    }
	}
	FREE(mem);
	return(error);
#endif
}/*lis_copyin_str*/


#if 0
/*  -------------------------------------------------------------------  */
/*
 * lis_down
 *
 * This routine is patterned after the Linux __down() function from
 * linux/sched.c.  This version watches for signals and returns
 * < 0 if the process has been signalled.  It returns 0 for a
 * good wakeup.
 */
int     
lis_down(struct semaphore *sem)
{
#ifdef VERSION_2
	return(down_interruptible(sem)) ;
#else
        struct wait_queue wait = { current, NULL };
	int		  rslt = 0 ;

	if (sem->count > 0)		/* available */
	{
	    ((volatile)sem->count)-- ;		/* one less on counter */
	    return(0) ;
	}

        add_wait_queue(&sem->wait, &wait);
        current->state = TASK_INTERRUPTIBLE;
        while (sem->count <= 0) {
                schedule();
		/*
		 * Check for signal
		 */
		if (current->signal & ~current->blocked)
		{
		    rslt = -EINTR ;	/* interrupted system call */
		    break ;
		}

                current->state = TASK_INTERRUPTIBLE;
        }
        current->state = TASK_RUNNING;
        remove_wait_queue(&sem->wait, &wait);
	if (rslt >= 0)				/* if not signalled */
	    ((volatile)sem->count)-- ;		/* one less on counter */

	return(rslt) ;

#endif

} /* lis_down */
#endif


/*  -------------------------------------------------------------------  */
/* get a err msg for error level
 */
char  *lis_errmsg( int lvl )
{
    switch(lvl){
    case LIS_PANIC:
	return(KERN_EMERG);
    case LIS_ERROR:
	return(KERN_ERR);
    case LIS_WARN:
	return(KERN_WARNING);
    case LIS_DEBUG:
	return(KERN_DEBUG);
    default:
	return("unknown severity");
    }
} /*lis_errmsg*/

/*  -------------------------------------------------------------------  */
/*				    Timeouts                             */

void
lis_tmout(struct timer_list *tl, void (*fn)(ulong), long arg, long ticks)
{
    init_timer(tl);
    tl->function= fn;
    tl->data	= arg;
    tl->expires	= jiffies + ticks;
    add_timer(tl);
}

void
lis_untmout( struct timer_list *tl)
{
    del_timer(tl);
}

/************************************************************************
*                            lis_time_till                             *
*************************************************************************
*									*
* Given a target time in terms of elapsed milli-seconds, in other words,*
* the same units as jiffies if jiffies were in milli-seconds, return	*
* the number of milli-seconds that it will take to reach to target time.*
*									*
* The return is negative if the target time is in the past, zero	*
* if it is the same as the current target time, and positive if		*
* target time is in the future.						*
*									*
************************************************************************/
long	lis_time_till(long target_time)
{
    return( target_time - jiffies*(1000/HZ) ) ;

} /* lis_time_till */

/************************************************************************
*                           lis_target_time                             *
*************************************************************************
*									*
* Convert the milli_sec interval to an absolute target time expressed	*
* in milli-seconds.  We compute this relative to the Linux software	*
* clock, jiffies, converted to milliseconds.				*
*									*
************************************************************************/
long	lis_target_time(long milli_sec)
{
    return( jiffies*(1000/HZ) + milli_sec ) ;

} /* lis_target_time */



/*  -------------------------------------------------------------------  */

/* find the head for a file descriptor.
 * STATUS: complete, untested
 */
stdata_t *
lis_fd2str( int fd )
{
#if defined(KERNEL_2_1)			/* kernel 2.1.0 or later */

    struct file		* file;
    stdata_t		* hd ;

    if ( (file = fcheck(fd)) == NULL)
	return(NULL) ;

    hd = INO_STR(file->f_dentry->d_inode) ;
    if (hd != NULL && hd->magic == STDATA_MAGIC)
	return(hd) ;			/* looks like a stdata structure */

    return(NULL) ;			/* not a streams file */

#else					/* kernel 2.0.xx  */
    struct file		* file;
    struct inode	* inode;
    stdata_t		* hd ;

    if (fd>=NR_OPEN || !(file=current->files->fd[fd]) || !(inode=file->f_inode))
	    return(NULL);

    hd = INO_STR(inode);
    if (hd != NULL && hd->magic == STDATA_MAGIC)
	return(hd) ;			/* looks like a stdata structure */

    return(NULL) ;			/* not a streams file */
#endif
}


/************************************************************************
*                          lis_select                                   *
*************************************************************************
*									*
* Called from the Linux sys_select routine to interrogate the status	*
* of a stream.								*
*									*
* Return 1 if the stream is ready for the requested operation.  	*
* Return 0, and set up the wait queue entry, if it is not.		*
*									*
************************************************************************/
#if !defined(LINUX_POLL)
int      lis_select(struct inode *inode, struct file *file,
		    int sel_type, select_table *wait)
{
    stdata_t		*hd;
    polldat_t		 pdat ;
    int			 evts ;
    int			 msk ;
    int			 rtn ;

    memset(&pdat, 0, sizeof(pdat)) ;		/* just in case */
retry:
    switch (sel_type)
    {
    case SEL_IN:
	pdat.pd_events = POLLIN | POLLPRI | POLLRDNORM | POLLRDBAND ;
	msk = POLLNVAL ;
	break ;
    case SEL_OUT:
	pdat.pd_events = POLLOUT | POLLWRNORM | POLLWRBAND ;
	msk = POLLNVAL ;
	break ;
    case SEL_EX:
    default:
	pdat.pd_events = POLLMSG ;			/* SIG-POLL msg */
	msk = POLLERR | POLLHUP | POLLNVAL ;
	break ;
    }

    evts = lis_strpoll(inode, file, &pdat) ;	/* get stream status */
    rtn = ( (evts & (pdat.pd_events | msk)) != 0 ) ;/* expected status */
    if (rtn) return(1) ;			/* yes, can do that */

    /*
     * Requested events are not present, add ourselves to the
     * wait queue for the stream.  Set a flag to be noticed by
     * str_rput.
     *
     * If we get back from select_wait and the flag has been unset then
     * lis_select_wakeup got called before returning to us and possibly
     * before we even got into the wait list.  In that case, we go around
     * again and return the actual events.
     */
    hd = INO_STR(inode) ;
    F_SET(hd->sd_flag,STRSELPND);
    select_wait(&hd->sd_select.sel_wait, wait) ;
    if (!F_ISSET(hd->sd_flag,STRSELPND))	/* lost the race */
	goto retry ;

    return(0) ;

} /* lis_select */
#endif

/************************************************************************
*                        lis_select_wakeup                              *
*************************************************************************
*									*
* Wake up those waiting on this stream.					*
*									*
************************************************************************/
#if !defined(LINUX_POLL)
void lis_select_wakeup(stdata_t *hd)
{
    F_CLR(hd->sd_flag,STRSELPND);	/* no longer waiting on select */
    if (   hd == NULL
	|| hd->magic != STDATA_MAGIC
	|| hd->sd_select.sel_wait == NULL
	|| hd->sd_select.sel_wait == WAIT_QUEUE_HEAD(&hd->sd_select.sel_wait)
       )
	return ;

    wake_up_interruptible(&hd->sd_select.sel_wait) ;

} /* lis_select_wakeup */
#endif

/************************************************************************
*                            lis_pick_inode                             *
*************************************************************************
*									*
* This routine is called from stropen().				*
*									*
* 'old' is the inode passed to stropen for a clone open and 'new' is	*
* NULL.  We need to allocate a new inode, transfer some attributes	*
* into it and return a pointer to it.  We are done with the 'old'	*
* inode since it was the one that pointed to the /dev/clone and		*
* the new one points to the actual stream device opened.		*
*									*
************************************************************************/
struct inode *lis_pick_inode(struct file  *oldf,
			     struct inode *old,
			     struct inode *new,
			     dev_t         dev)
{
    if (new != NULL)			/* calling error */
	printk("lis_pick_inode: parameter 'new' expected to be NULL\n") ;

    if (old != NULL && old->i_rdev == dev)	/* old has same rdev */
	return(old) ;				/* use old one */

    new = get_empty_inode() ;			/* system routine */
    if (new != NULL)
    {
	new->i_rdev = dev ;			/* set desired dev */
	if (old != NULL)			/* new is cloned */
	{					/* new will own the stream */
	    INO_STR(new) = INO_STR(old) ;
	    new->i_mode  = old->i_mode ;	/* inherit mode */
	    new->i_ctime = old->i_ctime;
	    new->i_atime = old->i_atime;
	    new->i_mtime = old->i_mtime;
	}
	else
	    printk("lis_pick_inode: clone requires old inode non-NULL\n") ;
    }

    if (old != NULL)
    {
	INO_STR(old) = NULL ;
#ifdef KERNEL_2_1				/* linux >= 2.1.0 */
	dput(oldf->f_dentry) ;			/* free old inode/dentry */
	oldf->f_dentry = d_alloc_root(new, NULL) ;/* get new dentry */
#else						/* linux <= 2.0.xx */
	iput(old) ;				/* free old inode */
	STR_FILE_INODE(oldf) = new ;		/* hook file to new inode */
#endif
    }

    return(new) ;				/* use new one */

} /* lis_pick_inode */




/*  -------------------------------------------------------------------  */
/*				    Module                               */
#ifdef LINUX

static int lis_major;
char lis_kernel_version[]= UTS_RELEASE;

extern char	lis_version[] ;
extern char	lis_date[] ;

extern void     hook_syscalls(void);
extern void     restore_syscalls(void);


#ifdef LIS_LOADABLE_SUPPORT

/*
 *  lis_loadable_load - load a loadable module.
 *
 *  This function may sleep. On failure it may return a negative errno, or
 *  it may return 0. The module should have been loaded on return from this
 *  function, and the caller should check for this.
 */
int lis_loadable_load(const char *name)
{
	return request_module(name);
}

#endif /* ifdef LIS_LOADABLE_SUPPORT */


int lis_init_module( void )
{
    lis_print_trace = lis_print_trace_func ;
    lis_major = register_chrdev(0,"streams",&lis_streams_fops);
    if	(lis_major < 0)
    {
	printk("Unable to register Streams Subsystem\n");
	return -EIO;
    }
    /* Initialize every global variable to a default value, if there're
     * modules w/ not-exported globals we should create init_() functions for
     * them and call them from here. */

    lis_init_head();

    hook_syscalls();
    /* lis_runqueues(); */		/* gets kernel thread created */

    printk ("Linux Streams Subsystem ready. Major device number %d.\n"
            "Version %s %s. Compiled for kernel version %s.\n",
	     lis_major, lis_version, lis_date, lis_kernel_version);
    return(0);
}

#ifdef MODULE			/* for loadable module support */

/*
 * Magic named routine called by kernel module support code.
 */
int init_module( void )
{
    return(lis_init_module()) ;
}

/*
 * Magic named routine called by kernel module support code.
 */
void cleanup_module( void )
{
   extern void	lis_kill_qsched(void) ;
   /*
    * Make sure no streams modules are running,
    * and de-register devices unregister_netdev (dev);
    */

    restore_syscalls();
    lis_kill_qsched() ;			/* drivers/str/runq.c */
    lis_terminate_head();

    unregister_chrdev(lis_major,"streams");
    printk ("Linux Streams Subsystem removed\n");
}

#endif

/************************************************************************
*                       lis_print_trace_func                            *
*************************************************************************
*									*
* This function is pointed to by the lis_print_trace pointer.  It makes	*
* cmn_err strings come out in the log. 					*
*									*
************************************************************************/
void lis_print_trace_func(char *msg)
{
    printk("%s", msg) ;

} /* lis_print_trace_func */

/************************************************************************
*                          streams_init                                 *
*************************************************************************
*									*
* This routine is called from the Linux main.c.  It registers the 	*
* streams driver and initializes some streams variables.		*
*									*
* No memory is allocated.						*
*									*
************************************************************************/
void	streams_init(void)
{

    lis_init_module() ;		/* register STREAMS with Linux */

} /* streams_init */

/************************************************************************
*                        lis_inc_mod_cnt                                *
*************************************************************************
*									*
* Increment the module counter for STREAMS.  This prevents us from	*
* being unloaded while we have files open.				*
*									*
************************************************************************/
void	lis_inc_mod_cnt(void)
{
    MOD_INC_USE_COUNT ;

} /* lis_inc_mod_cnt */

/************************************************************************
*                          lis_dec_mod_cnt                              *
*************************************************************************
*									*
* Decrement our module use count.					*
*									*
************************************************************************/
void	lis_dec_mod_cnt(void)
{
    MOD_DEC_USE_COUNT ;

} /* lis_dec_mod_cnt */

#endif /* LINUX */

/************************************************************************
*                         lis_hitime                                    *
*************************************************************************
*									*
* High precision timer support.  Use the kernel do_gettimeofday		*
* routine to get the seconds/micro-seconds timer.  Condense it to	*
* a long with rollover every 64 seconds.  The resulting integer has	*
* micro-second precision.						*
*									*
************************************************************************/
unsigned long	lis_hitime(void)
{
    struct timeval	 tv ;

    do_gettimeofday(&tv) ;		/* kernel time function */
    return( (tv.tv_sec & 0x3f) * 1000000 + tv.tv_usec ) ;

} /* lis_hitime */

/************************************************************************
*                             lis_splstr                                *
*************************************************************************
*									*
* Disable interrupts, return the previous state.			*
*									*
************************************************************************/
static int		 disableinx ;
static unsigned long	 disabletim ;
static int		 disablestate ;

int	lis_splstr(char *file, int line)
{
    int			 prev ;
    spl_track_t		*p ;

    save_flags(prev) ;
    cli() ;

    if ( !LIS_DEBUG_SPL_TRACE ) return(prev) ;

    p = &lis_spl_track[lis_spl_track_inx] ;
    p->file = file ;
    p->line = line ;
    p->type = TRACK_SPLSTR ;
    p->flags = prev ;

    p->tim = lis_hitime() ;
    p->state = disablestate ;
    if (disablestate == 0)		/* first disable */
    {
	disableinx = lis_spl_track_inx ;
	disabletim = p->tim ;
	disablestate = 1 ;
    }

    p->dtim = disabletim ;
    if (++lis_spl_track_inx >= SPL_TRACK_SIZE)
	lis_spl_track_inx = 0 ;

    return(prev) ;

} /* lis_splstr */

/************************************************************************
*                             lis_splx                                  *
*************************************************************************
*									*
* Restore state.							*
*									*
************************************************************************/
void	lis_splx(int x, char *file, int line)
{
    spl_track_t		*p ;

    if ( !LIS_DEBUG_SPL_TRACE )
    {
	restore_flags(x) ;
	return ;
    }

    p = &lis_spl_track[lis_spl_track_inx] ;
    if (++lis_spl_track_inx >= SPL_TRACK_SIZE)
	lis_spl_track_inx = 0 ;

    p->tim = lis_hitime() ;
    p->file = file ;
    p->line = line ;
    p->type = TRACK_SPLX ;
    p->flags = x ;
    p->state = disablestate ;
    p->dtim = disabletim ;

    if ( x & 0x200 )			/* enabling */
    {
	disablestate = 0 ;
#if 0
	if (p->intvl < 1000)		/* short disable */
	{				/* erase all these entries */
	    p->type = 0 ;
	    lis_spl_track_inx = disableinx ;	/* set back index */
	    lis_spl_track[lis_spl_track_inx].type = 0 ;
	}
#endif
    }

    restore_flags(x) ;

} /* lis_splx */

/************************************************************************
*                         lis_print_spl_track                           *
*************************************************************************
*									*
* Print out the spl tracking table.					*
*									*
************************************************************************/
void	lis_print_spl_track(void)
{
    spl_track_t		*p ;
    int			 i ;
    int			 n ;

    printk("lis_print_spl_track: lis_spl_track_inx=%d time=%lu\n",
		lis_spl_track_inx,
		lis_hitime()
    		) ;

    for (i = lis_spl_track_inx, p = &lis_spl_track[i], n = SPL_TRACK_SIZE; n--;)
    {
	if (p->type != 0)
	{
	    printk("%4d: %s %9lu %9lu I=%ld S=%d F=%03x %s #%u\n",
		    i,
		    p->type == TRACK_SPLSTR ? "splstr" : "splx  ",
		    p->dtim,
		    p->tim,
		    p->tim - p->dtim,
		    p->state,
		    p->flags & 0x200,
		    p->file, p->line) ;
	}

	p++ ;
	if (++i >= SPL_TRACK_SIZE)
	{
	    i = 0 ;
	    p = lis_spl_track ;
	}
    }

    printk("lis_print_spl_track: end of output\n") ;
    
} /* lis_print_spl_track */

/************************************************************************
*                       Queue Running                                   *
*************************************************************************
*									*
* The following routines are used to manage running the STREAMS queus.	*
* There are two different techniques, selectable via the Configure	*
* script.  One is to use a "bottom half" routine to do the work.  The	*
* other is to use a kernel thread.					*
*									*
* The bottom half technique is faster and more like SVR4 Unix.  It	*
* has in common with SVR4 the fact that the bottom half context prevents*
* STREAMS service procedures from using certain kernel functions which	*
* have to be called from blockable context.				*
*									*
* The kernel thread technique is slower and is more like the multi-	*
* threaded STREAMS of Solaris, except for being the degenerate case	*
* of only one thread.  Its advantage is the STREAMS service procedures	*
* run in blockable context and can use any kernel function.		*
*									*
* The routines lis_setqsched and lis_kill_qsched are called from outside*
* this module.  lis_setqsched is a request to run the queues.  It means	*
* "now would be a good time to _schedule_ the queues to be run".	*
* list_kill_qsched is called from module unload to kill the kernel	*
* thread.  There is a semaphore interlock on this process to make sure	*
* that the process has been killed before unloading the module.		*
*									*
************************************************************************/

#ifdef USE_BH

/*
 * This runs off of the kernel's bottom half mechanism.  This will
 * happen at kernel exit, return from interrupt or from the schedule
 * routine.  We want to make sure that the running of the service
 * queues does not interrupt driver execution via a putnext from
 * the stream head.  So we pay attention to a flag that is set
 * by the stream head whenever it does a putnext down to a streams
 * driver.
 *
 * Note that a putnext upwards from an interrupt routine will not
 * be preempted by this code since it has to unwind to an interrupt
 * exit before the bottom half mechanism runs.
 *
 * I am sure that this will have to be revisited for SMP.
 */
void	lis_tq_runqueues(void *p)	/* bottom half style */
{
    /* printk("lis_tq_runqueues called\n") ; */

    if (!lis_putnext_flag)
	lis_run_queues() ;
    else
	lis_runq_sched = 0 ;		/* ok to reschedule */

    (void) p;
}

void	lis_setqsched()			/* bottom half style */
{
    static int			initialized ;
    static struct tq_struct	lis_tq ;

    if (!initialized)
    {
	initialized	= 1 ;
	lis_tq.next	= NULL ;
	lis_tq.sync	= 0 ;
	lis_tq.routine	= lis_tq_runqueues ;
	lis_tq.data	= 0 ;
	/* printk("lis_setqsched: lis_tq initialized\n") ; */
    }

    if (lis_queues_running == 0 && lis_runq_sched == 0)
    {					/* no, then wake up qrun process */
	/* printk("lis_setqsched: queue_task and mark_bh\n") ; */
	lis_runq_sched = 1 ;		/* queues waiting to be run */
	queue_task(&lis_tq, &tq_immediate);
	mark_bh(IMMEDIATE_BH);
    }

} /* lis_setqsched */

/*
 * Called from head code when time to unload LiS module.
 */
void	lis_kill_qsched(void)
{
    /* nothing to kill in bottom half implementation */
}

void	lis_start_qsched(void)
{
    /* nothing to do in bottom half style */
}

#else					/* kernel thread technique */

/*
 * This is the entry point of a kernel thread
 */
int	lis_thread_runqueues(void *p)
{
#ifdef KERNEL_2_1
#define	MY_SIG		current->signal.sig[0]
#define	MY_BLKD		current->blocked.sig[0]
#else
#define MY_SIG		current->signal
#define MY_BLKD		current->blocked
#endif
    int		sig_cnt  = 0 ;

    (void) p;

    printk(KERN_INFO "LiS-RunQ running pid=%d\n", current->pid) ;
    exit_files(current) ;		/* close all files */
    exit_mm(current) ;			/* detach user pages */
    current->pgrp = 1 ;			/* detach from parent */
    current->session = 1 ;		/* now owned by "init" */
    current->uid = 0 ;			/* become root */
    current->euid = 0 ;			/* become root */
    current->tty = NULL ;		/* detach from any tty */
    current->policy = SCHED_FIFO ;	/* real-time: run when ready */
    current->rt_priority = 50 ;		/* middle value real-time priority */
    current->mm->arg_start = current->mm->arg_end = 0;
    sprintf(current->comm, "LiS-%s", lis_version) ;
    sigaddset(&current->blocked, SIGTERM) ;	/* inhibit SIGTERM */

    for (;;)
    {
	lis_run_queues() ;		/* run the STREAMS queues */

	/*
	 * When rebooting the kernel sends a TERM signal to all processes
	 * and then a KILL signal.  We want to keep running after the
	 * TERM so that we can close streams from user processes.  We
	 * will go away from a KILL or INT signal so that we can also
	 * be easily killed from the keyboard.
	 */
	if (lis_down(&lis_runq_sem) < 0)	/* interrupted */
	{
	    if (signal_pending(current)		/* killed */
		&& (   sigismember(&current->signal, SIGKILL)
		    || sigismember(&current->signal, SIGINT)
		   )
	       )
		break ;				/* process killed */

	    if (++sig_cnt >= 5)			/* nothing but signals */
	    {
		printk("%s: Signalled: signal=0x%lx blocked=0x%lx, exiting\n",
			current->comm, MY_SIG, MY_BLKD) ;
		break ;
	    }
	}
	else
	    sig_cnt = 0 ;
    }

    printk(KERN_INFO "%s exiting pid=%d\n", current->comm, current->pid) ;
    lis_runq_open_flag = 0 ;		/* logically "closed" */
    SEM_DESTROY(&lis_runq_sem) ;	/* de-init semaphore */
    up(&lis_runq_kill_sem) ;		/* OK, we're killed */
    lis_runq_pid = 0 ;			/* process gone */
    return(0) ;
}

/*
 * Called to start the queue running process.
 */
void	lis_start_qsched(void)
{
    if (lis_runq_pid > 0)		/* already running */
	return ;

    lis_runq_open_flag = 0 ;		/* logically "closed" */
    SET_SEM_COUNT(&lis_runq_sem, 0) ;	/* initialize semaphore */
    SET_SEM_COUNT(&lis_runq_kill_sem, 0) ;	/* initialize semaphore */
    lis_runq_pid = kernel_thread(lis_thread_runqueues, NULL, 0) ;
    if (lis_runq_pid < 0)
    {
	printk("lis_tq_fork: kernel_thread error %d\n", lis_runq_pid) ;
	return ;
    }

    lis_runq_open_flag = 1 ;		/* logically "opened" */
    lis_runq_sched = 0 ;		/* allow "up" on semaphore */
}

void	lis_setqsched()			/* kernel thread style */
{
    if (lis_runq_pid == 0)
    {
	static int	cnt ;
	if (cnt++ < 5)
	    printk("lis_setqsched: No qrun process\n");

	/*
	 * If the runq process is dead we are probably on the way
	 * down for a reboot.  Run the queues from here.
	 */
	lis_run_queues() ;		/* run the STREAMS queues */
	return ;
    }

    if (lis_runq_open_flag && lis_queues_running == 0 && lis_runq_sched == 0)
    {					/* no, then wake up qrun process */
	lis_runq_sched = 1 ;		/* queues waiting to be run */
	lis_up(&lis_runq_sem) ;		/* V() the semaphore */
	if (SEM_COUNT(&lis_runq_sem) > 2)
	    printk("lis_setqsched: "
	    	   "lis_runq_sem.count=%u lis_runq_sem.waking=%u\n",
			SEM_COUNT(&lis_runq_sem), lis_runq_sem.waking) ;
    }

} /* lis_setqsched */

/*
 * Called from head code when time to unload LiS module.
 */
void	lis_kill_qsched(void)
{
    if (lis_runq_pid > 0) /* Only stop the kernel thread if running */
    {
        kill_proc(lis_runq_pid, SIGKILL, 1) ;
        down(&lis_runq_kill_sem) ;
    }
}

#endif
