/*                               -*- Mode: C -*- 
 * mod.c --- module mgmt
 * Author          : Francisco J. Ballesteros Camara
 * Created On      : Thu Jun 29 11:03:59 1995
 * Last Modified By: David Grothe
 * Purpose         : provide LiS STREAMS modules glue
 * ----------------______________________________________________
 *
 *    Copyright (C) 1995  Francisco J. Ballesteros, Denis Froschauer
 *    Copyright (C) 1997  David Grothe, Gcom, Inc <dave@gcom.com>
 *    Copyright (C) 1999  Ole Husgaard <sparre@login.dknet.dk>
 *
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 * 
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330, Cambridge,
 * MA 02139, USA.
 * 
 *
 *    You can reach us by email to any of
 *    nemo@ordago.uc3m.es, 100741.1151@compuserve.com
 */

#ident "@(#) LiS mod.c 2.4 6/21/99 21:23:37 "


/*  -------------------------------------------------------------------  */
/*				 Dependencies                            */

#include <sys/strport.h>
#include <sys/strconfig.h>	/* config definitions	*/
#include <sys/stropts.h>	/* struct str_list	*/
#include <sys/LiS/share.h>	/* streams shared defs	*/
#include <sys/LiS/queue.h>	/* streams queues	*/
#include <sys/LiS/mod.h>	/* module interface	*/

/*  -------------------------------------------------------------------  */
/*				  Glob. vars                             */

struct fmodsw lis_fstr_sw[MAX_STRDEV];	/* streams devices */
struct fmodsw lis_fmod_sw[MAX_STRMOD];	/* streams modules */

extern struct file_operations lis_streams_fops;


/*  -------------------------------------------------------------------  */
/*				   Symbols                               */

int lis_fmodcnt; /* (largest module id used) + 1 */

/*
 *  Module autopush list
 */
typedef struct autopush_list {
	unsigned npush;
	modID_t mod[MAXAPUSH];
} autopush_list_t;

/*
 *  Module autopush entry
 *
 *  These form a chain for each major.
 *  The chain is sorted by the minor field.
 */
typedef struct autopush {
	struct autopush *next;
	unsigned cmd;
	minor_t minor;
	minor_t lastminor;
	autopush_list_t push;
} autopush_t;

static autopush_t *apush[MAX_STRDEV];
static unsigned *apush_nref[MAX_STRMOD];


/*  -------------------------------------------------------------------  */
/*		   Driver and module configuration                       */

/*
 *  The file head/<target>/modconf.c (generated by the strconf utility)
 *  used to be compiled as a seperate object file.
 *  But now it only needs to be known locally in this file.
 *  When we include it here instead, we can move the declarations for the
 *  types used in it from global scope in <sys/LiS/mod.h> to local scope in
 *  this file, reducing overall LiS complexity.
 *  To reflect this change, the name is changed from modconf.c to modconf.inc.
 */ 

/*
 * Module Configuration
 *
 * In order to initialize configured linked streams modules a table of the
 * following form must be linked into the streams code.
 * There is one entry per configured linked module.
 *
 * The fields are:
 *
 *      cnf_name        The ASCII string name of the module.
 *      cnf_str         A pointer to the streamtab entry for the module.
 * 
 * This table is used at initialization time to register all registered
 * linked modules with streams, thus inserting them into the fmod_sw table.
 * 
 * The final entry in the table consists of {"", NULL}.
 *
 * The file modconf.c provides such a table and is automatically
 * generated by the strconf utility.  DO NOT EDIT modconf.c BY HAND.
 */
typedef struct module_config
{
    char               cnf_name[FMNAMESZ+1];
    struct streamtab  *cnf_str;

} module_config_t ;

/*
 * Driver Configuration
 *
 * In order to initialize configured linked streams drivers a table of the
 * following form must be linked into the streams code.
 * There is one entry per configured linked driver.
 *
 * The fields are as follows:
 *
 *      cnf_name        ASCII string name associated with the driver.
 *      cnf_str         Pointer to the driver's streamtab entry.
 *      cnf_major       The major device number to associate with the driver.
 *                      This must correspond with the major device number
 *                      used in a mknod() system call from user level
 *                      to make the device corresponding to this driver.
 *      cnf_n_minors    The number of minors associated with the driver.
 *      cnf_init        A pointer to the init routine of the driver.  NULL
 *                      if the driver has no init routine.
 *
 * This table is used at initialization time to register all configured
 * linked drivers with streams, thus getting them into the fstr_sw table.
 *
 * The final entry in the table is of the form:
 *
 *      {"", NULL, 0, 0, NULL}
 * 
 * The file modconf.c provides such a table and is generated by the
 * strconf utility.  DO NOT EDIT modconf.c BY HAND.
 * 
 */
typedef struct driver_config
{
    char                 cnf_name[FMNAMESZ+1];
    struct streamtab    *cnf_str;
    int                  cnf_major;
    int                  cnf_n_minors;
    void                (*cnf_init)(void);

} driver_config_t ;

#ifdef LIS_LOADABLE_SUPPORT

/*
 * Loadable module object names
 *
 * This table contains the object names of all configured loadable modules.
 * It is sorted on the module name field.
 *
 * This tells LiS how to ask kerneld(8) to load the kernel module that
 * contains the streams module.
 *
 * The file modconf.c provides such a table and is generated by the
 * strconf utility.  DO NOT EDIT modconf.c BY HAND.
 * 
 */
typedef struct module_obj_name {
    const char  *name;          /* The module name                      */
    const char  *objname;       /* The kernel object name               */
} module_obj_name_t;

/*
 * Loadable driver object names
 *
 * This table contains the object names of all configured loadable drivers.
 * It is sorted on the driver major device number field.
 *
 * This tells LiS how to ask kerneld(8) to load the kernel module that
 * contains the streams driver.
 *
 * The file modconf.c provides such a table and is generated by the
 * strconf utility.  DO NOT EDIT modconf.c BY HAND.
 * 
 */
typedef struct driver_obj_name {
    const major_t       major;   /* The driver major device number      */
    const char         *initname;/* The initialization function name    */
    const char         *objname; /* The kernel object name              */
} driver_obj_name_t;

#endif /* ifdef LIS_LOADABLE_SUPPORT */

/*
 * Autopush initialzation specification
 *
 * This table contains the initial autopush specification. This is the
 * autopush configuration that is done when streams is started.
 *
 */
typedef struct autopush_init {
    const int          major;
    const int          minor;
    const int          lastminor;
    autopush_list_t    push;
} autopush_init_t;

/*
 *  The include itself
 *
 *  It is not included when doing dependencies,
 *  but we have a manual dependency in the makefile.
 */
#ifndef DEP
#include "modconf.inc"
#endif

/*
 *  We know the size of the tables
 */
#define DRV_CONFIG_SIZE \
	( sizeof(lis_driver_config) / sizeof(driver_config_t) )

#define MOD_CONFIG_SIZE \
	( sizeof(lis_module_config) / sizeof(module_config_t) )

#define AP_CONFIG_SIZE \
	( sizeof(lis_apush_init) / sizeof(autopush_init_t) )

#ifdef LIS_LOADABLE_SUPPORT

#define DRV_OBJNAMES_SIZE \
	( sizeof(lis_drv_objnames) / sizeof(driver_obj_name_t) )

#define MOD_OBJNAMES_SIZE \
	( sizeof(lis_mod_objnames) / sizeof(module_obj_name_t) )

#endif

/*
 *  Return the index into lis_mod_objnames[] of a module name,
 *  or -1 if not found.
 */
int find_mod_objname(const char *name)
{
#ifdef LIS_LOADABLE_SUPPORT
	int i;

	/* TODO: Use a binary search */
	for (i = 0; i < MOD_OBJNAMES_SIZE; ++i)
		if (!strcmp(name, lis_mod_objnames[i].name))
			return i;
#endif
	return -1;
}

/*
 *  Return the index into lis_drv_objnames[] of a driver major,
 *  or -1 if not found.
 */
int find_drv_objname(major_t major)
{
#ifdef LIS_LOADABLE_SUPPORT
	int i;

	/* TODO: Use a binary search */
	for (i = 0; i < DRV_OBJNAMES_SIZE; ++i)
		if (major == lis_drv_objnames[i].major)
			return i;
#endif
	return -1;
}


/*  -------------------------------------------------------------------  */
/*			   Local functions & macros                      */

/*
 *  Free an autopush entry
 */
static void apush_free(autopush_t *a)
{
	int i;

	ASSERT(a->push.npush >= 0);
	ASSERT(a->push.npush <= MAXAPUSH);

	for (i = 0; i < a->push.npush; ++i) {
		ASSERT(a->push.mod[i] > 0);
		ASSERT(a->push.mod[i] < MAX_STRMOD);
		ASSERT(lis_fmod_sw[a->push.mod[i]].f_str != NULL);
		ASSERT(apush_nref[a->push.mod[i]] > 0);

		--apush_nref[a->push.mod[i]];
	}

	FREE(a);
}

/*
 *  Clear the autopush configuration for a major device number
 */
static void apush_free_major(major_t major)
{
	autopush_t *a, *b;

	ASSERT(major < MAX_STRDEV);

	if ((a = apush[major]) == NULL)
		return;

	apush[major] = NULL;

	while (a != NULL) {
		b = a;
		a = a->next;
		apush_free(b);
	}
}

/*
 *  Clear the autopush configuration
 */
static void apush_free_all(void)
{
	int i;

	for (i = 0; i < MAX_STRDEV; ++i)
		apush_free_major(i);

	/* Pure paranoia: Can go away when we know this stuff works */
	for (i = 0; i < MAX_STRMOD; ++i)
		ASSERT(apush_nref[i] == 0);
}

/*
 *  Remove a module from all autopush lists
 */
static void apush_drop_mod(modID_t mod)
{
	int i;

	ASSERT(mod > 0);
	ASSERT(mod < MAX_STRMOD);

	if (apush_nref[mod] == 0)
		return;

	for (i = 0; i < MAX_STRDEV; ++i) {
		autopush_t *a;

		for (a = apush[i]; a != NULL; a = a->next) {
			int j = 0;

			while (j < a->push.npush) {
				if (a->push.mod[j] == mod) {
					int k;

					for (k = j + 1; k < a->push.npush; ++k)
					    a->push.mod[k-1] = a->push.mod[k];
					--a->push.npush;
					if (--apush_nref[mod] == 0)
						return;
				} else
					++j;
			}
		}
	}
}

/*
 *  Validate an autopush module list
 */
static int apush_validate(autopush_list_t *list)
{
	int i;

	if (list->npush < 0 || list->npush > MAXAPUSH)
		return -1;
	for (i = 0; i < list->npush; ++i) {
		modID_t mod = list->mod[i];

		if (lis_fmod_sw[mod].f_str != NULL)
			continue; /* registered */
#ifdef LIS_LOADABLE_SUPPORT
		if (mod > MOD_OBJNAMES_SIZE)
			return -1; /* not registered or configured */
#else
		return -1; /* not registered or configured */
#endif
	}
	return 0;
}

/*
 *  Do autopush configuration
 */
static int apush_conf(major_t major, autopush_t *a)
{
	autopush_t *ar, **ap;
	int i;

	if (major >= MAX_STRDEV)
		return -ENODEV;

	if (lis_fstr_sw[major].f_str == NULL) {
		/* Driver is not present */
#ifdef LIS_LOADABLE_SUPPORT
		/* Is it a configured loadable? */
		if (find_drv_objname(major) < 0)
			return -ENODEV;
#else
		return -ENODEV;
#endif
	}

	/*
	 *  Check for conflicting entries and insert new entry.
	 *
	 *  This code may look a bit hairy. I have tried to state
	 *  inherited predicates in the form of ASSERT's to clarify.
	 */
	switch (a->cmd) {
	    case SAP_CLEAR:
		if (apush[major] == NULL)
			return -ENODEV;
		ap = &apush[major];
		if (a->minor == 0 && (*ap)->cmd == SAP_ALL) {
			ASSERT((*ap)->next == NULL);
			ar = *ap;
			*ap = (*ap)->next;
			apush_free(ar);
			return 0;
		}
		for (;;) {
			ar = *ap;
			if (ar == NULL)
				return -ERANGE;
			if (ar->minor > a->minor)
				return -ERANGE;
			if (ar->minor == a->minor) {
				*ap = ar->next;
				apush_free(ar);
				return 0;
			}
			ap = &ar->next;
		}
		/*NOTREACHED*/
	   case SAP_ONE:
		ASSERT(apush_validate(&a->push) >= 0);
		ap = &apush[major];
		if (*ap != NULL) {
			if ((*ap)->cmd == SAP_ALL)
				return -EEXIST;
			for (;;) {
				ar = *ap;
				if (ar == NULL)
					break; /* insert at end of chain */
				if (ar->minor > a->minor)
					break; /* insert here */
				if (ar->cmd == SAP_ONE) {
					if (ar->minor == a->minor)
						return -EEXIST;
				} else {
					ASSERT(ar->cmd == SAP_RANGE);
					ASSERT(a->minor >= ar->minor);
					if (a->minor <= ar->lastminor)
						return -EEXIST;
				}
				ap = &ar->next;
			}
		}
		a->next = *ap;
		*ap = a;
		break;
	    case SAP_RANGE:
		ASSERT(apush_validate(&a->push) >= 0);
		if (a->minor >= a->lastminor)
			return -ERANGE;
		ap = &apush[major];
		if (*ap != NULL) {
			if ((*ap)->cmd == SAP_ALL)
				return -EEXIST;
			for (;;) {
				ar = *ap;
				if (ar == NULL)
					break; /* insert at end of chain */
				if (ar->minor > a->lastminor)
					break; /* insert here */
				if (ar->minor >= a->minor) {
					ASSERT(ar->minor <= a->lastminor);
					return -EEXIST;
				}
				if (ar->cmd != SAP_ONE) {
					ASSERT(ar->cmd == SAP_RANGE);
					ASSERT(ar->minor <= a->lastminor);
					ASSERT(ar->minor < a->minor);
					ASSERT(a->minor < a->lastminor);
					ASSERT(ar->minor < a->lastminor);
					if (ar->lastminor <= a->lastminor
					    || ar->lastminor >= a->minor)
						return -EEXIST;
				}
				ap = &ar->next;
			}
		}
		a->next = *ap;
		*ap = a;
		break;
	    case SAP_ALL:
		ASSERT(apush_validate(&a->push) >= 0);
		if (apush[major] != NULL)
			return -EEXIST;
		a->next = NULL;
		apush[major] = a;
		break;
	    default:
		return -EINVAL;
	}

	for (i = 0; i < a->push.npush; ++i)
		++apush_nref[a->push.mod[i]];

	return 0;
}

static struct autopush *find_apush_entry(major_t major, minor_t minor)
{
	struct autopush *a;

	ASSERT(major < MAX_STRDEV);

	for (a = apush[major]; a != NULL; a = a->next) {
		ASSERT(a->cmd==SAP_ALL || a->cmd==SAP_ONE || a->cmd==SAP_RANGE);

		if (a->cmd == SAP_ALL)
			break;
		if (a->minor > minor)
			return NULL;
		if (a->minor == minor)
			break;
		if (a->cmd == SAP_RANGE && a->lastminor <= minor)
			break;
	}

	ASSERT(a == NULL || a->minor <= minor);

	return a;
}


/*  -------------------------------------------------------------------  */
/*			Exported functions & macros                      */


/*  -------------------------------------------------------------------  */
/* register a new module
 */
modID_t
lis_register_strmod(struct streamtab *strtab, const char *name)
{
	modID_t id = LIS_NULL_MID;
	int i;

	ASSERT(name != NULL);
	ASSERT(strtab != NULL);

	if (name == NULL)
		return LIS_NULL_MID;

	if ((i = find_mod_objname(name)) >= 0) {
		id = i + 1;
		if (lis_fmod_sw[id].f_str != NULL) {
			printk("Unable to register configured "
			       "loadable module \"%s\", slot in use.\n",
			       name);
			return LIS_NULL_MID;
		}
	}

	if (id == LIS_NULL_MID) {
#ifdef LIS_LOADABLE_SUPPORT
		/* Not a configured loadable: Find a free id */
		for (id = MOD_OBJNAMES_SIZE + 1; id < lis_fmodcnt; ++id)
			if (lis_fmod_sw[id].f_str == NULL)
				break;
#else
		/* Find a free id */
		for (id = 1; id < lis_fmodcnt; ++id)
			if (lis_fmod_sw[id].f_str == NULL)
				break;
#endif
	}

	if (id == lis_fmodcnt) {
		if (id == MAX_STRMOD) {
			printk("Unable to register module \"%s\", "
			       "all lis_fmod_sw slots in use.\n", name);
			return LIS_NULL_MID;
		}
		lis_fmodcnt++;
	}
	lis_fmod_sw[id].f_str = strtab;
	lis_fmod_sw[id].f_count = 0;
	strncpy(lis_fmod_sw[id].f_name, name, FMNAMESZ);

	printk("Streams module \"%s\" registered, id %d\n",
	       lis_fmod_sw[id].f_name, id);

	return id;
} /* lis_register_strmod */

/*  -------------------------------------------------------------------  */
/* unregister this module
 */
int
lis_unregister_strmod(struct streamtab *strtab)
{
	modID_t id;
	fmodsw_t *slot;
	char name[FMNAMESZ + 1];

	for (slot = lis_fmod_sw + lis_fmodcnt; 
	     slot != lis_fmod_sw + 1 && slot->f_str != strtab; slot--)
		;
	if (slot->f_str != strtab)
		return -EINVAL;
	if (slot->f_count)
		return -EBUSY;

	strncpy(name, slot->f_name, FMNAMESZ);

	slot->f_str = NULL;
	slot->f_name[0] = '\0';

	id = slot - lis_fmod_sw;
#ifdef LIS_LOADABLE_SUPPORT
	if (id > MOD_OBJNAMES_SIZE)
		apush_drop_mod(id);
#else
	apush_drop_mod(id);
#endif

	printk("Streams module \"%s\" unregistered, id %d\n",
	       name, id);

	return 0;
} /* lis_unregister_strmod */

/*  -------------------------------------------------------------------  */
/* Find module name in lis_fmod_sw[] 
 */
modID_t
lis_findmod(const char *name)
{
	int i;
#ifdef LIS_LOADABLE_SUPPORT
	int configured;
	const char *objname;
	char req[FMNAMESZ + 10];
#endif

	/* Look for the module */
	for (i = lis_fmodcnt; i > 0; i--)
		if (!strcmp(lis_fmod_sw[i].f_name, name))
			return i; /* found it */

#ifdef LIS_LOADABLE_SUPPORT
	/* Find object name of this module */
	objname = name; /* default objname */
	configured = 0;
	if ((i = find_mod_objname(name)) >= 0) {
		objname = lis_mod_objnames[i].objname;
		configured = 1;
	}

	/* Ask kerneld to load the module and retry */
	sprintf(req, "streams-%s", objname);
	request_module(req);
	if (configured && !strcmp(lis_fmod_sw[i].f_name, name))
		return i;
		
	for (i = lis_fmodcnt; i > 0; i--)
		if (!strcmp(lis_fmod_sw[i].f_name, name))
			return i;
	if (configured)
		printk("Unable to demand load LiS objname %s, "
		       "STREAMS module %s\n ",
		       objname, (name) ? name : "(null)");
	else
		printk("Unable to demand load STREAMS module %s\n",
		       (name) ? name : "(null)");
#endif

	return LIS_NULL_MID;
} /* lis_findmod */

/*  -------------------------------------------------------------------  */
/* register a new streams device
 */
int 
lis_register_strdev(major_t major, 
		    struct streamtab *strtab, int nminor, const char *name)
{
	int rslt;
	fmodsw_t *slot;

	(void)nminor; /* ignore minor count, compiler happiness */

	rslt = register_chrdev(major, name, &lis_streams_fops);
	if (rslt < 0) {
		printk("Unable to get major %lu "
		       "for stream driver \"%s\", errno=%d\n",
		       major, (name != NULL) ? name : "", rslt);
		return -EIO;
	}

	if (major == 0)			/* register routine assigned it */
		major = rslt;

	slot = &lis_fstr_sw[major];

	if (slot->f_str != strtab && slot->f_str != NULL) {
		printk("Unable to register major %lu "
		       "for stream driver \"%s\": In use for driver \"%s\".\n",
		       major, (name != NULL) ? name : "", slot->f_name);
		return -EBUSY;
	}

	slot->f_str = strtab;
	slot->f_count = 0;
	if (name)
		strncpy(slot->f_name, name, FMNAMESZ);
	else
		*slot->f_name = '\0';

	printk("Streams driver \"%s\" registered, major %lu\n",
	       slot->f_name, major);

	return major;
} /* lis_register_strdev */

/*  -------------------------------------------------------------------  */
/* unregister a streams device
 */
int 
lis_unregister_strdev(major_t major)
{

	if (major >= MAX_STRDEV || lis_fstr_sw[major].f_str == NULL) {
		printk("Streams driver not registered, "
		       "cannot unregister major %lu\n", major);
		return -ENODEV;
	}

	if (lis_fstr_sw[major].f_count) {
		printk("Streams driver \"%s\" has open count %d, "
		       "cannot unregister major %lu\n",
		       lis_fstr_sw[major].f_name, lis_fstr_sw[major].f_count,
		       major);
		return -EBUSY;
	}

#ifdef LIS_LOADABLE_SUPPORT
	/* Is this driver a configured loadable */
	if (find_drv_objname(major) < 0)
		apush_free_major(major);
#else
	apush_free_major(major);
#endif

	lis_fstr_sw[major].f_str = NULL;
	unregister_chrdev(major, lis_fstr_sw[major].f_name);

	printk("Streams driver \"%s\" unregistered, major %lu\n",
	       lis_fstr_sw[major].f_name, major);

	return 0;
} /* lis_unregister_strdev */

/*  -------------------------------------------------------------------  */
/* Find a driver by major device number and return a pointer to
 * its streamtab entry.
 */
streamtab_t *
lis_find_strdev(major_t major)
{
	if (major >= MAX_STRDEV)
		return NULL;

#ifdef LIS_LOADABLE_SUPPORT
	if (lis_fstr_sw[major].f_str == NULL) {
		const char *objname = NULL, *initname = NULL;
		char name[30];
		int i;

		/* Find object name of this driver */
		if ((i = find_drv_objname(major)) >= 0) {
			objname = lis_drv_objnames[i].objname;
			initname = lis_drv_objnames[i].initname;
		}
		if (objname == NULL)
			sprintf(name, "char-major-%lu", major);
		else
			sprintf(name, "streams-%s", objname);

		request_module(name);

		if (lis_fstr_sw[major].f_str == NULL) {
			if (objname != NULL) {
			    sprintf(name, "char-major-%lu", major);
			    request_module(name);
			    if (lis_fstr_sw[major].f_str == NULL) {
				printk("Unable to demand load "
				       "configured STREAMS object %s, "
				       "device major %lu\n",
				       objname, major);
				return NULL;
			    }
			} else {
				printk("Unable to demand load "
				       "unconfigured STREAMS "
				       "device major %lu\n",
				       major);
				return NULL;
			}
		}
		if (initname != NULL) { /* call initialization */
			void (* init)(void);
			/*
			 *  get_module_symbol() author forgot to declare
			 *  untouched arguments as "const char *".
			 *  Avoid warning by declaring prototype here
			 *  instead of including <linux/module.h>.
			 *  This is a bit dirty, but I *did* check the
			 *  function source.
			 */
			extern void *get_module_symbol(const char *,
						       const char *);

			init = get_module_symbol(name, initname);
			if (init == NULL)
				printk("lis_find_strdev(): "
				       "Unable to resolve init function %s "
				       "in module %s for major %lu\n",
				       initname, objname, major);
			else
				(* init)();
		}
	}
#endif
	return lis_fstr_sw[major].f_str;
} /* lis_find_strdev */


/*  -------------------------------------------------------------------  */
/*				 Autopush                                */

/*
 *  Get list of autopush modules for a device
 */
int lis_apushm(dev_t dev, const char *mods[])
{
	int major = STR_MAJOR(dev);
	int minor = STR_MINOR(dev);
	struct autopush *a;
	int i, psw;

	if (major < 0 || major >= MAX_STRDEV)
		return 0;

	SPLSTR(psw);
	if ((a = find_apush_entry(major, minor)) == NULL) {
		SPLX(psw);
		return 0;
	}

	for (i = 0; i < a->push.npush; ++i) {
		modID_t mod = a->push.mod[i];

		ASSERT(mod > 0);
		ASSERT(mod < MAX_STRMOD);

#ifdef LIS_LOADABLE_SUPPORT
		if (lis_fmod_sw[mod].f_str != NULL)
			mods[i] = lis_fmod_sw[mod].f_name;
		else {
			ASSERT(mod <= MOD_OBJNAMES_SIZE);
			mods[i] = lis_mod_objnames[i+1].name;
		}
#else
		ASSERT(lis_fmod_sw[mod].f_str != NULL);
		mods[i] = lis_fmod_sw[mod].f_name;
#endif
	}
	SPLX(psw);

	return a->push.npush;
} /* lis_apushm */

int lis_apush_set(struct strapush *ap)
{
	int i, j, err, psw;
	autopush_t *a;

	ASSERT(ap != NULL);

	if (lis_fstr_sw[ap->sap_major].f_str == NULL) {
		/* Is this a configured loadable driver? */
		if (find_drv_objname(ap->sap_major) < 0)
			return -ENOSTR;
	}

	if (ap->sap_npush > MAXAPUSH)
		return -EINVAL;

	if ((a = ALLOC(sizeof(autopush_t))) == NULL)
		return -ENOSR;

	a->cmd = ap->sap_cmd;
	a->minor = ap->sap_minor;
	a->lastminor = ap->sap_lastminor;
	a->push.npush = ap->sap_npush;

	/* Get module IDs */
	for (i = 0; i < ap->sap_npush; ++i) {
		/* Is this module a configured loadable? */
		if ((j = find_mod_objname(ap->sap_list[i])) >= 0) {
			a->push.mod[i] = j + 1;
			continue;
		}
		/* Nope, is it registered? */
		for (j = lis_fmodcnt; j > 0; --j) {
			if (!strcmp(ap->sap_list[i], lis_fmod_sw[j].f_name)) {
				a->push.mod[i] = j;
				break;
			}
		}
		if (j == 0) {
			/* Unknown module */
			FREE(a);
			return -EINVAL;
		}
	}

	SPLSTR(psw);
	if (apush_validate(&a->push) < 0) {
		SPLX(psw);
		FREE(a);
		return -EINVAL;
	}
	if ((err = apush_conf(ap->sap_major, a)) < 0) {
		SPLX(psw);
		FREE(a);
		return err;
	}
	SPLX(psw);

	return 0;
} /* lis_apush_set */

int lis_apush_get(struct strapush *ap)
{
	autopush_t *a;
	int i, psw;

	if (ap->sap_major >= MAX_STRDEV)
		return -EINVAL;

	if (lis_fstr_sw[ap->sap_major].f_str == NULL) {
		/* Is this a configured loadable driver? */
		if (find_drv_objname(ap->sap_major) < 0)
			return -ENOSTR;
	}

	SPLSTR(psw);
	if ((a = find_apush_entry(ap->sap_major, ap->sap_minor)) == NULL) {
		SPLX(psw);
		return -ENODEV;
	}

	ap->sap_cmd = a->cmd;
	ASSERT(ap->sap_minor == a->minor);
	ap->sap_lastminor = a->lastminor;
	ap->sap_npush = a->push.npush;

	for (i = 0; i < a->push.npush; ++i) {
		modID_t id = a->push.mod[i];

		ASSERT(id > 0);
		ASSERT(id < MAX_STRMOD);

#ifdef LIS_LOADABLE_SUPPORT
		if (lis_fmod_sw[id].f_str != NULL)
			strncpy(ap->sap_list[i],
				lis_fmod_sw[id].f_name, FMNAMESZ + 1);
		else {
			ASSERT(id <= MOD_OBJNAMES_SIZE);
			strncpy(ap->sap_list[i],
				lis_mod_objnames[id + 1].name, FMNAMESZ + 1);
		}
#else
		ASSERT(lis_fmod_sw[id].f_str != NULL);
		strncpy(ap->sap_list[i], lis_fmod_sw[id].f_name, FMNAMESZ + 1);
#endif
	}
	SPLX(psw);

	return 0;
} /* lis_apush_get */

int lis_valid_mod_list(struct str_list ml)
{
	int i, j;
	struct str_mlist *lp;

	if (ml.sl_nmods <= 0)
		return -EINVAL;

	for (i = 0, lp = ml.sl_modlist; i < ml.sl_nmods; ++i, ++lp) {
		/* Is this module a configured loadable? */
		if (find_mod_objname(lp->l_name) >= 0)
			continue;
		/* Nope, is it registered? */
		for (j = lis_fmodcnt; j > 0; j--)
			if (!strcmp(lp->l_name, lis_fmod_sw[j].f_name))
				break;
		if (j == 0)
			return 1; /* Unknown module */
	}

	return 0;
} /* lis_validate_mod_list */


/*
 *  Initialization
 */
void lis_init_mod(void)
{
	int i;

	memset(lis_fstr_sw, 0, sizeof lis_fstr_sw);
	memset(lis_fmod_sw, 0, sizeof lis_fmod_sw);
	memset(apush,       0, sizeof apush);
	memset(apush_nref,  0, sizeof apush_nref);

#ifdef LIS_LOADABLE_SUPPORT
	lis_fmodcnt = MOD_OBJNAMES_SIZE + 1;
#else
	lis_fmodcnt = 1;
#endif

	for (i = 0; i < MOD_CONFIG_SIZE; ++i) {
		ASSERT(lis_fmodcnt < MAX_STRMOD);

		if (lis_register_strmod(lis_module_config[i].cnf_str,
					lis_module_config[i].cnf_name)
								== LIS_NULL_MID)
			printk("Failed to register module \"%s\".\n",
			       lis_module_config[i].cnf_name);
	}

	for (i = 0; i < DRV_CONFIG_SIZE; ++i) {
		if (lis_register_strdev(lis_driver_config[i].cnf_major,
					lis_driver_config[i].cnf_str,
					lis_driver_config[i].cnf_n_minors,
					lis_driver_config[i].cnf_name ) < 0) {
			printk("Failed to register driver \"%s\".\n",
			       lis_driver_config[i].cnf_name);
			continue;
		}
		if (lis_driver_config[i].cnf_init != NULL)
			(*lis_driver_config[i].cnf_init)();
	}

	for (i = 0; i < AP_CONFIG_SIZE; ++i) {
		autopush_t *a;
		int err;

		if ((a = ALLOC(sizeof(autopush_t))) == NULL) {
			printk("Failed to allocate autopush entry.\n");
			continue;
		}

		if (lis_apush_init[i].minor == -1)
			a->cmd = SAP_ALL;
		else if (lis_apush_init[i].lastminor == 0)
			a->cmd = SAP_ONE;
		else
			a->cmd = SAP_RANGE;
		if (a->cmd == SAP_ALL)
			a->minor = a->lastminor = 0;
		else {
			a->minor = lis_apush_init[i].minor;
			a->lastminor = lis_apush_init[i].lastminor;
		}
		a->push = lis_apush_init[i].push;

		if ((err = apush_conf(lis_apush_init[i].major, a)) < 0) {
			FREE(a);
			printk("Failed to add initial autopush entry, "
			       "error = %d.\n", -err);
			continue;
		}
	}
	if (AP_CONFIG_SIZE > 0)
		printk("Added initial autopush entries.\n");
}

/*
 *  Shutdown
 */
void lis_terminate_mod(void)
{
	int i;

	apush_free_all();

	for (i = 0; i < MOD_CONFIG_SIZE; ++i)
		if (lis_unregister_strmod(lis_module_config[i].cnf_str) < 0)
			printk("Failed to unregister module \"%s\".\n",
			       lis_module_config[i].cnf_name);

	for (i = 0; i < DRV_CONFIG_SIZE; ++i)
		if (lis_unregister_strdev(lis_driver_config[i].cnf_major) < 0)
			printk("Failed to unregister driver \"%s\".\n",
			       lis_driver_config[i].cnf_name);
}

