/************************************************************************
*                   Operating System Interface                          *
*************************************************************************
*									*
* The routines in this file provide a filtered interface to certain	*
* operating system routines.  These are routines that drivers are	*
* most likely to use such as PCI BIOS calls and mapping addresses.	*
*									*
* This code loads as a loadable module, but it has no open routine.	*
*									*
* Additions to this file are welcome.					*
*									*
*	Copyright (c) 1999 David Grothe <dave@gcom.com>			*
*									*
* This library is free software; you can redistribute it and/or		*
* modify it under the terms of the GNU Library General Public		*
* License as published by the Free Software Foundation; either		*
* version 2 of the License, or (at your option) any later version.	*
* 									*
* This library is distributed in the hope that it will be useful,	*
* but WITHOUT ANY WARRANTY; without even the implied warranty of	*
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU	*
* Library General Public License for more details.			*
*								 	*
* You should have received a copy of the GNU Library General Public	*
* License along with this library; if not, write to the	Free Software	*
* Foundation, Inc., 59 Temple Place - Suite 330, Cambridge, MA 02139 USA*
*									*
************************************************************************/

#ident "@(#) LiS osif.c 1.7 6/17/99"

#include <sys/stream.h>
#include <linux/autoconf.h>		/* Linux config defines */

#if 0
#if LINUX_VERSION_CODE >= 0x020100      /* 2.2 kernel */
#ifdef pcibios_read_config_byte
#undef pcibios_read_config_byte
#endif
#endif
#endif

#include <linux/pci.h>		/* PCI BIOS functions */
#include <linux/sched.h>		/* odd place for request_irq */
#include <linux/ioport.h>		/* request_region */
#include <asm/dma.h>
#include <linux/malloc.h>
#include <asm/io.h>			/* ioremap, virt_to_phys */
#if LINUX_VERSION_CODE < 0x020100       /* 2.0 kernel */
#include <linux/bios32.h>		/* old style PCI routines */
#include <linux/mm.h>			/* vremap */
#endif

#define	INCL_FROM_OSIF_DRIVER		/* do not change routine names */
#include <sys/osif.h>



/************************************************************************
*                       PCI BIOS Functions                              *
*************************************************************************
*									*
* These are filtered calls to the kernel's PCI BIOS routines.		*
*									*
************************************************************************/

int lis_pcibios_present(void)
{
    return(pcibios_present()) ;
}

/*
 * PCO BIOS init routine is not an exported symbol, but we define it
 * anyway in case some driver was calling it.
 */
#if LINUX_VERSION_CODE < 0x020100               /* 2.0 kernel */
unsigned long lis_pcibios_init(unsigned long memory_start,
			        unsigned long memory_end)
{
#if defined(CONFIG_PCI)
    return(1) ;			/* assume if configured then init'd by now */
#else
    return(0) ;			/* not configured, not initialized */
#endif
}
#else                                           /* 2.1 or 2.2 kernel */
void lis_pcibios_init(void)
{
}
#endif

int lis_pcibios_find_class(unsigned int   class_code,
			    unsigned short index,
                            unsigned char *bus,
			    unsigned char *dev_fn)
{
    return(pcibios_find_class(class_code, index, bus, dev_fn)) ;
}

int lis_pcibios_find_device(unsigned short vendor,
			     unsigned short dev_id,
			     unsigned short index,
			     unsigned char *bus,
			     unsigned char *dev_fn)
{
    return(pcibios_find_device(vendor, dev_id, index, bus, dev_fn)) ;
}

int lis_pcibios_read_config_byte(unsigned char  bus,
				  unsigned char  dev_fn,
				  unsigned char  where,
				  unsigned char *val)
{
#if LINUX_VERSION_CODE < 0x020100       /* 2.0 kernel */
    return pcibios_read_config_byte(bus, dev_fn, where, val);
#else					/* 2.1/2.2 kernel */
    if (where != PCI_INTERRUPT_LINE) 
	return pcibios_read_config_byte(bus, dev_fn, where, val);
    else {
	struct pci_dev *dev;

	if ((dev = pci_find_slot(bus, dev_fn)) != NULL) {
	    *val = dev->irq;
	    return PCIBIOS_SUCCESSFUL;
	}
    }
    *val = 0;
    return PCIBIOS_DEVICE_NOT_FOUND;
#endif
}

int lis_pcibios_read_config_word(unsigned char   bus,
				  unsigned char   dev_fn,
				  unsigned char   where,
				  unsigned short *val)
{
    return(pcibios_read_config_word(bus, dev_fn, where, val)) ;
}

int lis_pcibios_read_config_dword(unsigned char  bus,
				   unsigned char  dev_fn,
				   unsigned char  where,
				   unsigned int  *val)
{
    return(pcibios_read_config_dword(bus, dev_fn, where, val)) ;
}

int lis_pcibios_write_config_byte(unsigned char  bus,
				   unsigned char  dev_fn,
				   unsigned char  where,
				   unsigned char  val)
{
    return(pcibios_write_config_byte(bus, dev_fn, where, val)) ;
}

int lis_pcibios_write_config_word(unsigned char   bus,
				   unsigned char   dev_fn,
				   unsigned char   where,
				   unsigned short  val)
{
    return(pcibios_write_config_word(bus, dev_fn, where, val)) ;
}

int lis_pcibios_write_config_dword(unsigned char  bus,
				    unsigned char  dev_fn,
				    unsigned char  where,
				    unsigned int   val)
{
    return(pcibios_write_config_dword(bus, dev_fn, where, val)) ;
}

/************************************************************************
*                        IRQ Routines                                   *
************************************************************************/

int lis_request_irq(unsigned int  irq,
	             void        (*handler)(int, void *, struct pt_regs *),
	             unsigned long flags,
		     const char   *device,
		     void         *dev_id)
{
    return(request_irq(irq, handler, flags, device, dev_id)) ;
}

void lis_free_irq(unsigned int irq, void *dev_id)
{
    free_irq(irq, dev_id) ;
}


/************************************************************************
*                       Memory Mapping Routines                         *
*************************************************************************
*									*
* These are a subset of what is available.  Feel free to add more.	*
*									*
* The 2.0.x routines will call 2.2.x routines if we are being compiled	*
* for 2.2.x, and vice versa.  This gives drivers something in the way	*
* of version independence.  Wish the kernel guys would tend to that	*
* a bit more.								*
*									*
************************************************************************/

void *lis_ioremap(unsigned long offset, unsigned long size)
{
#if LINUX_VERSION_CODE < 0x020100               /* 2.0 kernel */
    return(vremap(offset, size)) ;
#else
    return(ioremap(offset, size)) ;
#endif
}

void *lis_ioremap_nocache(unsigned long offset, unsigned long size)
{
#if LINUX_VERSION_CODE < 0x020100               /* 2.0 kernel */
    return(vremap(offset, size)) ;
#else
    return(ioremap_nocache(offset, size)) ;
#endif
}

void *lis_vremap(unsigned long offset, unsigned long size)
{
#if LINUX_VERSION_CODE < 0x020100               /* 2.0 kernel */
    return(vremap(offset, size)) ;
#else
    return(ioremap_nocache(offset, size)) ;
#endif
}

unsigned long lis_virt_to_phys(volatile void *addr)
{
    return(virt_to_phys(addr)) ;
}

void *lis_phys_to_virt(unsigned long addr)
{
    return(phys_to_virt(addr)) ;
}


/************************************************************************
*                       I/O Ports Routines                              *
************************************************************************/

int lis_check_region(unsigned int from, unsigned int extent)
{
    return(check_region(from, extent)) ;
}

void lis_request_region(unsigned int from,
			 unsigned int extent,
			 const char  *name)
{
    request_region(from, extent, name) ;
}

void lis_release_region(unsigned int from, unsigned int extent)
{
    release_region(from, extent) ;
}

/************************************************************************
*                    Memory Allocation Routines                         *
************************************************************************/

void *lis_kmalloc(size_t nbytes, int type)
{
    return(kmalloc(nbytes, type)) ;
}

void  lis_kfree(const void *ptr)
{
    kfree((void *)ptr) ;
}

/************************************************************************
*                        DMA Routines                                   *
************************************************************************/

int  lis_request_dma(unsigned int dma_nr, const char *device_id)
{
    return(request_dma(dma_nr, device_id)) ;
}

void lis_free_dma(unsigned int dma_nr)
{
    free_dma(dma_nr) ;
}

