/*                               -*- Mode: C -*- 
 * stream.c --- STREAMS entry points and main routines 
 * Author          : Graham Wheeler, Francisco J. Ballesteros
 * Created On      : Tue May 31 22:25:19 1994
 * Last Modified By: David Grothe
 * RCS Id          : $Id: stream.c,v 1.4 1996/01/17 18:57:07 dave Exp $
 * ----------------______________________________________________
 *
 *   Copyright (C) 1995  Francisco J. Ballesteros, Graham Wheeler,
 *                       Denis Froschauer
 *   Copyright (C) 1997  David Grothe, Gcom, Inc <dave@gcom.com>
 *
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 * 
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330, Cambridge,
 * MA 02139, USA.
 * 
 *
 *    You can reach us by email to any of
 *    nemo@ordago.uc3m.es, gram@aztec.co.za, 100741.1151@compuserve.com
 */

#ident "@(#) LiS stream.c 2.1 4/4/99 11:34:25 "


/*  -------------------------------------------------------------------  */
/*				 Dependencies                            */

#include <sys/strport.h>
#include <sys/stream.h>	/* this will include the whole stuff */

/*  -------------------------------------------------------------------  */
/*				    Types                                */


/*  -------------------------------------------------------------------  */
/*				  Glob. Vars                             */

extern int	lis_queues_running ;



/*  -------------------------------------------------------------------  */
/*			   Local functions & macros                      */

/*  -------------------------------------------------------------------  */

/*  -------------------------------------------------------------------  */
/* service scheduler
 */
static void
queuerun(void)
{
    queue_t	*q;
    queue_t	*qhead ;
    stdata_t	*strmhd ;
    int		 psw;

    if (LIS_DEBUG_MONITOR_MEM)
	lis_check_mem() ;

    /* Process scan queue head list
     *
     * Take the scan queue head and tail pointers away into locals.
     * Clear the globals.  This allows us to traverse the queue
     * with interrrupts wide open.  
     */
    SPLSTR(psw);
    qhead = lis_scanqhead ;
    lis_scanqhead= lis_scanqtail = NULL ;
    SPLX(psw);

    for (q= qhead; q != NULL; 
	 qhead= q->q_scnxt, q->q_scnxt=NULL, q=qhead)
    {
	if (!LIS_CHECK_Q_MAGIC(q))
	{
	    printk("queuerun: scan queue is now invalid\n") ;
	    break ;
	}

	SPLSTR(psw);
	F_CLR(q->q_flag,QSCAN);
	SPLX(psw);
	strmhd = (stdata_t *) q->q_ptr ;
	if (strmhd->sd_wmsg != NULL)		/* held msg waiting */
	{
	    lis_putnext(LIS_WR(q),strmhd->sd_wmsg);	/* send it */
	    strmhd->sd_wmsg = NULL ;		/* don't reuse msg */
	}
    }

    if (lis_strbcflag)
      lis_dobufcall();

    /*
     * Make local copies of the service procedure queue head and
     * tail.  We can then traverse this list with interrupts on.
     */
    SPLSTR(psw);
    qhead = lis_qhead ;
    lis_qhead = lis_qtail = NULL ;
    SPLX(psw);

    /* for each scheduled queue on the qhead list... */
    for (q = qhead; q != NULL; q = qhead)
    {
	if (!LIS_CHECK_Q_MAGIC(q))
	{
	    printk("queuerun: service queue list is now invalid\n") ;
	    break ;
	}

	SPLSTR(psw);				/* protect queue fields */
	/* clear the QENAB flag */
	q->q_flag &= ~QENAB;

	/* remove it from the list */
	qhead     = q->q_link ;			/* next queue to examine */
	q->q_link = NULL ;			/* clear link of this q */
	LisDownCount(QSCHEDS) ;
	SPLX(psw);

	/* call its service procedure, if any */
	if (q->q_qinfo && q->q_qinfo->qi_srvp) 
	    (*q->q_qinfo->qi_srvp)(q);
    }

}/*queuerun*/


/*  -------------------------------------------------------------------  */
/*			Exported functions & macros                      */

/*  -------------------------------------------------------------------  */
/* lis_run_queues - entry point to the STREAMS scheduler. This is called
 *		    whenever we want the STREAMS queues to actually
 *		    be run.
 */
void 
lis_run_queues(void)
{
    int		psw;
    extern int	lis_runq_sched ;	/* linux-mdep.c */

    SPLSTR(psw);			/* intrs off to chk q heads */
    while (lis_qhead != NULL || lis_scanqhead != NULL || lis_strbcflag)
    {
	if (lis_queues_running)		/* recursion protection */
	{
	    SPLX(psw);
	    return;
	}

	lis_queues_running++;
	SPLX(psw);

	queuerun();			/* run q's w/intrs enabled */

	SPLSTR(psw);
	lis_queues_running--;
    }

    lis_runq_sched = 0 ;		/* OK to V semaphore now */
    SPLX(psw);				/* intrs enabled upon return */

}/*lis_run_queues*/

/* process a held msg tmout
 */
void  lis_scantmout_handler(unsigned long arg) {

  (void)arg;
  lis_runqueues();
}
