/*                               -*- Mode: C -*- 
 * syscalls.c --- 
 * Author          : Denis Froschauer
 * Created On      : 
 * Last Modified By: David Grothe (dave@gcom.com)
 * RCS Id          : $Id: syscalls.c,v 1.4 1996/01/23 19:40:13 dave Exp $
 * ----------------______________________________________________
 *   Copyright (C) 1995  Denis Froschauer
 *   Copyright (C) 1997  David Grothe, Gcom, Inc <dave@gcom.com>
 *
 *    This library is free software; you can redistribute it and/or
 *    modify it under the terms of the GNU Library General Public
 *    License as published by the Free Software Foundation; either
 *    version 2 of the License, or (at your option) any later version.
 *
 *    This library is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    Library General Public License for more details.
 *
 *    You should have received a copy of the GNU Library General Public
 *    License along with this library; if not, write to the Free
 *    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    You can reach me by email to
 *    100741.1151@compuserve.com
 */


#ident "@(#) LiS syscalls.c 2.2 5/25/99 18:11:20 "

#ifdef LINUX

#include <sys/stream.h>
#include <sys/stropts.h>	/* streams config symbols are here */
#include <sys/poll.h>		/* usr structs */
#include <linux/unistd.h>	/* kernel system call definitions */

/*
 * Do not evaluate the following if we are only doing
 * dependencies.
 */
#if !defined(DEP) && !defined(__NR_putpmsg)

# ifndef __NR_streams1
# error "unistd.h does not define __NR_putpmsg or __NR_streams2"
# else		/* Linus defined __NR_streams2 rather than __NR_putpmsg */
# define __NR_putpmsg	__NR_streams2
# endif

#endif

#define	NR_putpmsg_2_1	189	/* 2.1 and onward */


#if !defined(DEP) && !defined(__NR_getpmsg)

# ifndef __NR_streams1
# error "unistd.h does not define __NR_getpmsg or __NR_streams1"
# else		/* Linus defined __NR_streams1 rather than __NR_getpmsg */
# define __NR_getpmsg	__NR_streams1
# endif

#endif

#define	NR_getpmsg_2_1	188	/* 2.1 and onward */

/*
 * In head.c
 */
extern int
lis_strputpmsg(struct inode *i, struct file *fp, 
	       void *ctlp, void *datp, int band, int flags) ;
extern int
lis_strgetpmsg(struct inode *i, struct file *fp, 
	       void *ctlp, void *datp, int *bandp, int *flagsp,
	       int doit) ;

/*
 * In poll.c
 */
#if defined(PORTABLE_POLL)
extern int	lis_poll(struct poll *fds, int nfds, int time_out) ;
#endif

/*
 * To lock or not to lock: only SMP kernels need to lock
 */
#if defined(KERNEL_2_1) && defined(__SMP__)
#include <linux/smp_lock.h>
#define	Klock	lock_kernel()
#define	Kunlock	unlock_kernel()
#else
#define	Klock	
#define	Kunlock	
#endif


/*  -------------------------------------------------------------------  */
/* sys_poll -  poll() syscall entry point
 *
 * STATUS: complete, tested
 */
#if defined(PORTABLE_POLL)
asmlinkage int
sys_poll(void *ubuff, ulong n, int tmout)
{
    return(lis_poll((struct poll *) ubuff, n, tmout)) ;

}/*lis_syspoll*/
#endif


/*  -------------------------------------------------------------------  */
/* sys_putpmsg -  putpmsg syscall entry point
 *
 * STATUS: complete, tested
 */
asmlinkage int
sys_putpmsg(int fd,
	    struct strbuf *ctl,
	    struct strbuf *dat,
	    int band,
	    int flags)
{
    struct file		* file;
    struct inode	* inode;
    int			  ret ;

    /*
     * The following checks are patterned after sys_write
     */
    Klock ;
#if defined(KERNEL_2_1)
    if (   (file = fget(fd)) == NULL
	|| file->f_dentry == NULL
	|| (inode = file->f_dentry->d_inode) == NULL
       )
    {
	if (file != NULL)
	    fput(file) ;

	Kunlock ;
	return -EBADF;
    }
#else
    if (fd>=NR_OPEN || !(file=current->files->fd[fd]) || !(inode=file->f_inode))
    {
	Kunlock ;
	return -EBADF;
    }
#endif
    if (!(file->f_mode & 2))
    {
#if defined(KERNEL_2_1)
	fput(file) ;
#endif
	Kunlock ;
	return -EBADF;
    }
    if (!file->f_op || !file->f_op->write)
    {
#if defined(KERNEL_2_1)
	fput(file) ;
#endif
	Kunlock ;
	return -EINVAL;
    }

    down(&inode->i_sem) ;
    ret = lis_strputpmsg(inode,file,ctl,dat,band,flags);
    up(&inode->i_sem) ;
#if defined(KERNEL_2_1)
    fput(file) ;
#endif
    Kunlock ;
    return(ret) ;

}/*sys_putpmsg*/

/*  -------------------------------------------------------------------  */
/* sys_getpmsg -  getpmsg syscall entry point
 *
 * STATUS: complete, tested
 */
asmlinkage int
sys_getpmsg(int fd,
	    struct strbuf *ctl,
	    struct strbuf *dat,
	    int *bandp,
	    int *flagsp)
{
    struct file		* file;
    struct inode	* inode;
    int			  ret ;

    /*
     * The following checks are patterned after sys_read
     */
    Klock ;
#if defined(KERNEL_2_1)
    if (   (file = fget(fd)) == NULL
	|| file->f_dentry == NULL
	|| (inode = file->f_dentry->d_inode) == NULL
       )
    {
	if (file != NULL)
	    fput(file) ;

	Kunlock ;
	return -EBADF;
    }
#else
    if (fd>=NR_OPEN || !(file=current->files->fd[fd]) || !(inode=file->f_inode))
    {
	Kunlock ;
	return -EBADF;
    }
#endif
    if (!(file->f_mode & 1))
    {
#if defined(KERNEL_2_1)
	fput(file) ;
#endif
	Kunlock ;
	return -EBADF;
    }
    if (!file->f_op || !file->f_op->read)
    {
#if defined(KERNEL_2_1)
	fput(file) ;
#endif
	Kunlock ;
	return -EINVAL;
    }

    down(&inode->i_sem) ;
    ret = lis_strgetpmsg(inode,file,ctl,dat,bandp,flagsp,1);
    up(&inode->i_sem) ;
#if defined(KERNEL_2_1)
    fput(file) ;
#endif
    Kunlock ;
    return(ret) ;

}/*sys_getpmsg*/

extern void     *sys_call_table[];

void
hook_syscalls(void)
{
        sys_call_table[__NR_putpmsg]		= (void *)sys_putpmsg;
        sys_call_table[  NR_putpmsg_2_1]	= (void *)sys_putpmsg;
        sys_call_table[__NR_getpmsg]		= (void *)sys_getpmsg;
        sys_call_table[  NR_getpmsg_2_1]	= (void *)sys_getpmsg;
#if defined(PORTABLE_POLL)
        sys_call_table[__NR_poll]		= (void *)sys_poll;
#endif
        return;
}

void
restore_syscalls(void)
{
        sys_call_table[__NR_putpmsg]		= 0;
        sys_call_table[  NR_putpmsg_2_1]	= 0;
        sys_call_table[__NR_getpmsg]		= 0;
        sys_call_table[  NR_getpmsg_2_1]	= 0;
#if defined(PORTABLE_POLL)
        sys_call_table[__NR_poll]		= 0;
#endif
        return;
}

#endif				/* LINUX */
