/*                               -*- Mode: C -*- 
 * wait.c --- wait queues management
 * Author          : Francisco J. Ballesteros
 * Created On      : Tue May 31 22:25:19 1994
 * Last Modified By: David Grothe
 * RCS Id          : $Id: wait.c,v 1.2 1995/12/27 01:02:53 dave Exp $
 * Purpose         : keep close all the wait related stuff.
 * ----------------______________________________________________
 *
 *   Copyright (C) 1995  Francisco J. Ballesteros
 *   Copyright (C) 1997  David Grothe, Gcom, Inc <dave@gcom.com>
 *
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 * 
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330, Cambridge,
 * MA 02139, USA.
 * 
 *
 *    You can reach me by email to
 *    nemo@ordago.uc3m.es
 */

#ident "@(#) LiS wait.c 2.1 4/4/99 11:34:25 "


/*  -------------------------------------------------------------------  */
/*				 Dependencies                            */

#include <sys/strport.h>
#include <sys/strconfig.h>	/* config definitions */
#include <sys/LiS/share.h>	/* streams shared defs*/
#include <sys/LiS/wait.h>	/* interface */
#include <sys/LiS/head.h>	/* stream head */
#include <sys/LiS/queue.h>	/* streams queues */


/*  -------------------------------------------------------------------  */
/*				   Symbols                               */

/* Timer tick, this will in be the worst case 
 */
#define LIS_TICK	0x00

/*  -------------------------------------------------------------------  */
/*				 Global vars                             */

/*  -------------------------------------------------------------------  */
/*			Exported functions & macros                      */

/* The functions below are to achieve mutual exclusion on the stream
 * head. They use the sd_wopen wait queue to sleep and the sd_lock char to
 * flag the lock.
 * The lock/unlock are the complete ones.
 * if you just want to wait you may use wait_on/wake_up.
 *
 * You can lock(L)/wait(W)/sleep(S) on the following events:
 * L/W wopen:  the 1st opener completes its job
 * W/S wwrite: there's room in the write queue to issue a downstream msg.
 * STATUS: complete(?), untested
 */
/*  -------------------------------------------------------------------  */
int 
lis_sleep_on_wopen(struct stdata * sd)
{
    int		rslt ;

    rslt = lis_down(&sd->sd_wopen);
    if (rslt < 0) return(rslt) ;
    F_SET(sd->sd_flag,STWOPEN);
    return(rslt) ;

}/*lis_sleep_on_wopen*/
/*  -------------------------------------------------------------------  */
int 
lis_sleep_on_wioc(struct stdata * sd)
{
    int		rslt ;

    rslt = lis_down(&sd->sd_wioc);
    if (rslt < 0) return(rslt) ;
    return(rslt) ;

}/*lis_sleep_on_wopen*/
/*  -------------------------------------------------------------------  */
int 
lis_sleep_on_wwrite(struct stdata * sd)
{

    return(lis_down(&sd->sd_wwrite));

}/*lis_sleep_on_wwrite*/
/*  -------------------------------------------------------------------  */
int 
lis_sleep_on_wread(struct stdata * sd)
{

    return(lis_down(&sd->sd_wread));

}/*lis_sleep_on_wread*/

/*  -------------------------------------------------------------------  */
int 
lis_sleep_on_wiocing(struct stdata * sd)
{

    return(lis_down(&sd->sd_wiocing));

}/*lis_sleep_on_wiocing*/




/*----------------------------------------------------------------------
# Local Variables:      ***
# change-log-default-name: "~/src/prj/streams/src/NOTES" ***
# End: ***
  ----------------------------------------------------------------------*/
