/*                               -*- Mode: C -*- 
 * head.h --- streams head handling
 * Author          : Graham Wheeler, Francisco J. Ballesteros
 * Created On      : Tue May 31 22:25:19 1994
 * Last Modified By: David Grothe
 * RCS Id          : $Id: head.h,v 1.13 1996/01/28 21:59:25 dave Exp $
 * Purpose         : here you have utilites to handle str heads.
 * ----------------______________________________________________
 *
 *    Copyright (C) 1995  Graham Wheeler, Francisco J. Ballesteros,
 *                        Denis Froschauer
 *    Copyright (C) 1997  David Grothe, Gcom, Inc <dave@gcom.com>
 *
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 * 
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330, Cambridge,
 * MA 02139, USA.
 * 
 *
 *    You can reach us by email to any of
 *    gram@aztec.co.za, nemo@ordago.uc3m.es, 100741.1151@compuserve.com
 *    dave@gcom.com
 */

#ifndef _HEAD_H
#define _HEAD_H 1

#ident "@(#) LiS head.h 2.1 4/4/99 11:34:29 "

/*  -------------------------------------------------------------------  */
/*				 Dependencies                            */

#ifndef _STRPORT_H
#include <sys/strport.h>	/* porting environment tailoring */
#endif
#ifndef _LIS_CONFIG_H
#include <sys/strconfig.h>	/* streams config symbols are here */
#endif
#ifndef _STR_MSG_H
#include <sys/strmsg.h>		/* streams msgs */
#endif
#ifndef _STR_OPTS_H
#include <sys/stropts.h>	/* streams operations */
#endif
#ifndef _MSG_H
#include <sys/LiS/msg.h>	/* streams msg symbols & types */
#endif
#ifndef _MSGUTL_H
#include <sys/LiS/msgutl.h>	/* streams msg utilities  */
#endif
#ifndef _QUEUE_H
#include <sys/LiS/queue.h>	/* streams queue symbols & types */
#endif
#ifndef _EVENTS_H
#include <sys/LiS/events.h>	/* streams events  */
#endif
#ifndef _MOD_H
#include <sys/LiS/mod.h>	/* streams module symbols & types */
#endif
#if !defined(KERNEL_2_1)
# ifndef _POLL_H
# include <sys/LiS/poll.h>	/* streams module symbols & types */
# endif
#endif

/*  -------------------------------------------------------------------  */
/*				   Symbols                               */

/* Close time max.
 */
#define LIS_MAX_CLTIME	1000*60*5	/* in m.seconds -> 5min */

/* Link time, default & max.
 */
#define LIS_MAX_LNTIME  1000*60*5
#define LIS_LNTIME	1000*15

/* str head default high/low water marks
 */
#define STRHIGH 5120
#define STRLOW  1024

/* str head flags 
 */
#define STWOPEN		0x00000001L /* someone's opening this */
#define	RSLEEP		0x00000002L /* someone wants to read/recv msg */
#define	WSLEEP		0x00000004L /* someone wants to write */
#define	IOCWAIT		0x00000008L /* someone wants to do ioctl */
#define	STRCLOSE	0x00000010L /*!wait for a close to complete */
#define	STRPLUMB	0x00000020L /*!push/pop pending */
#define STRHOLD		0x00000040L /* use stream hold on write feature */
#define	STPLEX		0x00001000L /* multiplexed stream */
#define	STRISTTY	0x00002000L /* stream is a terminal */
#define	STRTOSTOP	0x00004000L /* block background writes */
#define	STRSIGPIPE	0x00008000L /* send SIGPIPE on write errors */
#define	STRPRI		0x00010000L /* An M_PCPROTO is at stream head */
#define	STRDERR		0x00020000L /* fatal read error from M_ERROR */
#define	STWRERR		0x00040000L /* fatal write error from M_ERROR */
#define	STRHUP		0x00080000L /* Device has vanished */
#define	SNDMREAD	0x00100000L /* used for read notification */
#define	OLDNDELAY	0x00200000L /* use old TTY semantics for */
				    /* NDELAY reads and writes */
#define	STRDELIM	0x00400000L /* generate delimited messages ???*/
#define	STRSELPND	0x00800000L /* select waiting on this stream */
#define STIOCTMR	0x01000000L /* ioctl timer in progress */
/*  -------------------------------------------------------------------  */

/* Should be in head->magic, it's used to cach 254/255 ;) 
 * danglind references to a stream heads in critical places
 */
#define STDATA_MAGIC (0x11110000L | sizeof(stdata_t))



/*  -------------------------------------------------------------------  */
/*				    Types                                */

/*
 * This structure is embedded in the stdata struct.  It is used for
 * maintaining the topology of multiplexors.  When the user does
 * an ioctl(fd, I_LINK, muxfd), the stream represented by muxfd
 * is linked "below" the stream driver represented by fd.  The
 * stdata struct of fd uses mx_hd as a list head to thread together
 * all such streams that were linked below it.  The mx_next field
 * carries the links.
 *
 * The muxfd stream is marked as belonging to a multiplexor and
 * cannot be operated on with file I/O routines, except close.
 *
 * The muxfd stream has its refcnt incremented so that if it is
 * closed it will not be deallocated.
 *
 * When it is time to unlink the stream it is removed from its list.
 * The refcnt is then decremented and if it reaches zero the stream
 * is deallocated.
 *
 * If you close the stream that heads a list (a so-called "control stream")
 * then its list is traversed and each stream on the list is unlinked.
 *
 * The I_LINK ioctl returns a reference number, l_index, that is used
 * by the user to unlink the stream with ioctl(fd, I_UNLINK, l_index).
 * The list of streams headed by fd is scanned and the one with the
 * mx_index equal to l_index is unlinked.
 */
#ifdef __KERNEL__
typedef 
struct stmux
{
    int            mx_cmd;	/* I_LINK or I_PLINK */
    int		   mx_index;	/* l_index for this mux */
    struct stdata *mx_hd;	/* list head for control stream */
    struct stdata *mx_next;	/* list threaded through here */

} stmux_t;
#endif


/*
 * The stdata struct. This is the main stream structure and is pointed by
 * the file system inode for the stream.
 * It correspond with the STREAM head.
 */

#ifdef __KERNEL__
typedef
struct stdata
{
        long              magic; 	/* should be always STDATA_MAGIC */
	struct stdata	 *sd_next ;	/* all stdatas are linked together */
	struct stdata	 *sd_prev ;
        struct streamtab *sd_strtab;    /* pointer to streamtab for stream */
        struct queue     *sd_wq;        /* write queue */
	mblk_t           *sd_wmsg;      /* buff for write msg */
	mblk_t           *sd_iocblk; 	/* data block for ioctl */
	int  sd_open_flags ;		/* flags that opened stream */
	int  sd_scantimer;		/* scanq timer handle */
	int  sd_iocseq;			/* ioc seq # */
        int  sd_session;                /* controlling session id */
        int  sd_pgrp;                   /* controlling process group id */
        long sd_flag;                   /* state/flags */
        int  sd_pushcnt;                /* number of pushes done on stream */
        int  sd_sigflags;               /* logical OR of all siglist events */
        int  sd_events;                 /* logical OR of all eventlist events*/
        int  sd_rdopt;			/* read options */
        int  sd_wropt;			/* write options */
        int  sd_closetime;              /* time to wait to drain q in close */
        int  sd_close_timer;		/* timer handle for close timer */
        int  sd_rerror;                 /* read error to set u.u_error */
        int  sd_werror;                 /* write error to set u.u_error */
	int  sd_maxpushcnt;	        /* currently there's no limit in 
					 * # of pushed mods, but let's set up
					 * an artificial one to aid debugging*/
        long sd_maxpsz;                 /* max pkt size --should be a cache*/
        long sd_minpsz;                 /* min pkt size   of below module's */
	unsigned short     sd_wroff;	/* write offset for downstream data */
	struct strevent   *sd_siglist;  /* processes to be sent SIGPOLL */
#if defined(PORTABLE_POLL)
	struct pollhead    sd_polllist;	/* polling processes*/
#elif defined(LINUX_POLL)
	struct wait_queue *sd_task_list;/* tasks waiting on poll */
#else
#error "Either PORTABLE_POLL or LINUX_POLL must be defined"
#endif
	struct semaphore   sd_wopen;	/* Waiting for open/close to complete*/
        struct semaphore   sd_wwrite;   /* wait to room for write */
        struct semaphore   sd_wread;    /* wait for msg to arrive */
        struct semaphore   sd_wioc;     /* wait for ioctl */
	struct semaphore   sd_wiocing;  /* wait for iocl response */
	struct semaphore   sd_closing;	/* waiting to close */
	unsigned short     sd_refcnt;	/* reference count */
	unsigned short     sd_linkcnt;	/* # of I_PLINKs done via stream */
        struct stmux       sd_mux;      /* info for muxing streams */
        int		   sd_l_index;	/* muxid cntr for stream head */
        dev_t		   sd_dev ;	/* major/minor from inode */
#ifdef PORTABLE_POLL
	lis_select_t	   sd_select ;	/* abstract select structure */
					/* see *-mdep.h */
#endif
} stdata_t;

extern stdata_t		*lis_stdata_head ;	/* to list of stdatas */

#endif



/*  -------------------------------------------------------------------  */
/*				 Glob. Vars.                             */

/* Scheduling  & scan list
 */
#ifdef __KERNEL__
extern volatile queue_t *lis_sched_first, *lis_sched_last; /* sched list */
extern struct queue	*lis_qhead; /* first scheduled queue	*/
extern struct queue	*lis_qtail; /* last scheduled queue		*/
extern struct queue	*lis_scanqhead;	/* head of STREAMS scan queue	*/
extern struct queue	*lis_scanqtail;	/* tail of STREAMS scan queue	*/
extern char lis_qrunflag;	/*set if there is at least one enabled queue*/
extern char lis_queueflag;	/*the function queuerun is running    	*/

/* cfg. opts.
 */
extern int lis_nstrpush;	/* maximum # of pushed modules */
extern int lis_strhold;		/* if not zero str hold feature's activated*/
extern unsigned long lis_strthresh;	/* configurable STREAMS memory limit */
extern unsigned long lis_iocseq; /* ioctl id */
#endif


/*  -------------------------------------------------------------------  */
/*			Exported functions & macros                      */


/* The streams file ops
 * these are the entry points into the LiS subsystem.
 * in response to a system call the Linux kernel will dispatch to
 * one of these entries.
 * -- see "The design of the unix operating system" (Bach)
 */
#ifdef __KERNEL__
extern int 
lis_stropen( struct inode *, struct file *);
extern int
lis_strioctl(struct inode *, struct file *, unsigned int, unsigned long);
extern int
lis_strputpmsg(struct inode *, struct file *, void *, void *, int, int);
extern int
lis_strgetpmsg(struct inode *, struct file *, void *, void *, int *,int *,int);
extern int
lis_strpoll(struct inode *i, struct file *f, void *ptr) ;
#if defined(KERNEL_2_1)
extern int
lis_strclose(struct inode *i, struct file *f);
extern ssize_t
lis_strwrite(struct file *, const char *,  size_t, loff_t *);
extern ssize_t
lis_strread(struct file *, char *,  size_t, loff_t *);
#else
extern void
lis_strclose(struct inode *i, struct file *f);
extern int
lis_strwrite(struct inode *, struct file *, const char *,  int);
extern int
lis_strread(struct inode *, struct file *, char *,  int);
#endif

/* Initialize some glob vars...
 */
extern void
 lis_init_head( void );

extern void
 lis_terminate_head( void );

/*
 * Hook a new inode onto a stream head.
 */
int	lis_new_inode(struct inode *oldi, struct inode *newi) ;

/*
 * Run the STREAMS queues
 */
void	lis_run_queues(void) ;

#endif				/* __KERNEL__ */

/*  -------------------------------------------------------------------  */
#endif /*!_HEAD_H*/

/*----------------------------------------------------------------------
# Local Variables:      ***
# change-log-default-name: "~/src/prj/streams/src/NOTES" ***
# End: ***
  ----------------------------------------------------------------------*/
