/************************************************************************
*                      Operating System Interface                       *
*************************************************************************
*									*
* These are routines that call kernel routines indirectly.  This allows	*
* driver writers to reference the lis_... names in compiled code	*
* and the osif.c module itself will use the mangled names of the	*
* kernel routines when "modversions" is set.				*
*									*
* Include this file AFTER any linux includes.				*
*									*
************************************************************************/
#ifndef OSIF_H
#define OSIF_H		/* file included */

#ident "@(#) LiS osif.h 1.5 6/17/99"

#if !defined(INCL_FROM_OSIF_DRIVER)

/*
 * Redefine the "official" kernel names of these routines to the
 * "lis_..." names.
 */

#define	pcibios_present			lis_pcibios_present
#define	pcibios_init			lis_pcibios_init
#define	pcibios_find_class		lis_pcibios_find_class
#define	pcibios_find_device		lis_pcibios_find_device
#ifdef pcibios_read_config_byte
#undef pcibios_read_config_byte
#endif
#define	pcibios_read_config_byte	lis_pcibios_read_config_byte
#define	pcibios_read_config_word	lis_pcibios_read_config_word
#define	pcibios_read_config_dword	lis_pcibios_read_config_dword
#define	pcibios_write_config_byte	lis_pcibios_write_config_byte
#define	pcibios_write_config_word	lis_pcibios_write_config_word
#define	pcibios_write_config_dword	lis_pcibios_write_config_dword
#define	pcibios_strerror		lis_pcibios_strerror

#define	request_irq			lis_request_irq
#define	free_irq			lis_free_irq

#define	ioremap				lis_ioremap
#define	ioremap_nocache			lis_ioremap_nocache
#define	vremap				lis_vremap
#define virt_to_phys			lis_virt_to_phys
#define phys_to_virt			lis_phys_to_virt

#define	check_region			lis_check_region
#define	request_region			lis_request_region
#define	release_region			lis_release_region

#define	kmalloc				lis_kmalloc
#define	kfree				lis_kfree

#define	request_dma			lis_request_dma
#define	free_dma			lis_free_dma

#endif

#ifndef LINUX_VERSION_CODE
#include <linux/version.h>
#endif

/*
 * PCI BIOS routines
 */
int lis_pcibios_present(void) ;
#if LINUX_VERSION_CODE < 0x020100		/* 2.0 kernel */
unsigned long lis_pcibios_init(unsigned long memory_start,
			        unsigned long memory_end) ;
#else						/* 2.1 or 2.2 kernel */
void lis_pcibios_init(void) ;
#endif
int lis_pcibios_find_class(unsigned int   class_code,
			    unsigned short index,
                            unsigned char *bus,
			    unsigned char *dev_fn) ;
int lis_pcibios_find_device(unsigned short vendor,
			     unsigned short dev_id,
			     unsigned short index,
			     unsigned char *bus,
			     unsigned char *dev_fn) ;
int lis_pcibios_read_config_byte(unsigned char  bus,
				  unsigned char  dev_fn,
				  unsigned char  where,
				  unsigned char *val) ;
int lis_pcibios_read_config_word(unsigned char   bus,
				  unsigned char   dev_fn,
				  unsigned char   where,
				  unsigned short *val) ;
int lis_pcibios_read_config_dword(unsigned char  bus,
				   unsigned char  dev_fn,
				   unsigned char  where,
				   unsigned int  *val) ;
int lis_pcibios_write_config_byte(unsigned char  bus,
				   unsigned char  dev_fn,
				   unsigned char  where,
				   unsigned char  val) ;
int lis_pcibios_write_config_word(unsigned char   bus,
				   unsigned char   dev_fn,
				   unsigned char   where,
				   unsigned short  val) ;
int lis_pcibios_write_config_dword(unsigned char  bus,
				    unsigned char  dev_fn,
				    unsigned char  where,
				    unsigned int   val) ;
const char *lis_pcibios_strerror(int error) ;

/*
 * New style PCI interface
 */
#if LINUX_VERSION_CODE >= 0x020100		/* 2.0 kernel */
/* TBD */
#endif

/*
 * IRQ routines
 */
#if defined(_LINUX_PTRACE_H)
#define	OSIF_REGS_T	struct pt_regs
#else
#define	OSIF_REGS_T	void
#endif
int  lis_request_irq(unsigned int  irq,
	              void        (*handler)(int, void *, OSIF_REGS_T *),
	              unsigned long flags,
		      const char   *device,
		      void         *dev_id) ;
void lis_free_irq(unsigned int irq, void *dev_id) ;

/*
 * Memory mapping routines
 */
void *lis_ioremap(unsigned long offset, unsigned long size) ;
void *lis_ioremap_nocache(unsigned long offset, unsigned long size) ;
void *lis_vremap(unsigned long offset, unsigned long size) ;
unsigned long lis_virt_to_phys(volatile void *addr) ;
void         *lis_phys_to_virt(unsigned long addr) ;

/*
 * I/O port routines <linux/ioport.h>
 */
int  lis_check_region(unsigned int from, unsigned int extent) ;
void lis_request_region(unsigned int from,
			 unsigned int extent,
			 const char  *name) ;
void lis_release_region(unsigned int from, unsigned int extent) ;


/*
 * Memory allocator <linux/malloc.h>
 */
void *lis_kmalloc(size_t nbytes, int type) ;
void  lis_kfree(const void *ptr) ;


/*
 * DMA routines <asm/dma.h>
 */
int  lis_request_dma(unsigned int dma_nr, const char *device_id) ;
void lis_free_dma(unsigned int dma_nr) ;



#endif			/* from top of file */
