#!/bin/sh

. config.in

DIR=`pwd`

check_semaphore_patch()
{
    echo "Testing to see if the semaphore patch is needed."

    grep -s Grothe ./include/asm-i386/semaphore.h > /dev/null 2>&1

    if [ "$?" != "0" ]
    then
       echo "The semaphore patch needs to be added to your system."
       echo "Updating the kernel now."
       patch -p1 < $DIR/patches.semaphores > /tmp/plog 2>&1
       err=$?
       grep -s fail /tmp/plog
       if [ "$?" = "0" -o $err -ne 0 ]
       then
	  echo "An error was detected when patching the kernel source"
	  echo "for the semaphores.  Please check the file /tmp/plog"
	  echo "and fix the problem if possible."
	  exit 1
       fi
       echo "The semaphore patch was successful."
    else
      echo "The semaphore patch is not needed on this system."
    fi
}

chk_installed()
{
    if [ $PRESENT -ne 0 ]; then
	echo "STREAMS hooks already installed in $KVER kernel source at $KSRC"
	exit 0
    fi
}


PRESENT=0
if grep -s -q __NR_putpmsg $KSRC/include/asm/unistd.h ||
   grep -s -q __NR_streams $KSRC/include/asm/unistd.h
then
    PRESENT=1
fi

case $KVER in
    2.0.*)
	v=$KVER_PATCH
	if   [ $v -ge 36 ]; then		# 36 and up
	    if [ "$STRMS_QUEUES" = "b" ]; then
		echo "No patch needed for $KVER kernel using \"bottom half\"."
		exit 0
	    fi

	    if grep -s -q exit_files $KSRC/kernel/ksyms.c; then
		exit 0		# patch installed
	    fi
	    cd $KSRC
	    patch -p1 <$DIR/patches_ins.ksyms.36
	    exit $?
	elif [ $v -ge 35 ]; then		# 35 
	    chk_installed

	    cd $KSRC
	    patch -p1 <$DIR/patches_ins.hooks
	    exit $?
	elif [ $v -ge 28 ]; then		# 28 - 34
	    chk_installed

	    cd $KSRC
	    patch -p1 <$DIR/patches.kernel.28-34
	    exit $?
	elif [ $v -ge 24 ]; then		# 24 - 27
	    cd $KSRC
	    check_semaphore_patch
	    chk_installed

	    patch -p1 <$DIR/patches.kernel.24-27
	    exit $?
	else
	    echo "Cannot install STREAMS hooks in kernel version $KVER."
	    echo "Must be at lease kernel version 2.0.24 to install hooks."
	    echo "No hooks needed starting with 2.0.36."
	    exit 1
	fi
	;;
    2.1.*)
	chk_installed

	v=$KVER_PATCH
	if [ $v -ge 115 ]; then
	    echo "Kernel version $KVER should have the STREAMS"
	    echo "patches in the standard source (but doesn\'t)."
	else
	    echo "You need to use kernel version 2.1.115 or later"
	    echo "to install STREAMS."
	fi
	exit 1
	;;
    2.2.*)
	chk_installed

	echo "Kernel version $KVER should have the STREAMS"
	echo "patches in the standard source (but doesn\'t)."
	exit 1
	;;
    *)
	echo "cannot determine linux version from version.h."
	exit 1
	;;
esac
